// FB Alpha CPS-1 & CPS Changer driver module

#include "cps.h"
#include "bitswap.h"

//----------------------------------------------------------------------------------------------
// Game inputs

static struct BurnInputInfo wonders3InputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"   },
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"  },
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"     },
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"   },
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"   },
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"  },
	{"P1 Button 1"   , BIT_DIGITAL, CpsInp001+4, "p1 fire 1" },
	{"P1 Button 2"   , BIT_DIGITAL, CpsInp001+5, "p1 fire 2" },

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"   },
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"  },
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"     },
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"   },
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"   },
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"  },
	{"P2 Button 1"   , BIT_DIGITAL, CpsInp000+4, "p2 fire 1" },
	{"P2 Button 2"   , BIT_DIGITAL, CpsInp000+5, "p2 fire 2" },

	{"Reset"         , BIT_DIGITAL, &CpsReset  , "reset"     },
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"      },
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"   },
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"     },
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"     },
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"     },
};

STDINPUTINFO(wonders3);

static struct BurnInputInfo NTFOInputList[] =
{
	{"P1 Coin"    , BIT_DIGITAL, CpsInp018+0, "p1 coin"   },
	{"P1 Start"   , BIT_DIGITAL, CpsInp018+4, "p1 start"  },
	{"P1 Up"      , BIT_DIGITAL, CpsInp001+3, "p1 up"     },
	{"P1 Down"    , BIT_DIGITAL, CpsInp001+2, "p1 down"   },
	{"P1 Left"    , BIT_DIGITAL, CpsInp001+1, "p1 left"   },
	{"P1 Right"   , BIT_DIGITAL, CpsInp001+0, "p1 right"  },
	{"P1 Shot"    , BIT_DIGITAL, CpsInp001+4, "p1 fire 1" },
	{"P1 Special" , BIT_DIGITAL, CpsInp001+5, "p1 fire 2" },

	{"P2 Coin"    , BIT_DIGITAL, CpsInp018+1, "p2 coin"   },
	{"P2 Start"   , BIT_DIGITAL, CpsInp018+5, "p2 start"  },
	{"P2 Up"      , BIT_DIGITAL, CpsInp000+3, "p2 up"     },
	{"P2 Down"    , BIT_DIGITAL, CpsInp000+2, "p2 down"   },
	{"P2 Left"    , BIT_DIGITAL, CpsInp000+1, "p2 left"   },
	{"P2 Right"   , BIT_DIGITAL, CpsInp000+0, "p2 right"  },
	{"P2 Shot"    , BIT_DIGITAL, CpsInp000+4, "p2 fire 1" },
	{"P2 Special" , BIT_DIGITAL, CpsInp000+5, "p2 fire 2" },

	{"Reset"      , BIT_DIGITAL, &CpsReset,   "reset"     },
	{"Diagnostic" , BIT_DIGITAL, CpsInp018+6, "diag"      },
	{"Service"    , BIT_DIGITAL, CpsInp018+2, "service"   },
	{"Dip A"      , BIT_DIPSWITCH, &Cpi01A    , "dip"     },
	{"Dip B"      , BIT_DIPSWITCH, &Cpi01C    , "dip"     },
	{"Dip C"      , BIT_DIPSWITCH, &Cpi01E    , "dip"     },
};

STDINPUTINFO(NTFO);


static struct BurnInputInfo captcommInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"  },
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start" },
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"    },
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"  },
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"  },
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right" },
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"  },
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start" },
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"    },
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"  },
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"  },
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right" },
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},

	{"P3 Coin"       , BIT_DIGITAL, CpsInp177+6, "p3 coin"  },
	{"P3 Start"      , BIT_DIGITAL, CpsInp177+7, "p3 start" },
	{"P3 Up"         , BIT_DIGITAL, CpsInp177+3, "p3 up"    },
	{"P3 Down"       , BIT_DIGITAL, CpsInp177+2, "p3 down"  },
	{"P3 Left"       , BIT_DIGITAL, CpsInp177+1, "p3 left"  },
	{"P3 Right"      , BIT_DIGITAL, CpsInp177+0, "p3 right" },
	{"P3 Attack"     , BIT_DIGITAL, CpsInp177+4, "p3 fire 1"},
	{"P3 Jump"       , BIT_DIGITAL, CpsInp177+5, "p3 fire 2"},

	{"P4 Coin"       , BIT_DIGITAL, CpsInp179+6, "p4 coin"  },
	{"P4 Start"      , BIT_DIGITAL, CpsInp179+7, "p4 start" },
	{"P4 Up"         , BIT_DIGITAL, CpsInp179+3, "p4 up"    },
	{"P4 Down"       , BIT_DIGITAL, CpsInp179+2, "p4 down"  },
	{"P4 Left"       , BIT_DIGITAL, CpsInp179+1, "p4 left"  },
	{"P4 Right"      , BIT_DIGITAL, CpsInp179+0, "p4 right" },
	{"P4 Attack"     , BIT_DIGITAL, CpsInp179+4, "p4 fire 1"},
	{"P4 Jump"       , BIT_DIGITAL, CpsInp179+5, "p4 fire 2"},

	{"Reset"         , BIT_DIGITAL, &CpsReset  , "reset"    },
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"     },
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"  },
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"    },
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"    },
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"    },
};

STDINPUTINFO(captcomm);

static struct BurnInputInfo cawingInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"   },
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"  },
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"     },
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"   },
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"   },
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"  },
	{"P1 Shot"       , BIT_DIGITAL, CpsInp001+4, "p1 fire 1" },
	{"P1 Special"    , BIT_DIGITAL, CpsInp001+5, "p1 fire 2" },
	{"P1 Fire 3"     , BIT_DIGITAL, CpsInp001+6, "p1 fire 3" },

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"   },
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"  },
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"     },
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"   },
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"   },
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"  },
	{"P2 Shot"       , BIT_DIGITAL, CpsInp000+4, "p2 fire 1" },
	{"P2 Special"    , BIT_DIGITAL, CpsInp000+5, "p2 fire 2" },
	{"P2 Fire 3"     , BIT_DIGITAL, CpsInp000+6, "p2 fire 3" },

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"     },
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"      },
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"   },
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"     },
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"     },
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"     },
};

STDINPUTINFO(cawing);

static struct BurnInputInfo cworld2jInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Answer 1"   , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Answer 2"   , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Answer 3"   , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},
	{"P1 Answer 4"   , BIT_DIGITAL, CpsInp001+7, "p1 fire 4"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Answer 1"   , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Answer 2"   , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Answer 3"   , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},
	{"P2 Answer 4"   , BIT_DIGITAL, CpsInp000+7, "p2 fire 4"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(cworld2j);

static struct BurnInputInfo dinoInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"  },
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start" },
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"    },
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"  },
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"  },
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right" },
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"  },
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start" },
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"    },
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"  },
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"  },
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right" },
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},

	{"P3 Coin"       , BIT_DIGITAL, CpsInpc001+6, "p3 coin"  },
	{"P3 Start"      , BIT_DIGITAL, CpsInpc001+7, "p3 start" },
	{"P3 Up"         , BIT_DIGITAL, CpsInpc001+3, "p3 up"    },
	{"P3 Down"       , BIT_DIGITAL, CpsInpc001+2, "p3 down"  },
	{"P3 Left"       , BIT_DIGITAL, CpsInpc001+1, "p3 left"  },
	{"P3 Right"      , BIT_DIGITAL, CpsInpc001+0, "p3 right" },
	{"P3 Attack"     , BIT_DIGITAL, CpsInpc001+4, "p3 fire 1"},
	{"P3 Jump"       , BIT_DIGITAL, CpsInpc001+5, "p3 fire 2"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(dino);

static struct BurnInputInfo dwInputList[] =
{
	{"P1 Coin"         , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"        , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"           , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"         , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"         , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"        , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Attack Left"  , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Attack Right" , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Special"      , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},

	{"P2 Coin"         , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"        , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"           , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"         , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"         , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"        , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Attack Left"  , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Attack Right" , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Special"      , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},

	{"Reset"           , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"      , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"         , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"           , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"           , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"           , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(dw);

static struct BurnInputInfo ffightInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"   },
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"  },
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"     },
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"   },
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"   },
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"  },
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1" },
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2" },
	{"P1 Special"    , BIT_DIGITAL, CpsInp001+6, "p1 fire 3" },

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"   },
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"  },
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"     },
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"   },
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"   },
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"  },
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1" },
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2" },
	{"P2 Special"    , BIT_DIGITAL, CpsInp000+6, "p2 fire 3" },

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(ffight);

#define A(a, b, c, d) {a, b, (unsigned char*)(c), d}

static struct BurnInputInfo forgottnInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL,		CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL,		CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL,		CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL,		CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL,		CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL,		CpsInp001+0, "p1 right"},
	{"P1 Attack"     , BIT_DIGITAL,		CpsInp001+4, "p1 fire 1"},
	A("P1 Turn"       , BIT_ANALOG_REL,	&CpsInp055,	 "p1 z-axis"),

	{"P2 Coin"       , BIT_DIGITAL,		CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL,		CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL,		CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL,		CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL,		CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL,		CpsInp000+0, "p2 right"},
	{"P2 Attack"     , BIT_DIGITAL,		CpsInp000+4, "p2 fire 1"},
	A("P2 Turn"       , BIT_ANALOG_REL,	&CpsInp05d,	 "p2 z-axis"),

	{"Reset"         , BIT_DIGITAL,		&CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL,		CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL,		CpsInp018+2, "service"},

	{"Dip A"         , BIT_DIPSWITCH,	&Cpi01A    , "dip"       },
	{"Dip B"         , BIT_DIPSWITCH,	&Cpi01C    , "dip"       },
	{"Dip C"         , BIT_DIPSWITCH,	&Cpi01E    , "dip"       },
};

#undef A

STDINPUTINFO(forgottn);

static struct BurnInputInfo ghoulsInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Fire"       , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Fire"       , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},

	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Reset"         , BIT_DIGITAL, &CpsReset  , "reset"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"       },
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"       },
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"       },
};

STDINPUTINFO(ghouls);

static struct BurnInputInfo knightsInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"   },
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"  },
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"     },
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"   },
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"   },
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"  },
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1" },
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2" },
	
	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"   },
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"  },
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"     },
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"   },
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"   },
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"  },
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1" },
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2" },
	
	{"P3 Coin"       , BIT_DIGITAL, CpsInp177+6, "p3 coin"   },
	{"P3 Start"      , BIT_DIGITAL, CpsInp177+7, "p3 start"  },
	{"P3 Up"         , BIT_DIGITAL, CpsInp177+3, "p3 up"     },
	{"P3 Down"       , BIT_DIGITAL, CpsInp177+2, "p3 down"   },
	{"P3 Left"       , BIT_DIGITAL, CpsInp177+1, "p3 left"   },
	{"P3 Right"      , BIT_DIGITAL, CpsInp177+0, "p3 right"  },
	{"P3 Attack"     , BIT_DIGITAL, CpsInp177+4, "p3 fire 1" },
	{"P3 Jump"       , BIT_DIGITAL, CpsInp177+5, "p3 fire 2" },
	
	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"     },
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"      },
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(knights);

static struct BurnInputInfo kodInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},

	{"P3 Coin"       , BIT_DIGITAL, CpsInp177+6, "p3 coin"  },
	{"P3 Start"      , BIT_DIGITAL, CpsInp177+7, "p3 start" },
	{"P3 Up"         , BIT_DIGITAL, CpsInp177+3, "p3 up"    },
	{"P3 Down"       , BIT_DIGITAL, CpsInp177+2, "p3 down"  },
	{"P3 Left"       , BIT_DIGITAL, CpsInp177+1, "p3 left"  },
	{"P3 Right"      , BIT_DIGITAL, CpsInp177+0, "p3 right" },
	{"P3 Attack"     , BIT_DIGITAL, CpsInp177+4, "p3 fire 1"},
	{"P3 Jump"       , BIT_DIGITAL, CpsInp177+5, "p3 fire 2"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(kod);

static struct BurnInputInfo megamanInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Fire"       , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Select"     , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Fire"       , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Select"     , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},

	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(megaman);

static struct BurnInputInfo mercsInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Shot"       , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Special"    , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"  },
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Shot"       , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Special"    , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},

	{"P3 Coin"       , BIT_DIGITAL, CpsInp177+6, "p3 coin"},
	{"P3 Start"      , BIT_DIGITAL, CpsInp177+7, "p3 start" },
	{"P3 Up"         , BIT_DIGITAL, CpsInp177+3, "p3 up"    },
	{"P3 Down"       , BIT_DIGITAL, CpsInp177+2, "p3 down"  },
	{"P3 Left"       , BIT_DIGITAL, CpsInp177+1, "p3 left"  },
	{"P3 Right"      , BIT_DIGITAL, CpsInp177+0, "p3 right" },
	{"P3 Shot"       , BIT_DIGITAL, CpsInp177+4, "p3 fire 1"},
	{"P3 Special"    , BIT_DIGITAL, CpsInp177+5, "p3 fire 2"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(mercs);

static struct BurnInputInfo mswordInputList[] =
{
	{"P1 Coin"       , 0, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , 0, CpsInp018+4, "p1 start"},
	{"P1 Up"         , 0, CpsInp001+3, "p1 up"},
	{"P1 Down"       , 0, CpsInp001+2, "p1 down"},
	{"P1 Left"       , 0, CpsInp001+1, "p1 left"},
	{"P1 Right"      , 0, CpsInp001+0, "p1 right"},
	{"P1 Attack"     , 0, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , 0, CpsInp001+5, "p1 fire 2"},
	{"P1 Fire 3"     , 0, CpsInp001+6, "p1 fire 3"},

	{"P2 Coin"       , 0, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , 0, CpsInp018+5, "p2 start"},
	{"P2 Up"         , 0, CpsInp000+3, "p2 up"},
	{"P2 Down"       , 0, CpsInp000+2, "p2 down"},
	{"P2 Left"       , 0, CpsInp000+1, "p2 left"},
	{"P2 Right"      , 0, CpsInp000+0, "p2 right"},
	{"P2 Attack"     , 0, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , 0, CpsInp000+5, "p2 fire 2"},
	{"P2 Fire 3"     , 0, CpsInp000+6, "p2 fire 3"},

	{"Reset"         , 0, &CpsReset,   "reset"},
	{"Diagnostic"    , 0, CpsInp018+6, "diag"},
	{"Service"       , 0, CpsInp018+2, "service"},
	{"Dip A"         , 2, &Cpi01A    , "dip"},
	{"Dip B"         , 2, &Cpi01C    , "dip"},
	{"Dip C"         , 2, &Cpi01E    , "dip"},
};

STDINPUTINFO(msword);

static struct BurnInputInfo mtwinsInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Special"    , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},
	
	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Special"    , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},
	
	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(mtwins);

static struct BurnInputInfo nemoInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"  },
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start" },
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"    },
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"  },
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"  },
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right" },
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Fire 3"     , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"  },
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start" },
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"    },
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"  },
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"  },
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right" },
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Fire 3"     , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"    },
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"     },
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(nemo);

static struct BurnInputInfo pang3InputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Shot 1"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Shot 2"     , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Shot 1"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Shot 2"     , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(pang3);

static struct BurnInputInfo pnickjInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Turn 1"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Turn 2"     , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Turn 1"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Turn 2"     , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(pnickj);

static struct BurnInputInfo punisherInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(punisher);

static struct BurnInputInfo qadInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Answer 1"   , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Answer 2"   , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Answer 3"   , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},
	{"P1 Answer 4"   , BIT_DIGITAL, CpsInp001+7, "p1 fire 4"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Answer 1"   , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Answer 2"   , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Answer 3"   , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},
	{"P2 Answer 4"   , BIT_DIGITAL, CpsInp000+7, "p2 fire 4"},

	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},

	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(qad);

static struct BurnInputInfo Sf2ueInputList[]=
{
	{"P1 Coin",			BIT_DIGITAL, CpsInp018+0, "p1 coin"},	// 0x00
	{"P1 Start",			BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up",			BIT_DIGITAL, CpsInp001+3, "p1 up"},		// 0x02
	{"P1 Down",			BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left",			BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right",			BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Weak punch",		BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},	// 0x06
	{"P1 Medium punch",		BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Strong punch",		BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},

	{"P1 Weak kick",		BIT_DIGITAL, CpsInp1fd+0, "p1 fire 4"},
	{"P1 Medium kick",		BIT_DIGITAL, CpsInp1fd+1, "p1 fire 5"},
	{"P1 Strong kick",		BIT_DIGITAL, CpsInp1fd+2, "p1 fire 6"},

	{"P2 Coin",			BIT_DIGITAL, CpsInp018+1, "p2 coin"},	// 0x0C
	{"P2 Start",			BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up",			BIT_DIGITAL, CpsInp000+3, "p2 up"},		// 0x0E
	{"P2 Down",			BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left",			BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right",			BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Weak punch",		BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},	// 0x12
	{"P2 Medium punch",		BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Strong Punch",		BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},
	{"P2 Weak kick",		BIT_DIGITAL, CpsInp1fd+4, "p2 fire 4"},
	{"P2 Medium kick",		BIT_DIGITAL, CpsInp1fd+5, "p2 fire 5"},
	{"P2 Strong kick",		BIT_DIGITAL, CpsInp1fd+6, "p2 fire 6"},
	{"Diagnostic",			BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Reset",			BIT_DIGITAL, &CpsReset,	  "reset"},
	{"Dip A",			BIT_DIPSWITCH, &Cpi01A,   "dip"},
	{"Dip B",			BIT_DIPSWITCH, &Cpi01C,   "dip"},
	{"Dip C",			BIT_DIPSWITCH, &Cpi01E,   "dip"},
};

STDINPUTINFO(Sf2ue);


static struct BurnInputInfo sf2InputList[] =
{
	{"P1 Coin",			BIT_DIGITAL, CpsInp018+0, "p1 coin"},	// 0x00
	{"P1 Start",			BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up",			BIT_DIGITAL, CpsInp001+3, "p1 up"},		// 0x02
	{"P1 Down",			BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left",			BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right",			BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Weak punch",		BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},	// 0x06
	{"P1 Medium punch",		BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Strong punch",		BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},
	{"P1 Weak kick",		BIT_DIGITAL, CpsInp177+0, "p1 fire 4"},
	{"P1 Medium kick",		BIT_DIGITAL, CpsInp177+1, "p1 fire 5"},
	{"P1 Strong kick",		BIT_DIGITAL, CpsInp177+2, "p1 fire 6"},
	{"P2 Coin",			BIT_DIGITAL, CpsInp018+1, "p2 coin"},	// 0x0C
	{"P2 Start",			BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up",			BIT_DIGITAL, CpsInp000+3, "p2 up"},		// 0x0E
	{"P2 Down",			BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left",			BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right",			BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Weak punch",		BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},	// 0x12
	{"P2 Medium punch",		BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Strong Punch",		BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},
	{"P2 Weak kick",		BIT_DIGITAL, CpsInp177+4, "p2 fire 4"},
	{"P2 Medium kick",		BIT_DIGITAL, CpsInp177+5, "p2 fire 5"},
	{"P2 Strong kick",		BIT_DIGITAL, CpsInp177+6, "p2 fire 6"},
	{"Diagnostic",			BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Reset",			BIT_DIGITAL, &CpsReset,	  "reset"},
	{"Service",			BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A",			BIT_DIPSWITCH, &Cpi01A,   "dip"},
	{"Dip B",			BIT_DIPSWITCH, &Cpi01C,   "dip"},
	{"Dip C",			BIT_DIPSWITCH, &Cpi01E,   "dip"},
};

STDINPUTINFO(sf2);

static struct BurnInputInfo Sf2yycInputList[]=
{
	{"P1 Coin",			BIT_DIGITAL, CpsInp019+0, "p1 coin"},	// 0x00
	{"P1 Start",			BIT_DIGITAL, CpsInp019+4, "p1 start"},
	{"P1 Up",			BIT_DIGITAL, CpsInp001+3, "p1 up"},		// 0x02
	{"P1 Down",			BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left",			BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right",			BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Weak punch",		BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},	// 0x06
	{"P1 Medium punch",		BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Strong punch",		BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},
	{"P1 Weak kick",		BIT_DIGITAL, CpsInp176+0, "p1 fire 4"},
	{"P1 Medium kick",		BIT_DIGITAL, CpsInp176+1, "p1 fire 5"},
	{"P1 Strong kick",		BIT_DIGITAL, CpsInp176+2, "p1 fire 6"},
	{"P2 Coin",			BIT_DIGITAL, CpsInp019+1, "p2 coin"},	// 0x0C
	{"P2 Start",			BIT_DIGITAL, CpsInp019+5, "p2 start"},
	{"P2 Up",			BIT_DIGITAL, CpsInp000+3, "p2 up"},		// 0x0E
	{"P2 Down",			BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left",			BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right",			BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Weak punch",		BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},	// 0x12
	{"P2 Medium punch",		BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Strong Punch",		BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},
	{"P2 Weak kick",		BIT_DIGITAL, CpsInp176+4, "p2 fire 4"},
	{"P2 Medium kick",		BIT_DIGITAL, CpsInp176+5, "p2 fire 5"},
	{"P2 Strong kick",		BIT_DIGITAL, CpsInp176+6, "p2 fire 6"},
	{"Diagnostic",		BIT_DIGITAL, CpsInp019+6, "diag"},
	{"Reset",			BIT_DIGITAL, &CpsReset,	  "reset"},
	{"Service",			BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A",			BIT_DIPSWITCH, &Cpi01A,   "dip"},
	{"Dip B",			BIT_DIPSWITCH, &Cpi01C,   "dip"},
	{"Dip C",			BIT_DIPSWITCH, &Cpi01E,   "dip"},
};

STDINPUTINFO(Sf2yyc);

static struct BurnInputInfo slammastInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Pin"        , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Pin"        , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},

	{"P3 Coin"       , BIT_DIGITAL, CpsInpc001+6, "p3 coin"},
	{"P3 Start"      , BIT_DIGITAL, CpsInpc001+7, "p3 start"},
	{"P3 Up"         , BIT_DIGITAL, CpsInpc001+3, "p3 up"},
	{"P3 Down"       , BIT_DIGITAL, CpsInpc001+2, "p3 down"},
	{"P3 Left"       , BIT_DIGITAL, CpsInpc001+1, "p3 left"},
	{"P3 Right"      , BIT_DIGITAL, CpsInpc001+0, "p3 right"},
	{"P3 Attack"     , BIT_DIGITAL, CpsInpc001+4, "p3 fire 1"},
	{"P3 Jump"       , BIT_DIGITAL, CpsInpc001+5, "p3 fire 2"},
	{"P3 Pin"        , BIT_DIGITAL, CpsInp001+7 , "p3 fire 3"},

	{"P4 Coin"       , BIT_DIGITAL, CpsInpc003+6, "p4 coin"},
	{"P4 Start"      , BIT_DIGITAL, CpsInpc003+7, "p4 start"},
	{"P4 Up"         , BIT_DIGITAL, CpsInpc003+3, "p4 up"},
	{"P4 Down"       , BIT_DIGITAL, CpsInpc003+2, "p4 down"},
	{"P4 Left"       , BIT_DIGITAL, CpsInpc003+1, "p4 left"},
	{"P4 Right"      , BIT_DIGITAL, CpsInpc003+0, "p4 right"},
	{"P4 Attack"     , BIT_DIGITAL, CpsInpc003+4, "p4 fire 1"},
	{"P4 Jump"       , BIT_DIGITAL, CpsInpc003+5, "p4 fire 2"},
	{"P4 Pin"        , BIT_DIGITAL, CpsInp000+7 , "p4 fire 3"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(slammast);

static struct BurnInputInfo striderInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Fire 3"     , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Fire 3"     , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"       },
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"       },
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"       },
};

STDINPUTINFO(strider);


static struct BurnInputInfo unsquadInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Shot"       , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Special"    , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Fire 3"     , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Shot"       , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Special"    , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Fire 3"     , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},

};

STDINPUTINFO(unsquad);

static struct BurnInputInfo varthInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Shot"       , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Bomb"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Fire 3"     , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Shot"       , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Bomb"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Fire 3"     , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(varth);

static struct BurnInputInfo willowInputList[] =
{
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},
	{"P1 Fire 3"     , BIT_DIGITAL, CpsInp001+6, "p1 fire 3"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},
	{"P2 Fire 3"     , BIT_DIGITAL, CpsInp000+6, "p2 fire 3"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip A"         , BIT_DIPSWITCH, &Cpi01A    , "dip"},
	{"Dip B"         , BIT_DIPSWITCH, &Cpi01C    , "dip"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(willow);

static struct BurnInputInfo wofInputList[] = {
	{"P1 Coin"       , BIT_DIGITAL, CpsInp018+0, "p1 coin"},
	{"P1 Start"      , BIT_DIGITAL, CpsInp018+4, "p1 start"},
	{"P1 Up"         , BIT_DIGITAL, CpsInp001+3, "p1 up"},
	{"P1 Down"       , BIT_DIGITAL, CpsInp001+2, "p1 down"},
	{"P1 Left"       , BIT_DIGITAL, CpsInp001+1, "p1 left"},
	{"P1 Right"      , BIT_DIGITAL, CpsInp001+0, "p1 right"},
	{"P1 Attack"     , BIT_DIGITAL, CpsInp001+4, "p1 fire 1"},
	{"P1 Jump"       , BIT_DIGITAL, CpsInp001+5, "p1 fire 2"},

	{"P2 Coin"       , BIT_DIGITAL, CpsInp018+1, "p2 coin"},
	{"P2 Start"      , BIT_DIGITAL, CpsInp018+5, "p2 start"},
	{"P2 Up"         , BIT_DIGITAL, CpsInp000+3, "p2 up"},
	{"P2 Down"       , BIT_DIGITAL, CpsInp000+2, "p2 down"},
	{"P2 Left"       , BIT_DIGITAL, CpsInp000+1, "p2 left"},
	{"P2 Right"      , BIT_DIGITAL, CpsInp000+0, "p2 right"},
	{"P2 Attack"     , BIT_DIGITAL, CpsInp000+4, "p2 fire 1"},
	{"P2 Jump"       , BIT_DIGITAL, CpsInp000+5, "p2 fire 2"},

	{"P3 Coin"       , BIT_DIGITAL, CpsInpc001+6, "p3 coin"  },
	{"P3 Start"      , BIT_DIGITAL, CpsInpc001+7, "p3 start" },
	{"P3 Up"         , BIT_DIGITAL, CpsInpc001+3, "p3 up"    },
	{"P3 Down"       , BIT_DIGITAL, CpsInpc001+2, "p3 down"  },
	{"P3 Left"       , BIT_DIGITAL, CpsInpc001+1, "p3 left"  },
	{"P3 Right"      , BIT_DIGITAL, CpsInpc001+0, "p3 right" },
	{"P3 Attack"     , BIT_DIGITAL, CpsInpc001+4, "p3 fire 1"},
	{"P3 Jump"       , BIT_DIGITAL, CpsInpc001+5, "p3 fire 2"},

	{"Reset"         , BIT_DIGITAL, &CpsReset,   "reset"},
	{"Diagnostic"    , BIT_DIGITAL, CpsInp018+6, "diag"},
	{"Service"       , BIT_DIGITAL, CpsInp018+2, "service"},
	{"Dip C"         , BIT_DIPSWITCH, &Cpi01E    , "dip"},
};

STDINPUTINFO(wof);

// CPS Changer (sfzch/sfach)
static struct BurnInputInfo cpschInputList[] = {
	{"P1 Pause",		BIT_DIGITAL, CpsInp018+2, ""          },
	{"P1 Start",		BIT_DIGITAL, CpsInp018+4, "p1 start"  },
	{"P1 Up",		BIT_DIGITAL, CpsInp001+3, "p1 up"     },
	{"P1 Down",		BIT_DIGITAL, CpsInp001+2, "p1 down"   },
	{"P1 Left",		BIT_DIGITAL, CpsInp001+1, "p1 left"   },
	{"P1 Right",		BIT_DIGITAL, CpsInp001+0, "p1 right"  },
	{"P1 Weak punch",	BIT_DIGITAL, CpsInp001+4, "p1 fire 1" },
	{"P1 Medium punch",	BIT_DIGITAL, CpsInp001+5, "p1 fire 2" },
	{"P1 Strong punch",	BIT_DIGITAL, CpsInp001+6, "p1 fire 3" },
	{"P1 Weak kick",	BIT_DIGITAL, CpsInp001+7, "p1 fire 4" },
	{"P1 Medium kick",	BIT_DIGITAL, CpsInp018+0, "p1 fire 5" },
	{"P1 Strong kick",	BIT_DIGITAL, CpsInp018+6, "p1 fire 6" },

	{"P2 Pause",		BIT_DIGITAL, CpsInp018+3, ""          },
	{"P2 Start",		BIT_DIGITAL, CpsInp018+5, "p2 start"  },
	{"P2 Up",		BIT_DIGITAL, CpsInp000+3, "p2 up"     },
	{"P2 Down",		BIT_DIGITAL, CpsInp000+2, "p2 down"   },
	{"P2 Left",		BIT_DIGITAL, CpsInp000+1, "p2 left"   },
	{"P2 Right",		BIT_DIGITAL, CpsInp000+0, "p2 right"  },
	{"P2 Weak punch",	BIT_DIGITAL, CpsInp000+4, "p2 fire 1" },
	{"P2 Medium punch",	BIT_DIGITAL, CpsInp000+5, "p2 fire 2" },
	{"P2 Strong punch",	BIT_DIGITAL, CpsInp000+6, "p2 fire 3" },
	{"P2 Weak kick",	BIT_DIGITAL, CpsInp000+7, "p2 fire 4" },
	{"P2 Medium kick",	BIT_DIGITAL, CpsInp018+1, "p2 fire 5" },
	{"P2 Strong kick",	BIT_DIGITAL, CpsInp018+7, "p2 fire 6" },

	{"Reset",		BIT_DIGITAL, &CpsReset,   "reset"     },
};

STDINPUTINFO(cpsch);


//----------------------------------------------------------------------------------------------
// Dip switches


static struct BurnDIPInfo threewondersDIPList[]=
{
	// Defaults
        {0x13, 0xff, 0xff, 0x00, NULL                     },
        {0x14, 0xff, 0xff, 0x66, NULL                     },
        {0x15, 0xff, 0xff, 0x66, NULL                     },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin 1"                 },
	{0x13, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x13, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x13, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x13, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x13, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "Coin 2"                 },
	{0x13, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x13, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x13, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x13, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x13, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x13, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x13, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x13, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "2C to Start, 1 to Cont" },
	{0x13, 0x01, 0x40, 0x00, "Off"                    },
	{0x13, 0x01, 0x40, 0x40, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x13, 0x01, 0x80, 0x00, "Off"                    },
	{0x13, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 4   , "Action Lives"           },
	{0x14, 0x01, 0x03, 0x00, "1"                      },
	{0x14, 0x01, 0x03, 0x01, "2"                      },
	{0x14, 0x01, 0x03, 0x02, "3"                      },
	{0x14, 0x01, 0x03, 0x03, "5"                      },

	{0   , 0xfe, 0   , 4   , "Action Game Level"      },
	{0x14, 0x01, 0x0c, 0x00, "Easy"                   },
	{0x14, 0x01, 0x0c, 0x04, "Normal"                 },
	{0x14, 0x01, 0x0c, 0x08, "Hard"                   },
	{0x14, 0x01, 0x0c, 0x0c, "Hardest"                },

	{0   , 0xfe, 0   , 4   , "Shooting Lives"         },
	{0x14, 0x01, 0x30, 0x00, "1"                      },
	{0x14, 0x01, 0x30, 0x10, "2"                      },
	{0x14, 0x01, 0x30, 0x20, "3"                      },
	{0x14, 0x01, 0x30, 0x30, "5"                      },

	{0   , 0xfe, 0   , 4   , "Shooting Game Level"    },
	{0x14, 0x01, 0xc0, 0x00, "Easy"                   },
	{0x14, 0x01, 0xc0, 0x40, "Normal"                 },
	{0x14, 0x01, 0xc0, 0x80, "Hard"                   },
	{0x14, 0x01, 0xc0, 0xc0, "Hardest"                },

	// Dip C
	{0   , 0xfe, 0   , 4   , "Puzzle Lives"           },
	{0x15, 0x01, 0x03, 0x00, "1"                      },
	{0x15, 0x01, 0x03, 0x01, "2"                      },
	{0x15, 0x01, 0x03, 0x02, "3"                      },
	{0x15, 0x01, 0x03, 0x03, "5"                      },

	{0   , 0xfe, 0   , 4   , "Puzzle Game Level"      },
	{0x15, 0x01, 0x0c, 0x00, "Easy"                   },
	{0x15, 0x01, 0x0c, 0x04, "Normal"                 },
	{0x15, 0x01, 0x0c, 0x08, "Hard"                   },
	{0x15, 0x01, 0x0c, 0x0c, "Hardest"                },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x15, 0x01, 0x10, 0x00, "Off"                    },
	{0x15, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x15, 0x01, 0x20, 0x00, "Off"                    },
	{0x15, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x15, 0x01, 0x40, 0x00, "Off"                    },
	{0x15, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(threewonders);

static struct BurnDIPInfo NTFODIPList[]=
{
        {0x13, 0xff, 0xff, 0x00, NULL                     },
        {0x14, 0xff, 0xff, 0x03, NULL                     },
        {0x15, 0xff, 0xff, 0x60, NULL                     },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin 1"                 },
	{0x13, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x13, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x13, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x13, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x13, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "Coin 2"                 },
	{0x13, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x13, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x13, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x13, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x13, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x13, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x13, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x13, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "2C to Start, 1 to Cont" },
	{0x13, 0x01, 0x40, 0x00, "Off"                    },
	{0x13, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown" },
//	{0x13, 0x01, 0x80, 0x00, "Off"                    },
//	{0x13, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"             },
	{0x14, 0x01, 0x07, 0x00, "00 (Easier)"            },
	{0x14, 0x01, 0x07, 0x01, "01"                     },
	{0x14, 0x01, 0x07, 0x02, "02"                     },
	{0x14, 0x01, 0x07, 0x03, "03"                     },
	{0x14, 0x01, 0x07, 0x04, "04"                     },
	{0x14, 0x01, 0x07, 0x05, "05"                     },
	{0x14, 0x01, 0x07, 0x06, "06"                     },
	{0x14, 0x01, 0x07, 0x07, "07 (Harder)"            },

	{0   , 0xfe, 0   , 4   , "Level Up Timer"         },
	{0x14, 0x01, 0x18, 0x00, "More Slowly"            },
	{0x14, 0x01, 0x18, 0x08, "Slowly"                 },
	{0x14, 0x01, 0x18, 0x10, "Quickly"                },
	{0x14, 0x01, 0x18, 0x18, "More Quickly"           },

	{0   , 0xfe, 0   , 4   , "Bullet's Speed"         },
	{0x14, 0x01, 0x60, 0x00, "Very Slow"              },
	{0x14, 0x01, 0x60, 0x20, "Slow"                   },
	{0x14, 0x01, 0x60, 0x40, "Fast"                   },
	{0x14, 0x01, 0x60, 0x60, "Very Fast"              },

	{0   , 0xfe, 0   , 2   , "Initital Vitality"      },
	{0x14, 0x01, 0x80, 0x00, "3 bars"                 },
	{0x14, 0x01, 0x80, 0x80, "4 bars"                 },

	// Dip C
//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x15, 0x01, 0x01, 0x00, "Off"                    },
//	{0x15, 0x01, 0x01, 0x01, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x15, 0x01, 0x02, 0x00, "Off"                    },
//	{0x15, 0x01, 0x02, 0x02, "On"                     },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x15, 0x01, 0x04, 0x00, "Off"                    },
	{0x15, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x15, 0x01, 0x08, 0x00, "Off"                    },
	{0x15, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x15, 0x01, 0x10, 0x00, "Off"                    },
	{0x15, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x15, 0x01, 0x20, 0x00, "Off"                    },
	{0x15, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x15, 0x01, 0x40, 0x00, "Off"                    },
	{0x15, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(NTFO);

static struct BurnDIPInfo captcommDIPList[]=
{
	// Defaults
        {0x23, 0xff, 0xff, 0x00, NULL                     },
        {0x24, 0xff, 0xff, 0x0B, NULL                     },
        {0x25, 0xff, 0xff, 0x60, NULL                     },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin 1"                 },
	{0x23, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x23, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x23, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x23, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x23, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x23, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x23, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x23, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "Coin 2"                 },
	{0x23, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x23, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x23, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x23, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x23, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x23, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x23, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x23, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "2C to Start, 1 to Cont" },
	{0x23, 0x01, 0x40, 0x00, "Off"                    },
	{0x23, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x23, 0x01, 0x80, 0x00, "Off"                    },
//	{0x23, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty 1"           },
	{0x24, 0x01, 0x07, 0x00, "Very Easy"              },
	{0x24, 0x01, 0x07, 0x01, "Easy 1"                 },
	{0x24, 0x01, 0x07, 0x02, "Easy 2"                 },
	{0x24, 0x01, 0x07, 0x03, "Normal"		  },
	{0x24, 0x01, 0x07, 0x04, "Difficult"              },
	{0x24, 0x01, 0x07, 0x05, "Very Difficult"         },
	{0x24, 0x01, 0x07, 0x06, "Hard"                   },
	{0x24, 0x01, 0x07, 0x07, "Hardest"                },

	{0   , 0xfe, 0   , 4   , "Difficulty 2"           },
	{0x24, 0x01, 0x18, 0x00, "1"                      },
	{0x24, 0x01, 0x18, 0x08, "2"                      },
	{0x24, 0x01, 0x18, 0x10, "3"                      },
	{0x24, 0x01, 0x18, 0x18, "4"                      },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x24, 0x01, 0x20, 0x00, "Off"                    },
//	{0x24, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 4   , "Max Players"            },
	{0x24, 0x01, 0xc0, 0x80, "1"                      },
	{0x24, 0x01, 0xc0, 0x00, "2"                      },
	{0x24, 0x01, 0xc0, 0x40, "3"                      },
	{0x24, 0x01, 0xc0, 0xc0, "4"                      },

	// Dip C
	{0   , 0xfe, 0   , 4   , "Lives"                  },
	{0x25, 0x01, 0x03, 0x03, "1"                      },
	{0x25, 0x01, 0x03, 0x00, "2"                      },
	{0x25, 0x01, 0x03, 0x01, "3"                      },
	{0x25, 0x01, 0x03, 0x02, "4"                      },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x25, 0x01, 0x04, 0x00, "Off"                    },
	{0x25, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x25, 0x01, 0x08, 0x00, "Off"                    },
	{0x25, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x25, 0x01, 0x10, 0x00, "Off"                    },
	{0x25, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x25, 0x01, 0x20, 0x00, "Off"                    },
	{0x25, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x25, 0x01, 0x40, 0x00, "Off"                    },
	{0x25, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(captcomm);

static struct BurnDIPInfo cawingDIPList[]=
{
	// Defaults
        {0x15, 0xff, 0xff, 0x40, NULL                     },
        {0x16, 0xff, 0xff, 0x03, NULL                     },
        {0x17, 0xff, 0xff, 0x60, NULL                     },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin 1"                 },
	{0x15, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "Coin 2"                 },
	{0x15, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x15, 0x01, 0x40, 0x00, "Off"                    },
	{0x15, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x15, 0x01, 0x80, 0x00, "Off"                    },
//	{0x15, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 5   , "Game Level"             },
	{0x16, 0x01, 0x07, 0x00, "Very Easy"              },
	{0x16, 0x01, 0x07, 0x02, "Easy"                   },
	{0x16, 0x01, 0x07, 0x03, "Normal"                 },
	{0x16, 0x01, 0x07, 0x04, "Difficult"              },
	{0x16, 0x01, 0x07, 0x07, "Very Difficult"         },

	{0   , 0xfe, 0   , 4   , "Nando Vital"            },
	{0x16, 0x01, 0x18, 0x08, "Easy"                   },
	{0x16, 0x01, 0x18, 0x00, "Normal"                 },
	{0x16, 0x01, 0x18, 0x10, "Difficult"              },
	{0x16, 0x01, 0x18, 0x18, "Very Difficult"         },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x20, 0x00, "Off"                    },
//	{0x16, 0x01, 0x20, 0x20, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x40, 0x00, "Off"                    },
//	{0x16, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x80, 0x00, "Off"                    },
//	{0x16, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x17, 0x01, 0x01, 0x00, "Off"                    },
//	{0x17, 0x01, 0x01, 0x01, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x17, 0x01, 0x02, 0x00, "Off"                    },
//	{0x17, 0x01, 0x02, 0x02, "On"                     },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x17, 0x01, 0x04, 0x00, "Off"                    },
	{0x17, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x17, 0x01, 0x08, 0x00, "Off"                    },
	{0x17, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x17, 0x01, 0x10, 0x00, "Off"                    },
	{0x17, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Sound"                  },
	{0x17, 0x01, 0x20, 0x00, "Off"                    },
	{0x17, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x17, 0x01, 0x40, 0x00, "Off"                    },
	{0x17, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(cawing);

static struct BurnDIPInfo cworld2jDIPList[]=
{
	// Defaults
        {0x0f, 0xff, 0xff, 0x40, NULL                     },
        {0x10, 0xff, 0xff, 0x08, NULL                     },
        {0x11, 0xff, 0xff, 0x60, NULL                     },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin"                   },
	{0x0f, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x0f, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x0f, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x0f, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x0f, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x0f, 0x01, 0x38, 0x00, "Off"                    },
//	{0x0f, 0x01, 0x38, 0x38, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x0f, 0x01, 0x40, 0x00, "Off"                    },
	{0x0f, 0x01, 0x40, 0x40, "On"                     },

	{0   , 0xfe, 0   , 2   , "Test"                   },
	{0x0f, 0x01, 0x80, 0x00, "Off"                    },
	{0x0f, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 4   , "Game Level"             },
	{0x10, 0x01, 0x07, 0x00, "0"                      },
	{0x10, 0x01, 0x07, 0x02, "1"                      },
	{0x10, 0x01, 0x07, 0x03, "2"                      },
	{0x10, 0x01, 0x07, 0x04, "3"                      },

	{0   , 0xfe, 0   , 3   , "Extend"                 },
	{0x10, 0x01, 0x18, 0x18, "D"                      },
	{0x10, 0x01, 0x18, 0x08, "E"                      },
	{0x10, 0x01, 0x18, 0x00, "N"                      },

	{0   , 0xfe, 0   , 5   , "Players Life"           },
	{0x10, 0x01, 0xe0, 0x80, "1"                      },
	{0x10, 0x01, 0xe0, 0x60, "2"                      },
	{0x10, 0x01, 0xe0, 0x00, "3"                      },
	{0x10, 0x01, 0xe0, 0x40, "4"                      },
	{0x10, 0x01, 0xe0, 0x20, "5"                      },

	// Dip C
//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x11, 0x01, 0x01, 0x00, "Off                     },
//	{0x11, 0x01, 0x01, 0x01, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x11, 0x01, 0x02, 0x00, "Off                     },
//	{0x11, 0x01, 0x02, 0x02, "On"                     },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x11, 0x01, 0x04, 0x00, "Off"                    },
	{0x11, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Stop"                   },
	{0x11, 0x01, 0x08, 0x00, "Off"                    },
	{0x11, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x11, 0x01, 0x10, 0x00, "Off"                    },
	{0x11, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x11, 0x01, 0x20, 0x00, "Off"                    },
	{0x11, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x11, 0x01, 0x40, 0x00, "Off"                    },
	{0x11, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(cworld2j);

static struct BurnDIPInfo dinoDIPList[]=
{
	// Defaults
	{0x1b, 0xff, 0xff, 0x00, NULL                     },

	// Dip C
	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x1b, 0x01, 0x08, 0x00, "Off"                    },
	{0x1b, 0x01, 0x08, 0x08, "On"                     },
};

STDDIPINFO(dino);

static struct BurnDIPInfo dwDIPList[]=
{
	// Defaults
        {0x15, 0xff, 0xff, 0x00, NULL                },
        {0x16, 0xff, 0xff, 0x00, NULL                },
        {0x17, 0xff, 0xff, 0x60, NULL                },

	// Dip A
        {0   , 0xfe, 0   , 7   , "Coin 1"            },
	{0x15, 0x01, 0x07, 0x06, "4 Coins 1 Credit"  },
	{0x15, 0x01, 0x07, 0x05, "3 Coins 1 Credit"  },
	{0x15, 0x01, 0x07, 0x04, "2 Coins 1 Credit"  },
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"   },
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"  },
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"  },
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"  },

        {0   , 0xfe, 0   , 7   , "Coin 2"            },
	{0x15, 0x01, 0x38, 0x30, "4 Coins 1 Credit"  },
	{0x15, 0x01, 0x38, 0x28, "3 Coins 1 Credit"  },
	{0x15, 0x01, 0x38, 0x20, "2 Coins 1 Credit"  },
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"   },
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"  },
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"  },
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"  },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x15, 0x01, 0x40, 0x00, "Off"               },
//	{0x15, 0x01, 0x40, 0x40, "On"                },

        {0   , 0xfe, 0   , 2   , "Free Play"         },
	{0x15, 0x01, 0x80, 0x00, "Off"               },
	{0x15, 0x01, 0x80, 0x80, "On"                },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"        },
	{0x16, 0x01, 0x07, 0x03, "Very Easy"         },
	{0x16, 0x01, 0x07, 0x02, "Easy 2"            },
	{0x16, 0x01, 0x07, 0x01, "Easy 1"            },
	{0x16, 0x01, 0x07, 0x00, "Normal"            },
	{0x16, 0x01, 0x07, 0x04, "Difficult 1"       },
	{0x16, 0x01, 0x07, 0x05, "Difficult 2"       },
	{0x16, 0x01, 0x07, 0x06, "Difficult 3"       },
	{0x16, 0x01, 0x07, 0x07, "Very Difficult"    },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x08, 0x00, "Off"               },
//	{0x16, 0x01, 0x08, 0x08, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x10, 0x00, "Off"               },
//	{0x16, 0x01, 0x10, 0x10, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x20, 0x00, "Off"               },
//	{0x16, 0x01, 0x20, 0x20, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x40, 0x00, "Off"               },
//	{0x16, 0x01, 0x40, 0x40, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x80, 0x00, "Off"               },
//	{0x16, 0x01, 0x80, 0x80, "On"                },

	// Dip C
	{0   , 0xfe, 0   , 2   , "Freeze"            },
	{0x17, 0x01, 0x01, 0x00, "Off"               },
	{0x17, 0x01, 0x01, 0x01, "On"                },

	{0   , 0xfe, 0   , 2   , "Turbo Mode"        },
	{0x17, 0x01, 0x02, 0x00, "Off"               },
	{0x17, 0x01, 0x02, 0x02, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x17, 0x01, 0x04, 0x00, "Off"               },
//	{0x17, 0x01, 0x04, 0x04, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x17, 0x01, 0x08, 0x00, "Off"               },
//	{0x17, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 2   , "Flip"              },
	{0x17, 0x01, 0x10, 0x00, "Off"               },
	{0x17, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 2   , "Demo Sound"        },
	{0x17, 0x01, 0x20, 0x00, "Off"               },
	{0x17, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue"          },
	{0x17, 0x01, 0x40, 0x00, "Off"               },
	{0x17, 0x01, 0x40, 0x40, "On"                },
};

STDDIPINFO(dw);

static struct BurnDIPInfo ffightDIPList[]=
{
	// Defaults
        {0x15, 0xff, 0xff, 0x40, NULL                     },
        {0x16, 0xff, 0xff, 0x0B, NULL                     },
        {0x17, 0xff, 0xff, 0x60, NULL                     },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin 1"                 },
	{0x15, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "Coin 2"                 },
	{0x15, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x15, 0x01, 0x40, 0x00, "Off"                    },
	{0x15, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x15, 0x01, 0x80, 0x00, "Off"                    },
//	{0x15, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"        },
	{0x16, 0x01, 0x07, 0x00, "01"            },
	{0x16, 0x01, 0x07, 0x01, "02"            },
	{0x16, 0x01, 0x07, 0x02, "03"            },
	{0x16, 0x01, 0x07, 0x03, "04"         },
	{0x16, 0x01, 0x07, 0x04, "05"       },
	{0x16, 0x01, 0x07, 0x05, "06"       },
	{0x16, 0x01, 0x07, 0x06, "07"       },
	{0x16, 0x01, 0x07, 0x07, "08"    },

	{0   , 0xfe, 0   , 4   , "Difficulty 2"      },
	{0x16, 0x01, 0x18, 0x00, "01"              },
	{0x16, 0x01, 0x18, 0x08, "02"            },
	{0x16, 0x01, 0x18, 0x10, "03"              },
	{0x16, 0x01, 0x18, 0x18, "04"           },

	{0   , 0xfe, 0   , 4   , "Bonus Life"        },
	{0x16, 0x01, 0x60, 0x00, "100k"              },
	{0x16, 0x01, 0x60, 0x20, "200k"              },
	{0x16, 0x01, 0x60, 0x40, "100k and every 200k"},
	{0x16, 0x01, 0x60, 0x60, "None"              },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x80, 0x00, "Off"               },
//	{0x16, 0x01, 0x80, 0x80, "On"                },

	// Dip C
	{0   , 0xfe, 0   , 4   , "Lives"             },
	{0x17, 0x01, 0x03, 0x03, "1"                 },
	{0x17, 0x01, 0x03, 0x00, "2"                 },
	{0x17, 0x01, 0x03, 0x01, "3"                 },
	{0x17, 0x01, 0x03, 0x02, "4"                 },

	{0   , 0xfe, 0   , 2   , "Free Play"         },
	{0x17, 0x01, 0x04, 0x00, "Off"               },
	{0x17, 0x01, 0x04, 0x04, "On"                },

	{0   , 0xfe, 0   , 2   , "Freeze"            },
	{0x17, 0x01, 0x08, 0x00, "Off"               },
	{0x17, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 2   , "Flip"              },
	{0x17, 0x01, 0x10, 0x00, "Off"               },
	{0x17, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 2   , "Demo Sound"        },
	{0x17, 0x01, 0x20, 0x00, "Off"               },
	{0x17, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue"          },
	{0x17, 0x01, 0x40, 0x00, "Off"               },
	{0x17, 0x01, 0x40, 0x40, "On"                },
};

STDDIPINFO(ffight);

static struct BurnDIPInfo forgottnDIPList[]=
{
	// Defaults
        {0x19, 0xff, 0xff, 0x40, NULL                     },
        {0x1a, 0xff, 0xff, 0x00, NULL                     },
        {0x1b, 0xff, 0xff, 0x00, NULL                     },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin 1"                 },
	{0x19, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x19, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x19, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x19, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x19, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x19, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x19, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x19, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "Coin 2"                 },
	{0x19, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x19, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x19, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x19, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x19, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x19, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x19, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x19, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x19, 0x01, 0x40, 0x00, "Off"                    },
	{0x19, 0x01, 0x40, 0x40, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x19, 0x01, 0x80, 0x00, "Off"                    },
	{0x19, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1a, 0x01, 0x01, 0x00, "Off"                    },
//	{0x1a, 0x01, 0x01, 0x01, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1a, 0x01, 0x02, 0x00, "Off"                    },
//	{0x1a, 0x01, 0x02, 0x02, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1a, 0x01, 0x04, 0x00, "Off"                    },
//	{0x1a, 0x01, 0x04, 0x04, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1a, 0x01, 0x08, 0x00, "Off"                    },
//	{0x1a, 0x01, 0x08, 0x08, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1a, 0x01, 0x10, 0x00, "Off"                    },
//	{0x1a, 0x01, 0x10, 0x10, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1a, 0x01, 0x20, 0x00, "Off"                    },
//	{0x1a, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Service Mode"           },
	{0x1a, 0x01, 0x40, 0x00, "Off"                    },
	{0x1a, 0x01, 0x40, 0x40, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x1a, 0x01, 0x80, 0x00, "Off"                    },
	{0x1a, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x01, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x01, 0x01, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x02, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x02, 0x02, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x04, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x04, 0x04, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x08, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x08, 0x08, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x10, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x10, 0x10, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x20, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x20, 0x20, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x40, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x80, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x80, 0x80, "On"                     },
};

STDDIPINFO(forgottn);

static struct BurnDIPInfo ghoulsDIPList[]=
{
	// Defaults
	{0x13, 0xff, 0xff, 0x00, NULL                     },
	{0x14, 0xff, 0xff, 0x00, NULL                     },
	{0x15, 0xff, 0xff, 0x00, NULL                     },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin 1"                 },
	{0x13, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x13, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x13, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x13, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x13, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "Coin 2"                 },
	{0x13, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x13, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x13, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x13, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x13, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x13, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x13, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x13, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 3   , "Cabinet"                },
	{0x13, 0x01, 0xc0, 0x00, "Upright 1 Player"       },
	{0x13, 0x01, 0xc0, 0x40, "Upright 2 Players"      },
	{0x13, 0x01, 0xc0, 0x80, "Cocktail"               },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"             },
	{0x14, 0x01, 0x07, 0x03, "Easier"              },
	{0x14, 0x01, 0x07, 0x02, "Very Easy"                 },
	{0x14, 0x01, 0x07, 0x01, "Easy"                 },
	{0x14, 0x01, 0x07, 0x00, "Normal"                 },
	{0x14, 0x01, 0x07, 0x04, "Difficult"            },
	{0x14, 0x01, 0x07, 0x05, "Hard"            },
	{0x14, 0x01, 0x07, 0x06, "Very Hard"            },
	{0x14, 0x01, 0x07, 0x07, "Hardest"         },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x14, 0x01, 0x08, 0x00, "Off"                    },
//	{0x14, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 4   , "Bonus Life"             },
	{0x14, 0x01, 0x30, 0x10, "10k, 30k and every 30k" },
	{0x14, 0x01, 0x30, 0x20, "20k, 50k and every 70k" },
	{0x14, 0x01, 0x30, 0x00, "30k, 60k and every 70k" },
	{0x14, 0x01, 0x30, 0x30, "40k, 70k and every 80k" },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x14, 0x01, 0x40, 0x00, "Off"                    },
//	{0x14, 0x01, 0x40, 0x40, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x14, 0x01, 0x80, 0x00, "Off"                    },
	{0x14, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
	{0   , 0xfe, 0   , 4   , "Lives"                  },
	{0x15, 0x01, 0x03, 0x00, "3"                      },
	{0x15, 0x01, 0x03, 0x01, "4"                      },
	{0x15, 0x01, 0x03, 0x02, "5"                      },
	{0x15, 0x01, 0x03, 0x03, "6"                      },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x15, 0x01, 0x04, 0x00, "Off"                    },
//	{0x15, 0x01, 0x04, 0x04, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x15, 0x01, 0x08, 0x00, "Off"                    },
//	{0x15, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x15, 0x01, 0x10, 0x00, "Off"                    },
	{0x15, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x15, 0x01, 0x20, 0x20, "Off"                    },
	{0x15, 0x01, 0x20, 0x00, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x15, 0x01, 0x40, 0x40, "Off"                    },
	{0x15, 0x01, 0x40, 0x00, "On"                     },
};

STDDIPINFO(ghouls);

static struct BurnDIPInfo knightsDIPList[]=
{
	// Defaults
        {0x1b, 0xff, 0xff, 0x00, NULL                },
        {0x1c, 0xff, 0xff, 0x03, NULL                },
        {0x1d, 0xff, 0xff, 0x60, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin"                   },
	{0x1b, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x1b, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x1b, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x1b, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x1b, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x08, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x08, 0x08, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x10, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x10, 0x10, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x20, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "2C to Start, 1 to Cont" },
	{0x1b, 0x01, 0x40, 0x00, "Off"                    },
	{0x1b, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x80, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Level 1"                },
	{0x1c, 0x01, 0x07, 0x02, "Easy 1"                 },
	{0x1c, 0x01, 0x07, 0x01, "Easy 2"                 },
	{0x1c, 0x01, 0x07, 0x00, "Easy 3"                 },
	{0x1c, 0x01, 0x07, 0x03, "Normal"                 },
	{0x1c, 0x01, 0x07, 0x04, "Difficult 1"            },
	{0x1c, 0x01, 0x07, 0x05, "Difficult 2"            },
	{0x1c, 0x01, 0x07, 0x06, "Difficult 3"            },
	{0x1c, 0x01, 0x07, 0x07, "Difficult 4"            },

	{0   , 0xfe, 0   , 8   , "Level 2"                },
	{0x1c, 0x01, 0x38, 0x28, "Easy 1"                 },
	{0x1c, 0x01, 0x38, 0x30, "Easy 2"                 },
	{0x1c, 0x01, 0x38, 0x38, "Easy 3"                 },
	{0x1c, 0x01, 0x38, 0x00, "Normal"                 },
	{0x1c, 0x01, 0x38, 0x08, "Difficult 1"            },
	{0x1c, 0x01, 0x38, 0x10, "Difficult 2"            },
	{0x1c, 0x01, 0x38, 0x18, "Difficult 3"            },
	{0x1c, 0x01, 0x38, 0x20, "Difficult 4"            },

	{0   , 0xfe, 0   , 2   , "Coin Shooter"           },
	{0x1c, 0x01, 0x40, 0x40, "1 Shooter"              },
	{0x1c, 0x01, 0x40, 0x00, "3 Shooter"              },

	{0   , 0xfe, 0   , 2   , "Play Type"              },
	{0x1c, 0x01, 0x80, 0x80, "2P"                     },
	{0x1c, 0x01, 0x80, 0x00, "3P"                     },

	// Dip C
	{0   , 0xfe, 0   , 4   , "Lives (varies by Diff)" },
	{0x1d, 0x01, 0x03, 0x03, "Least"                  },
	{0x1d, 0x01, 0x03, 0x00, "Less"                   },
	{0x1d, 0x01, 0x03, 0x01, "More"                   },
	{0x1d, 0x01, 0x03, 0x02, "Most"                   },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x1d, 0x01, 0x04, 0x00, "Off"                    },
	{0x1d, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x1d, 0x01, 0x08, 0x00, "Off"                    },
	{0x1d, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x1d, 0x01, 0x10, 0x00, "Off"                    },
	{0x1d, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x1d, 0x01, 0x20, 0x00, "Off"                    },
	{0x1d, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x1d, 0x01, 0x40, 0x00, "Off"                    },
	{0x1d, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(knights);

static struct BurnDIPInfo kodDIPList[]=
{
	// Defaults
        {0x1b, 0xff, 0xff, 0x00, NULL                },
        {0x1c, 0xff, 0xff, 0x83, NULL                },
        {0x1d, 0xff, 0xff, 0x60, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coinage"                },
	{0x1b, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x1b, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x1b, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x1b, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x1b, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Shooter"                },
	{0x1b, 0x01, 0x08, 0x08, "1"                      },
	{0x1b, 0x01, 0x08, 0x00, "3"                      },

	{0   , 0xfe, 0   , 2   , "Max Players"            },
	{0x1b, 0x01, 0x10, 0x10, "2"                      },
	{0x1b, 0x01, 0x10, 0x00, "3"                      },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x20, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "2C to Start, 1 to Cont" },
	{0x1b, 0x01, 0x40, 0x00, "Off"                    },
	{0x1b, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x80, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"             },
	{0x1c, 0x01, 0x07, 0x00, "Easiest"                },
	{0x1c, 0x01, 0x07, 0x01, "Very Easy"              },
	{0x1c, 0x01, 0x07, 0x02, "Easy"                   },
	{0x1c, 0x01, 0x07, 0x03, "Normal"                 },
	{0x1c, 0x01, 0x07, 0x04, "Medium"                 },
	{0x1c, 0x01, 0x07, 0x05, "Hard"                   },
	{0x1c, 0x01, 0x07, 0x06, "Very Hard"              },
	{0x1c, 0x01, 0x07, 0x07, "Hardest"                },

	{0   , 0xfe, 0   , 8   , "Lives"                  },
	{0x1c, 0x01, 0x38, 0x08, "1"                      },
	{0x1c, 0x01, 0x38, 0x00, "2"                      },
	{0x1c, 0x01, 0x38, 0x10, "3"                      },
	{0x1c, 0x01, 0x38, 0x18, "4"                      },
	{0x1c, 0x01, 0x38, 0x20, "5"                      },
	{0x1c, 0x01, 0x38, 0x28, "6"                      },
	{0x1c, 0x01, 0x38, 0x30, "7"                      },
	{0x1c, 0x01, 0x38, 0x38, "8"                      },

	{0   , 0xfe, 0   , 4   , "Bonus Life"             },
	{0x1c, 0x01, 0xc0, 0x40, "80k and every 400k"     },
	{0x1c, 0x01, 0xc0, 0x80, "160k and every 450k"    },
	{0x1c, 0x01, 0xc0, 0x00, "200k and every 450k"    },
	{0x1c, 0x01, 0xc0, 0xc0, "None"                   },

	// Dip C
//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1d, 0x01, 0x01, 0x00, "Off"                    },
//	{0x1d, 0x01, 0x01, 0x01, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1d, 0x01, 0x02, 0x00, "Off"                    },
//	{0x1d, 0x01, 0x02, 0x02, "On"                     },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x1d, 0x01, 0x04, 0x00, "Off"                    },
	{0x1d, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x1d, 0x01, 0x08, 0x00, "Off"                    },
	{0x1d, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x1d, 0x01, 0x10, 0x00, "Off"                    },
	{0x1d, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x1d, 0x01, 0x20, 0x00, "Off"                    },
	{0x1d, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x1d, 0x01, 0x40, 0x00, "Off"                    },
	{0x1d, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(kod);

static struct BurnDIPInfo megamanDIPList[]=
{
	// Defaults
        {0x15, 0xff, 0xff, 0x00, NULL                },
        {0x16, 0xff, 0xff, 0x45, NULL                },
        {0x17, 0xff, 0xff, 0x00, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 19  , "Coin Mode"         },
	{0x15, 0x01, 0x1f, 0x00, "1 Coin  1 Credit"  },
	{0x15, 0x01, 0x1f, 0x01, "1 Coin  2 Credits" },
	{0x15, 0x01, 0x1f, 0x02, "1 Coin  3 Credits" },
	{0x15, 0x01, 0x1f, 0x03, "1 Coin  4 Credits" },
	{0x15, 0x01, 0x1f, 0x04, "1 Coin  5 Credits" },
	{0x15, 0x01, 0x1f, 0x05, "1 Coin  6 Credits" },
	{0x15, 0x01, 0x1f, 0x06, "1 Coin  7 Credits" },
	{0x15, 0x01, 0x1f, 0x07, "1 Coin  8 Credits" },
	{0x15, 0x01, 0x1f, 0x08, "1 Coin  9 Credits" },
	{0x15, 0x01, 0x1f, 0x09, "2 Coins 1 Credit"  },
	{0x15, 0x01, 0x1f, 0x0a, "3 Coins 1 Credit"  },
	{0x15, 0x01, 0x1f, 0x0b, "4 Coins 1 Credit"  },
	{0x15, 0x01, 0x1f, 0x0c, "5 Coins 1 Credit"  },
	{0x15, 0x01, 0x1f, 0x0d, "6 Coins 1 Credit"  },
	{0x15, 0x01, 0x1f, 0x0e, "7 Coins 1 Credit"  },
	{0x15, 0x01, 0x1f, 0x0f, "8 Coins 1 Credit"  },
	{0x15, 0x01, 0x1f, 0x10, "9 Coins 1 Credit"  },
	{0x15, 0x01, 0x1f, 0x11, "2 Coins Start 1 Coin Continue"},
	{0x15, 0x01, 0x1f, 0x12, "Free Play"         },

	{0   , 0xfe, 0   , 4   , "Chute Type"        },
	{0x15, 0x01, 0x60, 0x40, "1 Chute Single"    },
	{0x15, 0x01, 0x60, 0x00, "2 Chutes Single"   },
	{0x15, 0x01, 0x60, 0x20, "2 Chutes Multi"    },

	// Dip C
	{0   , 0xfe, 0   , 2   , "Continue"          },
	{0x17, 0x01, 0x04, 0x04, "Off"               },
	{0x17, 0x01, 0x04, 0x00, "On"                },

	{0   , 0xfe, 0   , 2   , "Monitor"           },
	{0x17, 0x01, 0x01, 0x00, "Normal"            },
	{0x17, 0x01, 0x01, 0x01, "Flip"              },

	{0   , 0xfe, 0   , 2   , "Demo Sounds"       },
	{0x17, 0x01, 0x20, 0x20, "Off"               },
	{0x17, 0x01, 0x20, 0x00, "On"                },

	// Dip B
	{0   , 0xfe, 0   , 4   , "Difficulty"        },
	{0x16, 0x01, 0x03, 0x00, "1 Easy"            },
	{0x16, 0x01, 0x03, 0x01, "2 Normal"          },
	{0x16, 0x01, 0x03, 0x02, "3 Hard"            },
	{0x16, 0x01, 0x03, 0x03, "4 Very Hard"       },

	{0   , 0xfe, 0   , 4   , "Start Time"        },
	{0x16, 0x01, 0x0c, 0x00, "1 Easy"            },
	{0x16, 0x01, 0x0c, 0x04, "2 Normal"          },
	{0x16, 0x01, 0x0c, 0x08, "3 Hard"            },
	{0x16, 0x01, 0x0c, 0x0c, "4 Very Hard"       },

	{0   , 0xfe, 0   , 2   , "Voice"             },
	{0x16, 0x01, 0x40, 0x00, "Off"               },
	{0x16, 0x01, 0x40, 0x40, "Normal"            },
};

STDDIPINFO(megaman);

static struct BurnDIPInfo mercsDIPList[]=
{
	// Defaults
        {0x1b, 0xff, 0xff, 0x00, NULL                },
        {0x1c, 0xff, 0xff, 0x03, NULL                },
        {0x1d, 0xff, 0xff, 0x60, NULL                },


	// Dip A
	{0   , 0xfe, 0   , 8   , "Coinage"                },
	{0x1b, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x1b, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x1b, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x1b, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x1b, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

//	{0   , 0xfe, 0   , 8   , "Unknown"                },
//	{0x1b, 0x01, 0x38, 0x38, "0"                      },
//	{0x1b, 0x01, 0x38, 0x30, "1"                      },
//	{0x1b, 0x01, 0x38, 0x28, "2"                      },
//	{0x1b, 0x01, 0x38, 0x00, "3"                      },
//	{0x1b, 0x01, 0x38, 0x08, "4"                      },
//	{0x1b, 0x01, 0x38, 0x10, "5"                      },
//	{0x1b, 0x01, 0x38, 0x18, "6"                      },
//	{0x1b, 0x01, 0x38, 0x20, "7"                      },

	{0   , 0xfe, 0   , 2   , "2C to Start, 1 to Cont" },
	{0x1b, 0x01, 0x40, 0x00, "Off"                    },
	{0x1b, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x1b, 0x01, 0x80, 0x00, "Off"               },
//	{0x1b, 0x01, 0x80, 0x80, "On"                },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Game Level"            },
	{0x1c, 0x01, 0x07, 0x00, "1 (Easiest)"           },
	{0x1c, 0x01, 0x07, 0x01, "2"                     },
	{0x1c, 0x01, 0x07, 0x02, "3"                     },
	{0x1c, 0x01, 0x07, 0x03, "4"                     },
	{0x1c, 0x01, 0x07, 0x04, "5"                     },
	{0x1c, 0x01, 0x07, 0x05, "6"                     },
	{0x1c, 0x01, 0x07, 0x06, "7"                     },
	{0x1c, 0x01, 0x07, 0x07, "8 (Hardest)"           },

	{0   , 0xfe, 0   , 2   , "Coin Shooter"          },
	{0x1c, 0x01, 0x08, 0x08, "1 Shooter"             },
	{0x1c, 0x01, 0x08, 0x00, "3 Shooter"             },

	{0   , 0xfe, 0   , 2   , "Game Player"           },
	{0x1c, 0x01, 0x10, 0x10, "2 Player"              },
	{0x1c, 0x01, 0x10, 0x00, "3 Player"              },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x1c, 0x01, 0x20, 0x00, "Off"               },
//	{0x1c, 0x01, 0x20, 0x20, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x1c, 0x01, 0x40, 0x00, "Off"               },
//	{0x1c, 0x01, 0x40, 0x40, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x1c, 0x01, 0x80, 0x00, "Off"               },
//	{0x1c, 0x01, 0x80, 0x80, "On"                },

	// Dip C
//	{0   , 0xfe, 0   , 8   , "Unknown"                },
//	{0x1d, 0x01, 0x07, 0x07, "0"                      },
//	{0x1d, 0x01, 0x07, 0x06, "1"                      },
//	{0x1d, 0x01, 0x07, 0x05, "2"                      },
//	{0x1d, 0x01, 0x07, 0x00, "3"                      },
//	{0x1d, 0x01, 0x07, 0x01, "4"                      },
//	{0x1d, 0x01, 0x07, 0x02, "5"                      },
//	{0x1d, 0x01, 0x07, 0x03, "6"                      },
//	{0x1d, 0x01, 0x07, 0x04. "7"                      },

	{0   , 0xfe, 0   , 2   , "Freeze"            },
	{0x1d, 0x01, 0x08, 0x00, "Off"               },
	{0x1d, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 2   , "Display FlipFlop"  },
	{0x1d, 0x01, 0x10, 0x00, "Off"               },
	{0x1d, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 2   , "Demonstration Sound"},
	{0x1d, 0x01, 0x20, 0x00, "Off"               },
	{0x1d, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue Play"     },
	{0x1d, 0x01, 0x40, 0x00, "Off"               },
	{0x1d, 0x01, 0x40, 0x40, "On"                },
};

STDDIPINFO(mercs);

static struct BurnDIPInfo mswordDIPList[]=
{
	// Defaults
        {0x15, 0xff, 0xff, 0x00, NULL                },
        {0x16, 0xff, 0xff, 0x43, NULL                },
        {0x17, 0xff, 0xff, 0x60, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "A Side"                 },
	{0x15, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "B Side"                 },
	{0x15, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "2C to Start, 1 to Cont" },
	{0x15, 0x01, 0x40, 0x00, "Off"                    },
	{0x15, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x15, 0x01, 0x80, 0x00, "Off"                    },
//	{0x15, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Level 1"                },
	{0x16, 0x01, 0x07, 0x00, "Easy 3"                 },
	{0x16, 0x01, 0x07, 0x01, "Easy 2"                 },
	{0x16, 0x01, 0x07, 0x02, "Easy 1"                 },
	{0x16, 0x01, 0x07, 0x03, "Normal"                 },
	{0x16, 0x01, 0x07, 0x04, "Difficult 1"            },
	{0x16, 0x01, 0x07, 0x05, "Difficult 2"            },
	{0x16, 0x01, 0x07, 0x06, "Difficult 3"            },
	{0x16, 0x01, 0x07, 0x07, "Difficult 4"            },

	{0   , 0xfe, 0   , 8   , "Level 2"                },
	{0x16, 0x01, 0x38, 0x18, "Easy 3"                 },
	{0x16, 0x01, 0x38, 0x10, "Easy 2"                 },
	{0x16, 0x01, 0x38, 0x08, "Easy 1"                 },
	{0x16, 0x01, 0x38, 0x00, "Normal"                 },
	{0x16, 0x01, 0x38, 0x20, "Difficult 1"            },
	{0x16, 0x01, 0x38, 0x28, "Difficult 2"            },
	{0x16, 0x01, 0x38, 0x30, "Difficult 3"            },
	{0x16, 0x01, 0x38, 0x38, "Difficult 4"            },

	{0   , 0xfe, 0   , 2   , "Stage Select"           },
	{0x16, 0x01, 0x40, 0x00, "Off"                    },
	{0x16, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x80, 0x00, "Off"                    },
//	{0x16, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
	{0   , 0xfe, 0   , 4   , "Vitality"               },
	{0x17, 0x01, 0x03, 0x03, "1"                      },
	{0x17, 0x01, 0x03, 0x00, "2"                      },
	{0x17, 0x01, 0x03, 0x01, "3"                      },
	{0x17, 0x01, 0x03, 0x02, "4"                      },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x17, 0x01, 0x04, 0x00, "Off"                    },
	{0x17, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x17, 0x01, 0x08, 0x00, "Off"                    },
	{0x17, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x17, 0x01, 0x10, 0x00, "Off"                    },
	{0x17, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x17, 0x01, 0x20, 0x00, "Off"                    },
	{0x17, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x17, 0x01, 0x40, 0x00, "Off"                    },
	{0x17, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(msword);


static struct BurnDIPInfo mtwinsDIPList[]=
{
	// Defaults
        {0x15, 0xff, 0xff, 0x40, NULL                },
        {0x16, 0xff, 0xff, 0x23, NULL                },
        {0x17, 0xff, 0xff, 0x60, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin 1"                 },
	{0x15, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "Coin 2"                 },
	{0x15, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x15, 0x01, 0x40, 0x00, "Off"                    },
	{0x15, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x15, 0x01, 0x80, 0x00, "Off"               },
//	{0x15, 0x01, 0x80, 0x80, "On"                },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"        },
	{0x16, 0x01, 0x07, 0x00, "0 (Easiest)"       },
	{0x16, 0x01, 0x07, 0x01, "1"                 },
	{0x16, 0x01, 0x07, 0x02, "2"                 },
	{0x16, 0x01, 0x07, 0x03, "3"                 },
	{0x16, 0x01, 0x07, 0x04, "4"                 },
	{0x16, 0x01, 0x07, 0x05, "5"                 },
	{0x16, 0x01, 0x07, 0x06, "6"                 },
	{0x16, 0x01, 0x07, 0x07, "7"                 },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x08, 0x00, "Off"               },
//	{0x16, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 3   , "Lives"             },
	{0x16, 0x01, 0x30, 0x00, "1"                 },
	{0x16, 0x01, 0x30, 0x20, "2"                 },
	{0x16, 0x01, 0x30, 0x30, "3"                 },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x40, 0x00, "Off"               },
//	{0x16, 0x01, 0x40, 0x40, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x80, 0x00, "Off"               },
//	{0x16, 0x01, 0x80, 0x80, "On"                },

	// Dip C
//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x17, 0x01, 0x01, 0x00, "Off"               },
//	{0x17, 0x01, 0x01, 0x01, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x17, 0x01, 0x02, 0x00, "Off"               },
//	{0x17, 0x01, 0x02, 0x02, "On"                },

	{0   , 0xfe, 0   , 2   , "Free Play"         },
	{0x17, 0x01, 0x04, 0x00, "Off"               },
	{0x17, 0x01, 0x04, 0x04, "On"                },

	{0   , 0xfe, 0   , 2   , "Freeze"            },
	{0x17, 0x01, 0x08, 0x00, "Off"               },
	{0x17, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 2   , "Flip"              },
	{0x17, 0x01, 0x10, 0x00, "Off"               },
	{0x17, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 2   , "Demo Sound"        },
	{0x17, 0x01, 0x20, 0x00, "Off"               },
	{0x17, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue"          },
	{0x17, 0x01, 0x40, 0x00, "Off"               },
	{0x17, 0x01, 0x40, 0x40, "On"                },
};

STDDIPINFO(mtwins);

static struct BurnDIPInfo nemoDIPList[]=
{
	// Defaults
        {0x15, 0xff, 0xff, 0x40, NULL                },
        {0x16, 0xff, 0xff, 0x03, NULL                },
        {0x17, 0xff, 0xff, 0x60, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "A Side"                 },
	{0x15, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "B Side"                 },
	{0x15, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x15, 0x01, 0x40, 0x00, "Off"                    },
	{0x15, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x15, 0x01, 0x80, 0x00, "Off"                    },
//	{0x15, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Game Level"             },
	{0x16, 0x01, 0x07, 0x00, "Very Easy (3)"          },
	{0x16, 0x01, 0x07, 0x01, "Easy (4)"               },
	{0x16, 0x01, 0x07, 0x02, "Easy (5)"               },
	{0x16, 0x01, 0x07, 0x03, "Normal (6)"             },
	{0x16, 0x01, 0x07, 0x04, "Difficult (7)"          },
	{0x16, 0x01, 0x07, 0x05, "Difficult (8)"          },
	{0x16, 0x01, 0x07, 0x06, "Difficult (9)"          },
	{0x16, 0x01, 0x07, 0x07, "Very Difficult (a)"     },

	{0   , 0xfe, 0   , 3   , "Life"                   },
	{0x16, 0x01, 0x18, 0x18, "Minimum"                },
	{0x16, 0x01, 0x18, 0x00, "Medium"                 },
	{0x16, 0x01, 0x18, 0x10, "Maximum"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x20, 0x00, "Off"                    },
//	{0x16, 0x01, 0x20, 0x20, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x40, 0x00, "Off"                    },
//	{0x16, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x80, 0x00, "Off"                    },
//	{0x16, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
//	{0   , 0xfe, 0   , 4   , "Unknown"                },
//	{0x17, 0x01, 0x03, 0x00, "1"                      },
//	{0x17, 0x01, 0x03, 0x01, "2"                      },
//	{0x17, 0x01, 0x03, 0x02, "3"                      },
//	{0x17, 0x01, 0x03, 0x03, "4"                      },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x17, 0x01, 0x04, 0x00, "Off"                    },
	{0x17, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x17, 0x01, 0x08, 0x00, "Off"                    },
	{0x17, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x17, 0x01, 0x10, 0x00, "Off"                    },
	{0x17, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Sound"                  },
	{0x17, 0x01, 0x20, 0x00, "Off"                    },
	{0x17, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x17, 0x01, 0x40, 0x00, "Off"                    },
	{0x17, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(nemo);

static struct BurnDIPInfo pang3DIPList[]=
{
	// Defaults
	{0x13, 0xff, 0xff, 0x00, NULL                },

	// Dip C
	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x13, 0x01, 0x08, 0x00, "Off"                    },
	{0x13, 0x01, 0x08, 0x08, "On"                     },
};

STDDIPINFO(pang3);

static struct BurnDIPInfo pnickjDIPList[]=
{
	// Defaults
        {0x13, 0xff, 0xff, 0x00, NULL                },
        {0x14, 0xff, 0xff, 0x03, NULL                },
        {0x15, 0xff, 0xff, 0x60, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin"                   },
	{0x13, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x13, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x13, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x13, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x13, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x13, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Chuter Mode"            },
	{0x13, 0x01, 0x08, 0x00, "1 Chuter"               },
	{0x13, 0x01, 0x08, 0x08, "2 Chuter"               },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x13, 0x01, 0x10, 0x00, "Off"               },
//	{0x13, 0x01, 0x10, 0x10, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x13, 0x01, 0x20, 0x00, "Off"               },
//	{0x13, 0x01, 0x20, 0x20, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x13, 0x01, 0x40, 0x00, "Off"               },
//	{0x13, 0x01, 0x40, 0x40, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x13, 0x01, 0x80, 0x00, "Off"               },
//	{0x13, 0x01, 0x80, 0x80, "On"                },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"        },
	{0x14, 0x01, 0x07, 0x00, "Easiest"           },
	{0x14, 0x01, 0x07, 0x01, "Very Easy"         },
	{0x14, 0x01, 0x07, 0x02, "Easy"              },
	{0x14, 0x01, 0x07, 0x03, "Normal"            },
	{0x14, 0x01, 0x07, 0x04, "Hard"              },
	{0x14, 0x01, 0x07, 0x05, "Very Hard"         },
	{0x14, 0x01, 0x07, 0x06, "Hardest"           },
	{0x14, 0x01, 0x07, 0x07, "Master Level"      },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x14, 0x01, 0x08, 0x00, "Off"               },
//	{0x14, 0x01, 0x08, 0x08, "On"                },

//	{0   , 0xfe, 0   , 4   , "Unknown"           },
//	{0x14, 0x01, 0x30, 0x00, "1"                 },
//	{0x14, 0x01, 0x30, 0x10, "2"                 },
//	{0x14, 0x01, 0x30, 0x20, "3"                 },
//	{0x14, 0x01, 0x30, 0x30, "4"                 },

	{0   , 0xfe, 0   , 4   , "Vs. Play Mode"     },
	{0x14, 0x01, 0xc0, 0x00, "1 Game  Match"     },
	{0x14, 0x01, 0xc0, 0x40, "3 Games Match"     },
	{0x14, 0x01, 0xc0, 0x80, "5 Games Match"     },
	{0x14, 0x01, 0xc0, 0xc0, "7 Games Match"     },

	// Dip C
//	{0   , 0xfe, 0   , 4   , "Unknown"                },
//	{0x15, 0x01, 0x03, 0x00, "1"                      },
//	{0x15, 0x01, 0x03, 0x01, "2"                      },
//	{0x15, 0x01, 0x03, 0x02, "3"                      },
//	{0x15, 0x01, 0x03, 0x03, "4"                      },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x15, 0x01, 0x04, 0x00, "Off"               },
//	{0x15, 0x01, 0x04, 0x04, "On"                },

	{0   , 0xfe, 0   , 2   , "Stop Mode"         },
	{0x15, 0x01, 0x08, 0x00, "Off"               },
	{0x15, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 2   , "Flip"              },
	{0x15, 0x01, 0x10, 0x00, "Off"               },
	{0x15, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 2   , "Demo Sound"        },
	{0x15, 0x01, 0x20, 0x00, "Off"               },
	{0x15, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue"          },
	{0x15, 0x01, 0x40, 0x00, "Off"               },
	{0x15, 0x01, 0x40, 0x40, "On"                },
};

STDDIPINFO(pnickj);

static struct BurnDIPInfo punishDIPList[]=
{
	// Defaults
	{0x13, 0xff, 0xff, 0x00, NULL                },

	// Dip C
	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x13, 0x01, 0x08, 0x00, "Off"                    },
	{0x13, 0x01, 0x08, 0x08, "On"                     },
};

STDDIPINFO(punish);

static struct BurnDIPInfo qadDIPList[]=
{
	// Defaults
        {0x0f, 0xff, 0xff, 0x40, NULL                },
        {0x10, 0xff, 0xff, 0x0B, NULL                },
        {0x11, 0xff, 0xff, 0x00, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin"                   },
	{0x0f, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x0f, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x0f, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x0f, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x0f, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x08, 0x00, "Off"               },
//	{0x0f, 0x01, 0x08, 0x08, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x10, 0x00, "Off"               },
//	{0x0f, 0x01, 0x10, 0x10, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x20, 0x00, "Off"               },
//	{0x0f, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x0f, 0x01, 0x40, 0x00, "Off"                    },
	{0x0f, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x80, 0x00, "Off"               },
//	{0x0f, 0x01, 0x80, 0x80, "On"                },

	// Dip B
	{0   , 0xfe, 0   , 5   , "Difficulty"        },
	{0x10, 0x01, 0x07, 0x00, "Very Easy"            },
//	{0x10, 0x01, 0x07, 0x01, "Very Easy"            },
	{0x10, 0x01, 0x07, 0x02, "Easy"            },
	{0x10, 0x01, 0x07, 0x03, "Normal"         },
	{0x10, 0x01, 0x07, 0x04, "Hard"       },
	{0x10, 0x01, 0x07, 0x05, "Very Hard"       },
//	{0x10, 0x01, 0x07, 0x06, "Very Hard"       },
//	{0x10, 0x01, 0x07, 0x07, "Very Hard"    },

	{0   , 0xfe, 0   , 4   , "Wisdom of Enemy"            },
	{0x10, 0x01, 0x18, 0x00, "Low"               },
	{0x10, 0x01, 0x18, 0x08, "Normal"            },
	{0x10, 0x01, 0x18, 0x10, "High"              },
	{0x10, 0x01, 0x18, 0x18, "Brilliant"         },

	{0   , 0xfe, 0   , 3   , "Initial Vitality"  },
	{0x10, 0x01, 0xe0, 0x80, "1"                 },
//	{0x10, 0x01, 0xe0, 0xa0, "1"                 },
//	{0x10, 0x01, 0xe0, 0xc0, "1"                 },
//	{0x10, 0x01, 0xe0, 0xe0, "1"                 },
	{0x10, 0x01, 0xe0, 0x60, "2"                 },
	{0x10, 0x01, 0xe0, 0x00, "5"                 },

	// Dip C
	{0   , 0xfe, 0   , 2   , "Free Play"         },
	{0x11, 0x01, 0x04, 0x00, "Off"               },
	{0x11, 0x01, 0x04, 0x04, "On"                },

	{0   , 0xfe, 0   , 2   , "Stop"              },
	{0x11, 0x01, 0x08, 0x00, "Off"               },
	{0x11, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 2   , "Flip"              },
	{0x11, 0x01, 0x10, 0x00, "Off"               },
	{0x11, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 2   , "Demo Sound"        },
	{0x11, 0x01, 0x20, 0x20, "Off"               },
	{0x11, 0x01, 0x20, 0x00, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue"          },
	{0x11, 0x01, 0x40, 0x40, "Off"               },
	{0x11, 0x01, 0x40, 0x00, "On"                },
};

STDDIPINFO(qad);

static struct BurnDIPInfo qadjDIPList[]=
{
	// Defaults
	{0x0f, 0xff, 0xff, 0x00, NULL                },
	{0x10, 0xff, 0xff, 0x00, NULL                },
	{0x11, 0xff, 0xff, 0x00, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin"                   },
	{0x0f, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x0f, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x0f, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x0f, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x0f, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x08, 0x00, "Off"               },
//	{0x0f, 0x01, 0x08, 0x08, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x10, 0x00, "Off"               },
//	{0x0f, 0x01, 0x10, 0x10, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x20, 0x00, "Off"               },
//	{0x0f, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x0f, 0x01, 0x40, 0x00, "Off"                    },
	{0x0f, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x80, 0x00, "Off"               },
//	{0x0f, 0x01, 0x80, 0x80, "On"                },

	// Dip B
	{0   , 0xfe, 0   , 5   , "Difficulty"        },
	{0x10, 0x01, 0x07, 0x00, "Very Easy"            },
//	{0x10, 0x01, 0x07, 0x01, "Very Easy"            },
	{0x10, 0x01, 0x07, 0x02, "Easy"            },
	{0x10, 0x01, 0x07, 0x03, "Normal"         },
	{0x10, 0x01, 0x07, 0x04, "Hard"       },
	{0x10, 0x01, 0x07, 0x05, "Very Hard"       },
//	{0x10, 0x01, 0x07, 0x06, "Very Hard"       },
//	{0x10, 0x01, 0x07, 0x07, "Very Hard"    },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x10, 0x01, 0x08, 0x00, "Off"               },
//	{0x10, 0x01, 0x08, 0x08, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x10, 0x01, 0x10, 0x00, "Off"               },
//	{0x10, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 3   , "Initial Vitality"  },
//	{0x10, 0x01, 0xe0, 0xc0, "1"                 },
//	{0x10, 0x01, 0xe0, 0xe0, "1"                 },
	{0x10, 0x01, 0xe0, 0x60, "1"                 },
	{0x10, 0x01, 0xe0, 0xa0, "2"                 },
//	{0x10, 0x01, 0xe0, 0x80, "3"                 },
	{0x10, 0x01, 0xe0, 0x00, "3"                 },

	// Dip C
//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x11, 0x01, 0x01, 0x00, "Off"               },
//	{0x11, 0x01, 0x01, 0x01, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x11, 0x01, 0x02, 0x00, "Off"               },
//	{0x11, 0x01, 0x02, 0x02, "On"                },

	{0   , 0xfe, 0   , 2   , "Free Play"         },
	{0x11, 0x01, 0x04, 0x00, "Off"               },
	{0x11, 0x01, 0x04, 0x04, "On"                },

	{0   , 0xfe, 0   , 2   , "Stop"              },
	{0x11, 0x01, 0x08, 0x00, "Off"               },
	{0x11, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 2   , "Flip"              },
	{0x11, 0x01, 0x10, 0x00, "Off"               },
	{0x11, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 2   , "Demo Sound"        },
	{0x11, 0x01, 0x20, 0x00, "Off"               },
	{0x11, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue"          },
	{0x11, 0x01, 0x40, 0x40, "Off"               },
	{0x11, 0x01, 0x40, 0x00, "On"                },
};

STDDIPINFO(qadj);

static struct BurnDIPInfo qtono2DIPList[]=
{
	// Defaults
        {0x0f, 0xff, 0xff, 0x40, NULL                },
        {0x10, 0xff, 0xff, 0xA0, NULL                },
        {0x11, 0xff, 0xff, 0x60, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin"                   },
	{0x0f, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x0f, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x0f, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x0f, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x0f, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x0f, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x08, 0x00, "Off"               },
//	{0x0f, 0x01, 0x08, 0x08, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x10, 0x00, "Off"               },
//	{0x0f, 0x01, 0x10, 0x10, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x20, 0x00, "Off"               },
//	{0x0f, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x0f, 0x01, 0x40, 0x00, "Off"                    },
	{0x0f, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x0f, 0x01, 0x80, 0x00, "Off"               },
//	{0x0f, 0x01, 0x80, 0x80, "On"                },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"        },
	{0x10, 0x01, 0x07, 0x03, "Very Easy"         },
	{0x10, 0x01, 0x07, 0x02, "Easy 2"            },
	{0x10, 0x01, 0x07, 0x01, "Easy 1"            },
	{0x10, 0x01, 0x07, 0x00, "Normal"            },
	{0x10, 0x01, 0x07, 0x04, "Difficult 1"       },
	{0x10, 0x01, 0x07, 0x05, "Difficult 2"       },
	{0x10, 0x01, 0x07, 0x06, "Difficult 3"       },
	{0x10, 0x01, 0x07, 0x07, "Very Difficult"    },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x10, 0x01, 0x08, 0x00, "Off"               },
//	{0x10, 0x01, 0x08, 0x08, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x10, 0x01, 0x10, 0x00, "Off"               },
//	{0x10, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 6   , "Initital Vitality" },
	{0x10, 0x01, 0xe0, 0x60, "1"                 },
	{0x10, 0x01, 0xe0, 0x80, "2"                 },
	{0x10, 0x01, 0xe0, 0xa0, "3"                 },
	{0x10, 0x01, 0xe0, 0xc0, "4"                 },
	{0x10, 0x01, 0xe0, 0xe0, "5"                 },
	{0x10, 0x01, 0xe0, 0x00, "6"                 },

	// Dip C
//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x11, 0x01, 0x01, 0x00, "Off"               },
//	{0x11, 0x01, 0x01, 0x01, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x11, 0x01, 0x02, 0x00, "Off"               },
//	{0x11, 0x01, 0x02, 0x02, "On"                },

	{0   , 0xfe, 0   , 2   , "Free Play"         },
	{0x11, 0x01, 0x04, 0x00, "Off"               },
	{0x11, 0x01, 0x04, 0x04, "On"                },

	{0   , 0xfe, 0   , 2   , "Stop"              },
	{0x11, 0x01, 0x08, 0x00, "Off"               },
	{0x11, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 2   , "Flip"              },
	{0x11, 0x01, 0x10, 0x00, "Off"               },
	{0x11, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 2   , "Demo Sound"        },
	{0x11, 0x01, 0x20, 0x00, "Off"               },
	{0x11, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue"          },
	{0x11, 0x01, 0x40, 0x00, "Off"               },
	{0x11, 0x01, 0x40, 0x40, "On"                },
};

STDDIPINFO(qtono2);

static struct BurnDIPInfo sf2DIPList[]=
{
	// Defaults
        {0x1b, 0xff, 0xff, 0x40, NULL                },
        {0x1c, 0xff, 0xff, 0x03, NULL                }, 
        {0x1d, 0xff, 0xff, 0x60, NULL                }, 


	// Dip A
	{0   , 0xfe, 0   , 8   , "A Side"                 },
	{0x1b, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x1b, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x1b, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x1b, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x1b, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "B Side"                 },
	{0x1b, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x1b, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x1b, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x1b, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x1b, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x1b, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x1b, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x1b, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x1b, 0x01, 0x40, 0x00, "Off"                    },
	{0x1b, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x80, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Game Level"             },
	{0x1c, 0x01, 0x07, 0x00, "0"                 },
	{0x1c, 0x01, 0x07, 0x01, "1"                 },
	{0x1c, 0x01, 0x07, 0x02, "2"                 },
	{0x1c, 0x01, 0x07, 0x03, "3"              },
	{0x1c, 0x01, 0x07, 0x04, "4"            },
	{0x1c, 0x01, 0x07, 0x05, "5"            },
	{0x1c, 0x01, 0x07, 0x06, "6"            },
	{0x1c, 0x01, 0x07, 0x07, "7"         },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1c, 0x01, 0x08, 0x00, "Off"                    },
//	{0x1c, 0x01, 0x08, 0x08, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1c, 0x01, 0x10, 0x00, "Off"                    },
//      {0x1c, 0x01, 0x10, 0x10, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1c, 0x01, 0x20, 0x00, "Off"                    },
//	{0x1c, 0x01, 0x20, 0x20, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1c, 0x01, 0x40, 0x00, "Off"                    },
//      {0x1c, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1c, 0x01, 0x80, 0x00, "Off"                    },
//      {0x1c, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
//	{0   , 0xfe, 0   , 4   , "Unknown"                },
//      {0x1d, 0x01, 0x03, 0x00, "1"                      },
//      {0x1d, 0x01, 0x03, 0x01, "2"                      },
//      {0x1d, 0x01, 0x03, 0x02, "3"                      },
//      {0x1d, 0x01, 0x03, 0x03, "4"                      },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
        {0x1d, 0x01, 0x04, 0x00, "Off"                    },
        {0x1d, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Stop"                   },
        {0x1d, 0x01, 0x08, 0x00, "Off"                    },
        {0x1d, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
        {0x1d, 0x01, 0x10, 0x00, "Off"                    },
        {0x1d, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
        {0x1d, 0x01, 0x20, 0x00, "Off"                    },
        {0x1d, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
        {0x1d, 0x01, 0x40, 0x00, "Off"                    },
        {0x1d, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(sf2);

static struct BurnDIPInfo sf2jDIPList[]=
{
	// Defaults
        {0x1b, 0xff, 0xff, 0x40, NULL                },
        {0x1c, 0xff, 0xff, 0x03, NULL                }, 
        {0x1d, 0xff, 0xff, 0x60, NULL                }, 


	// Dip A
	{0   , 0xfe, 0   , 8   , "A Side"                 },
	{0x1b, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x1b, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x1b, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x1b, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x1b, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "B Side"                 },
	{0x1b, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x1b, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x1b, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x1b, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x1b, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x1b, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x1b, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x1b, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x1b, 0x01, 0x40, 0x00, "Off"                    },
	{0x1b, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x80, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Game Level"             },
	{0x1c, 0x01, 0x07, 0x00, "0"                 },
	{0x1c, 0x01, 0x07, 0x01, "1"                 },
	{0x1c, 0x01, 0x07, 0x02, "2"                 },
	{0x1c, 0x01, 0x07, 0x03, "3"              },
	{0x1c, 0x01, 0x07, 0x04, "4"            },
	{0x1c, 0x01, 0x07, 0x05, "5"            },
	{0x1c, 0x01, 0x07, 0x06, "6"            },
	{0x1c, 0x01, 0x07, 0x07, "7"         },

	{0   , 0xfe, 0   , 2   , "2 Players Game"         },
	{0x1c, 0x01, 0x08, 0x00, "1 Credit/No Continue"   },
	{0x1c, 0x01, 0x08, 0x08, "2 Credits/Winner Continue"},

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1c, 0x01, 0x10, 0x00, "Off"                    },
//      {0x1c, 0x01, 0x10, 0x10, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1c, 0x01, 0x20, 0x00, "Off"                    },
//	{0x1c, 0x01, 0x20, 0x20, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1c, 0x01, 0x40, 0x00, "Off"                    },
//      {0x1c, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1c, 0x01, 0x80, 0x00, "Off"                    },
//      {0x1c, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
//	{0   , 0xfe, 0   , 4   , "Unknown"                },
//      {0x1d, 0x01, 0x03, 0x00, "1"                      },
//      {0x1d, 0x01, 0x03, 0x01, "2"                      },
//      {0x1d, 0x01, 0x03, 0x02, "3"                      },
//      {0x1d, 0x01, 0x03, 0x03, "4"                      },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
        {0x1d, 0x01, 0x04, 0x00, "Off"                    },
        {0x1d, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Stop"                   },
        {0x1d, 0x01, 0x08, 0x00, "Off"                    },
        {0x1d, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
        {0x1d, 0x01, 0x10, 0x00, "Off"                    },
        {0x1d, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
        {0x1d, 0x01, 0x20, 0x00, "Off"                    },
        {0x1d, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
        {0x1d, 0x01, 0x40, 0x00, "Off"                    },
        {0x1d, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(sf2j);

static struct BurnDIPInfo Sf2ceDIPList[]=
{
	// Defaults
        {0x1a, 0xff, 0xff, 0x40, NULL                },
        {0x1b, 0xff, 0xff, 0x03, NULL                },
        {0x1c, 0xff, 0xff, 0x60, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "A Side"                 },
	{0x1a, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x1a, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x1a, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x1a, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x1a, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x1a, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x1a, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x1a, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "B Side"                 },
	{0x1a, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x1a, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x1a, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x1a, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x1a, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x1a, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x1a, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x1a, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x1a, 0x01, 0x40, 0x00, "Off"                    },
	{0x1a, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1a, 0x01, 0x80, 0x00, "Off"                    },
//	{0x1a, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Game Level"             },
	{0x1b, 0x01, 0x07, 0x00, "0"                 },
	{0x1b, 0x01, 0x07, 0x01, "1"                 },
	{0x1b, 0x01, 0x07, 0x02, "2"                 },
	{0x1b, 0x01, 0x07, 0x03, "3"              },
	{0x1b, 0x01, 0x07, 0x04, "4"            },
	{0x1b, 0x01, 0x07, 0x05, "5"            },
	{0x1b, 0x01, 0x07, 0x06, "6"            },
	{0x1b, 0x01, 0x07, 0x07, "7"         },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x08, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x08, 0x08, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1b, 0x01, 0x10, 0x00, "Off"                    },
//      {0x1b, 0x01, 0x10, 0x10, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x20, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x20, 0x20, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1b, 0x01, 0x40, 0x00, "Off"                    },
//      {0x1b, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1b, 0x01, 0x80, 0x00, "Off"                    },
//      {0x1b, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
//	{0   , 0xfe, 0   , 4   , "Unknown"                },
//      {0x1c, 0x01, 0x03, 0x00, "1"                      },
//      {0x1c, 0x01, 0x03, 0x01, "2"                      },
//      {0x1c, 0x01, 0x03, 0x02, "3"                      },
//      {0x1c, 0x01, 0x03, 0x03, "4"                      },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
        {0x1c, 0x01, 0x04, 0x00, "Off"                    },
        {0x1c, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Stop"                   },
        {0x1c, 0x01, 0x08, 0x00, "Off"                    },
        {0x1c, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
        {0x1c, 0x01, 0x10, 0x00, "Off"                    },
        {0x1c, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
        {0x1c, 0x01, 0x20, 0x00, "Off"                    },
        {0x1c, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
        {0x1c, 0x01, 0x40, 0x00, "Off"                    },
        {0x1c, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(Sf2ce);

static struct BurnDIPInfo Sf2cehacksDIPList[]=
{
	// Defaults
        {0x1b, 0xff, 0xff, 0x40, NULL                }, // or 1b on hacks 
        {0x1c, 0xff, 0xff, 0x03, NULL                }, // or 1c "      "
        {0x1d, 0xff, 0xff, 0x60, NULL                }, // or 1d "      " 

	// Dip A
	{0   , 0xfe, 0   , 8   , "A Side"                 },
	{0x1b, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x1b, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x1b, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x1b, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x1b, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x1b, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "B Side"                 },
	{0x1b, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x1b, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x1b, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x1b, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x1b, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x1b, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x1b, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x1b, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x1b, 0x01, 0x40, 0x00, "Off"                    },
	{0x1b, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x80, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Game Level"             },
	{0x1c, 0x01, 0x07, 0x00, "0"                 	  },
	{0x1c, 0x01, 0x07, 0x01, "1" 	                  },
	{0x1c, 0x01, 0x07, 0x02, "2"	                  },
	{0x1c, 0x01, 0x07, 0x03, "3"		              },
	{0x1c, 0x01, 0x07, 0x04, "4"   			          },
	{0x1c, 0x01, 0x07, 0x05, "5"		              },
	{0x1c, 0x01, 0x07, 0x06, "6"		              },
	{0x1c, 0x01, 0x07, 0x07, "7"			          },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1c, 0x01, 0x08, 0x00, "Off"                    },
//	{0x1c, 0x01, 0x08, 0x08, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1c, 0x01, 0x10, 0x00, "Off"                    },
//	{0x1c, 0x01, 0x10, 0x10, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1c, 0x01, 0x20, 0x00, "Off"                    },
//	{0x1c, 0x01, 0x20, 0x20, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1c, 0x01, 0x40, 0x00, "Off"                    },
//	{0x1c, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1c, 0x01, 0x80, 0x00, "Off"                    },
//	{0x1c, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
//	{0   , 0xfe, 0   , 4   , "Unknown"                },
//	{0x1d, 0x01, 0x03, 0x00, "1"                      },
//	{0x1d, 0x01, 0x03, 0x01, "2"                      },
//	{0x1d, 0x01, 0x03, 0x02, "3"                      },
//	{0x1d, 0x01, 0x03, 0x03, "4"                      },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x1d, 0x01, 0x04, 0x00, "Off"                    },
	{0x1d, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Stop"                   },
	{0x1d, 0x01, 0x08, 0x00, "Off"                    },
	{0x1d, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x1d, 0x01, 0x10, 0x00, "Off"                    },
	{0x1d, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x1d, 0x01, 0x20, 0x00, "Off"                    },
	{0x1d, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x1d, 0x01, 0x40, 0x00, "Off"                    },
	{0x1d, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(Sf2cehacks);

static struct BurnDIPInfo Sf2cejDIPList[]=
{
	// Defaults
        {0x1a, 0xff, 0xff, 0x40, NULL                },
        {0x1b, 0xff, 0xff, 0x03, NULL                },
        {0x1c, 0xff, 0xff, 0x60, NULL                },

	// Dip A
	{0   , 0xfe, 0   , 8   , "A Side"                 },
	{0x1a, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x1a, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x1a, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x1a, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x1a, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x1a, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x1a, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x1a, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "B Side"                 },
	{0x1a, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x1a, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x1a, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x1a, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x1a, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x1a, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x1a, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x1a, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 2   , "Continue Coin"          },
	{0x1a, 0x01, 0x40, 0x00, "Off"                    },
	{0x1a, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1a, 0x01, 0x80, 0x00, "Off"                    },
//	{0x1a, 0x01, 0x80, 0x80, "On"                     },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Game Level"             },
	{0x1b, 0x01, 0x07, 0x00, "0"                 },
	{0x1b, 0x01, 0x07, 0x01, "1"                 },
	{0x1b, 0x01, 0x07, 0x02, "2"                 },
	{0x1b, 0x01, 0x07, 0x03, "3"              },
	{0x1b, 0x01, 0x07, 0x04, "4"            },
	{0x1b, 0x01, 0x07, 0x05, "5"            },
	{0x1b, 0x01, 0x07, 0x06, "6"            },
	{0x1b, 0x01, 0x07, 0x07, "7"         },

	{0   , 0xfe, 0   , 2   , "2 Players Game"         },
	{0x1b, 0x01, 0x08, 0x00, "1 Credit/No Continue"   },
	{0x1b, 0x01, 0x08, 0x08, "2 Credits/Winner Continue"},

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1b, 0x01, 0x10, 0x00, "Off"                    },
//      {0x1b, 0x01, 0x10, 0x10, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x1b, 0x01, 0x20, 0x00, "Off"                    },
//	{0x1b, 0x01, 0x20, 0x20, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1b, 0x01, 0x40, 0x00, "Off"                    },
//      {0x1b, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//      {0x1b, 0x01, 0x80, 0x00, "Off"                    },
//      {0x1b, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
//	{0   , 0xfe, 0   , 4   , "Unknown"                },
//      {0x1c, 0x01, 0x03, 0x00, "1"                      },
//      {0x1c, 0x01, 0x03, 0x01, "2"                      },
//      {0x1c, 0x01, 0x03, 0x02, "3"                      },
//      {0x1c, 0x01, 0x03, 0x03, "4"                      },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
        {0x1c, 0x01, 0x04, 0x00, "Off"                    },
        {0x1c, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Stop"                   },
        {0x1c, 0x01, 0x08, 0x00, "Off"                    },
        {0x1c, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
        {0x1c, 0x01, 0x10, 0x00, "Off"                    },
        {0x1c, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
        {0x1c, 0x01, 0x20, 0x00, "Off"                    },
        {0x1c, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
        {0x1c, 0x01, 0x40, 0x00, "Off"                    },
        {0x1c, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(Sf2cej);

static struct BurnDIPInfo slammastDIPList[]=
{
	// Defaults
	{0x23, 0xff, 0xff, 0x08, NULL                },

	// Dip C
	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x23, 0x01, 0x08, 0x08, "Off"                    },
	{0x23, 0x01, 0x08, 0x00, "On"                     },
};

STDDIPINFO(slammast);

static struct BurnDIPInfo striderDIPList[]=
{
	// Defaults
	{0x15, 0xff, 0xff, 0x00, NULL                     },
	{0x16, 0xff, 0xff, 0x00, NULL                     },
	{0x17, 0xff, 0xff, 0x00, NULL                     },

	// Dip A
	{0   , 0xfe, 0   , 8   , "A Side"                 },
	{0x15, 0x01, 0x07, 0x07, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x06, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x05, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x07, 0x04, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 8   , "B Side"                 },
	{0x15, 0x01, 0x38, 0x38, "4 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x30, "3 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x28, "2 Coins 1 Credit"       },
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"        },
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"       },
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"       },
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"       },
	{0x15, 0x01, 0x38, 0x20, "1 Coin 6 Credits"       },

	{0   , 0xfe, 0   , 3   , "Table Type"             },
	{0x15, 0x01, 0xc0, 0x00, "Upright 1P"            },
	{0x15, 0x01, 0xc0, 0x40, "Upright 2P"            },
	{0x15, 0x01, 0xc0, 0x80, "Cocktail"               },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Game Level"             },
	{0x16, 0x01, 0x07, 0x03, "1"                 },
	{0x16, 0x01, 0x07, 0x02, "2"                 },
	{0x16, 0x01, 0x07, 0x01, "3"              },
	{0x16, 0x01, 0x07, 0x00, "4"                 },
	{0x16, 0x01, 0x07, 0x04, "5"            },
	{0x16, 0x01, 0x07, 0x05, "6"            },
	{0x16, 0x01, 0x07, 0x06, "7"            },
	{0x16, 0x01, 0x07, 0x07, "8"         },

//	{0   , 0xfe, 0   , 2   , "Continue Coinage?"      },
//	{0x16, 0x01, 0x08, 0x00, "1 Coin"                 },
//	{0x16, 0x01, 0x08, 0x08, "2 Coins"                },

	{0   , 0xfe, 0   , 4   , "Bonus"                  },
	{0x16, 0x01, 0x30, 0x20, "20k, 60k"               },
	{0x16, 0x01, 0x30, 0x30, "30k, 60k"               },
	{0x16, 0x01, 0x30, 0x00, "20k, 40k, 60k"          },
	{0x16, 0x01, 0x30, 0x10, "30k, 50k, 70k"          },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x40, 0x00, "Off"                    },
//	{0x16, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x80, 0x00, "Off"                    },
//	{0x16, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
	{0   , 0xfe, 0   , 4   , "Player"                 },
	{0x17, 0x01, 0x03, 0x00, "3"                      },
	{0x17, 0x01, 0x03, 0x01, "4"                      },
	{0x17, 0x01, 0x03, 0x02, "5"                      },
	{0x17, 0x01, 0x03, 0x03, "6"                      },

	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x17, 0x01, 0x04, 0x00, "Off"                    },
	{0x17, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x17, 0x01, 0x08, 0x00, "Off"                    },
	{0x17, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip"                   },
	{0x17, 0x01, 0x10, 0x00, "Off"                    },
	{0x17, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Sound"                  },
	{0x17, 0x01, 0x20, 0x20, "Off"                    },
	{0x17, 0x01, 0x20, 0x00, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x17, 0x01, 0x40, 0x40, "Off"                    },
	{0x17, 0x01, 0x40, 0x00, "On"                     },
};

STDDIPINFO(strider);

static struct BurnDIPInfo unsquadDIPList[]=
{
	// Defaults
        {0x15, 0xff, 0xff, 0x00, NULL                },
        {0x16, 0xff, 0xff, 0x03, NULL                },
        {0x17, 0xff, 0xff, 0x60, NULL                },

	// Dip A
        {0   , 0xfe, 0   , 7   , "Coin A"            },
	{0x15, 0x01, 0x07, 0x06, "4 Coins 1 Credit"  },
	{0x15, 0x01, 0x07, 0x05, "3 Coins 1 Credit"  },
	{0x15, 0x01, 0x07, 0x04, "2 Coins 1 Credit"  },
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"   },
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"  },
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"  },
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"  },

        {0   , 0xfe, 0   , 7   , "Coin B"            },
	{0x15, 0x01, 0x38, 0x30, "4 Coins 1 Credit"  },
	{0x15, 0x01, 0x38, 0x28, "3 Coins 1 Credit"  },
	{0x15, 0x01, 0x38, 0x20, "2 Coins 1 Credit"  },
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"   },
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"  },
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"  },
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"  },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x15, 0x01, 0x40, 0x00, "Off"               },
//	{0x15, 0x01, 0x40, 0x40, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x15, 0x01, 0x80, 0x00, "Off"               },
//	{0x15, 0x01, 0x80, 0x80, "On"                },

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"        },
	{0x16, 0x01, 0x07, 0x00, "Super Easy"            },
	{0x16, 0x01, 0x07, 0x01, "Very Easy"            },
	{0x16, 0x01, 0x07, 0x02, "Easy"            },
	{0x16, 0x01, 0x07, 0x03, "Normal"         },
	{0x16, 0x01, 0x07, 0x04, "Difficult"       },
	{0x16, 0x01, 0x07, 0x05, "Very Difficult"       },
	{0x16, 0x01, 0x07, 0x06, "Super Difficult"       },
	{0x16, 0x01, 0x07, 0x07, "Ultra Super Difficult"    },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x08, 0x00, "Off"               },
//	{0x16, 0x01, 0x08, 0x08, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x10, 0x00, "Off"               },
//	{0x16, 0x01, 0x10, 0x10, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x20, 0x00, "Off"               },
//	{0x16, 0x01, 0x20, 0x20, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x40, 0x00, "Off"               },
//	{0x16, 0x01, 0x40, 0x40, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x16, 0x01, 0x80, 0x00, "Off"               },
//	{0x16, 0x01, 0x80, 0x80, "On"                },

	// Dip C
//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x17, 0x01, 0x01, 0x00, "Off"               },
//	{0x17, 0x01, 0x01, 0x01, "On"                },

//	{0   , 0xfe, 0   , 2   , "Unknown"           },
//	{0x17, 0x01, 0x02, 0x00, "Off"               },
//	{0x17, 0x01, 0x02, 0x02, "On"                },

	{0   , 0xfe, 0   , 2   , "Free Play"         },
	{0x17, 0x01, 0x04, 0x00, "Off"               },
	{0x17, 0x01, 0x04, 0x04, "On"                },

	{0   , 0xfe, 0   , 2   , "Stop"              },
	{0x17, 0x01, 0x08, 0x00, "Off"               },
	{0x17, 0x01, 0x08, 0x08, "On"                },

	{0   , 0xfe, 0   , 2   , "Flip"              },
	{0x17, 0x01, 0x10, 0x00, "Off"               },
	{0x17, 0x01, 0x10, 0x10, "On"                },

	{0   , 0xfe, 0   , 2   , "Demo Sound"        },
	{0x17, 0x01, 0x20, 0x00, "Off"               },
	{0x17, 0x01, 0x20, 0x20, "On"                },

	{0   , 0xfe, 0   , 2   , "Continue"          },
	{0x17, 0x01, 0x40, 0x00, "No"                },
	{0x17, 0x01, 0x40, 0x40, "Yes"               },
};

STDDIPINFO(unsquad);


static struct BurnDIPInfo varthDIPList[]=
{
	// Defaults
        {0x15, 0xff, 0xff, 0x40, NULL			},
        {0x16, 0xff, 0xff, 0x2B, NULL			},
        {0x17, 0xff, 0xff, 0x60, NULL			},

	// Dip A
	{0   , 0xfe, 0   , 8   , "Coin A"		},
	{0x15, 0x01, 0x07, 0x07, "4 Coins 1 Credit"	},
	{0x15, 0x01, 0x07, 0x06, "3 Coins 1 Credit"	},
	{0x15, 0x01, 0x07, 0x05, "2 Coins 1 Credit"	},
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"	},
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"	},
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"	},
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"	},
	{0x15, 0x01, 0x07, 0x04, "1 Coin 6 Credits"	},

	{0   , 0xfe, 0   , 8   , "Coin B"		},
	{0x15, 0x01, 0x38, 0x38, "4 Coins 1 Credit" 	},
	{0x15, 0x01, 0x38, 0x30, "3 Coins 1 Credit"	},
	{0x15, 0x01, 0x38, 0x28, "2 Coins 1 Credit"	},
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"	},
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"	},
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"	},
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"	},
	{0x15, 0x01, 0x38, 0x20, "1 Coin 6 Credits"	},

	{0   , 0xfe, 0   , 2   , "Continue Coin"	},
	{0x15, 0x01, 0x40, 0x00, "Off"			},
	{0x15, 0x01, 0x40, 0x40, "On"			},

//	{0   , 0xfe, 0   , 2   , "Unknown"		},
//	{0x15, 0x01, 0x80, 0x00, "Off"			},
//	{0x15, 0x01, 0x80, 0x80, "On"			},

	// Dip B
	{0   , 0xfe, 0   , 8   , "Difficulty"		},
	{0x16, 0x01, 0x07, 0x00, "Easiest"		},
	{0x16, 0x01, 0x07, 0x01, "Easy 1"		},
	{0x16, 0x01, 0x07, 0x02, "Easy 2"		},
	{0x16, 0x01, 0x07, 0x03, "Normal"		},
	{0x16, 0x01, 0x07, 0x04, "Difficult"		},
	{0x16, 0x01, 0x07, 0x05, "Very Difficult"	},
	{0x16, 0x01, 0x07, 0x06, "Hard"			},
	{0x16, 0x01, 0x07, 0x07, "Hardest"		},

	{0   , 0xfe, 0   , 4   , "Bonus Life"		},
	{0x16, 0x01, 0x18, 0x00, "600k and every 1400k"			},
	{0x16, 0x01, 0x18, 0x08, "600k, 2000k and every 4500k"		},
	{0x16, 0x01, 0x18, 0x10, "1200k, 3500k"				},
	{0x16, 0x01, 0x18, 0x18, "2000k only"		},

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x20, 0x00, "Off"                    },
//	{0x16, 0x01, 0x20, 0x20, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x40, 0x00, "Off"                    },
//	{0x16, 0x01, 0x40, 0x40, "On"                     },

//	{0   , 0xfe, 0   , 2   , "Unknown"                },
//	{0x16, 0x01, 0x80, 0x00, "Off"                    },
//	{0x16, 0x01, 0x80, 0x80, "On"                     },

	// Dip C
	{0   , 0xfe, 0   , 4   , "Lives"                  },
	{0x17, 0x01, 0x03, 0x01, "1"                      },
	{0x17, 0x01, 0x03, 0x02, "2"                      },
	{0x17, 0x01, 0x03, 0x00, "3"                      },
	{0x17, 0x01, 0x03, 0x03, "4"                      },

	{0   , 0xfe, 0   , 2   , "Free Play"              },
	{0x17, 0x01, 0x04, 0x00, "Off"                    },
	{0x17, 0x01, 0x04, 0x04, "On"                     },

	{0   , 0xfe, 0   , 2   , "Stop"                   },
	{0x17, 0x01, 0x08, 0x00, "Off"                    },
	{0x17, 0x01, 0x08, 0x08, "On"                     },

	{0   , 0xfe, 0   , 2   , "Flip Screen"            },
	{0x17, 0x01, 0x10, 0x00, "Off"                    },
	{0x17, 0x01, 0x10, 0x10, "On"                     },

	{0   , 0xfe, 0   , 2   , "Demo Sound"             },
	{0x17, 0x01, 0x20, 0x00, "Off"                    },
	{0x17, 0x01, 0x20, 0x20, "On"                     },

	{0   , 0xfe, 0   , 2   , "Continue"               },
	{0x17, 0x01, 0x40, 0x00, "Off"                    },
	{0x17, 0x01, 0x40, 0x40, "On"                     },
};

STDDIPINFO(varth);

static struct BurnDIPInfo willowDIPList[]=
{
	// Defaults
	{0x15, 0xff, 0xff, 0x00, NULL			},
	{0x16, 0xff, 0xff, 0x00, NULL			},
	{0x17, 0xff, 0xff, 0x00, NULL			},

	// Dip A
        {0   , 0xfe, 0   , 7   , "A Side"		},
	{0x15, 0x01, 0x07, 0x06, "4 Coins 1 Credit"	},
	{0x15, 0x01, 0x07, 0x05, "3 Coins 1 Credit"	},
	{0x15, 0x01, 0x07, 0x04, "2 Coins 1 Credit"	},
	{0x15, 0x01, 0x07, 0x00, "1 Coin 1 Credit"	},
	{0x15, 0x01, 0x07, 0x01, "1 Coin 2 Credits"	},
	{0x15, 0x01, 0x07, 0x02, "1 Coin 3 Credits"	},
	{0x15, 0x01, 0x07, 0x03, "1 Coin 4 Credits"	},

        {0   , 0xfe, 0   , 7   , "B Side"		},
	{0x15, 0x01, 0x38, 0x30, "4 Coins 1 Credit"	},
	{0x15, 0x01, 0x38, 0x28, "3 Coins 1 Credit"	},
	{0x15, 0x01, 0x38, 0x20, "2 Coins 1 Credit"	},
	{0x15, 0x01, 0x38, 0x00, "1 Coin 1 Credit"	},
	{0x15, 0x01, 0x38, 0x08, "1 Coin 2 Credits"	},
	{0x15, 0x01, 0x38, 0x10, "1 Coin 3 Credits"	},
	{0x15, 0x01, 0x38, 0x18, "1 Coin 4 Credits"	},

        {0   , 0xfe, 0   , 3   , "Table Type"		},
	{0x15, 0x01, 0xc0, 0x00, "Upright 1P"		},
	{0x15, 0x01, 0xc0, 0x40, "Upright 2P"		},
	{0x15, 0x01, 0xc0, 0xc0, "Cocktail"		},

	// Dip B
	{0   , 0xfe, 0   , 8   , "Game Level"		},
	{0x16, 0x01, 0x07, 0x03, "Very Easy"		},
	{0x16, 0x01, 0x07, 0x02, "Easy 2"		},
	{0x16, 0x01, 0x07, 0x01, "Easy 1"		},
	{0x16, 0x01, 0x07, 0x00, "Normal"		},
	{0x16, 0x01, 0x07, 0x04, "Difficult 1"		},
	{0x16, 0x01, 0x07, 0x05, "Difficult 2"		},
	{0x16, 0x01, 0x07, 0x06, "Difficult 3"		},
	{0x16, 0x01, 0x07, 0x07, "Very Difficult"	},

	{0   , 0xfe, 0   , 4   , "Nando Speed"		},
	{0x16, 0x01, 0x18, 0x00, "Normal"		},
	{0x16, 0x01, 0x18, 0x08, "Slow"			},
	{0x16, 0x01, 0x18, 0x10, "Fast"			},
	{0x16, 0x01, 0x18, 0x18, "Very Fast"		},

//	{0   , 0xfe, 0   , 2   , "Unknown"		},
//	{0x16, 0x01, 0x20, 0x00, "Off"			},
//	{0x16, 0x01, 0x20, 0x20, "On"			},

//	{0   , 0xfe, 0   , 2   , "Unknown"		},
//	{0x16, 0x01, 0x40, 0x00, "Off"			},
//	{0x16, 0x01, 0x40, 0x40, "On"			},

	{0   , 0xfe, 0   , 2   , "Stage Magic Continue"	},
	{0x16, 0x01, 0x80, 0x00, "Off"			},
	{0x16, 0x01, 0x80, 0x80, "On"			},

	// Dip C
	{0   , 0xfe, 0   , 4   , "Player"		},
	{0x17, 0x01, 0x03, 0x01, "1 Player"		},
	{0x17, 0x01, 0x03, 0x00, "2 Players"		},
	{0x17, 0x01, 0x03, 0x02, "3 Players"		},
	{0x17, 0x01, 0x03, 0x03, "4 Players"		},

	{0   , 0xfe, 0   , 4   , "Life"			},
	{0x17, 0x01, 0x0c, 0x0c, "2"			},
	{0x17, 0x01, 0x0c, 0x00, "3"			},
	{0x17, 0x01, 0x0c, 0x04, "4"			},
	{0x17, 0x01, 0x0c, 0x08, "5"			},

	{0   , 0xfe, 0   , 2   , "Screen"		},
	{0x17, 0x01, 0x10, 0x00, "Off"			},
	{0x17, 0x01, 0x10, 0x10, "On"			},

	{0   , 0xfe, 0   , 2   , "Sound"		},
	{0x17, 0x01, 0x20, 0x20, "Off"			},
	{0x17, 0x01, 0x20, 0x00, "On"			},

	{0   , 0xfe, 0   , 2   , "Continue"		},
	{0x17, 0x01, 0x40, 0x40, "Off"			},
	{0x17, 0x01, 0x40, 0x00, "On"			},
};

STDDIPINFO(willow);

static struct BurnDIPInfo wofDIPList[]=
{
	// Defaults
	{0x1b, 0xff, 0xff, 0x80, NULL			},

	// Dip C
	{0   , 0xfe, 0   , 2   , "Freeze"		},
	{0x1b, 0x01, 0x08, 0x08, "Off"			},
	{0x1b, 0x01, 0x08, 0x00, "On"			},
};

STDDIPINFO(wof);


//----------------------------------------------------------------------------------------------
// Drivers


// Forgotten Worlds (US)

static struct BurnRomInfo ForgottnRomDesc[] = {
	{ "lwu11a",        0x020000, 0xddf78831, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "lwu15a",        0x020000, 0xf7ce2097, 1 | BRF_ESS | BRF_PRG },
	{ "lwu10a",        0x020000, 0x8cb38c81, 1 | BRF_ESS | BRF_PRG },
	{ "lwu14a",        0x020000, 0xd70ef9fd, 1 | BRF_ESS | BRF_PRG },
	{ "lw-07",         0x080000, 0xfd252a26, 1 | BRF_ESS | BRF_PRG },

	{ "lw-02",         0x080000, 0x43e6c5c8, 2 | BRF_GRA },
	{ "lw-09",         0x080000, 0x899cb4ad, 2 | BRF_GRA },
	{ "lw-06",         0x080000, 0x5b9edffc, 2 | BRF_GRA },
	{ "lw-13",         0x080000, 0x8e058ef5, 2 | BRF_GRA },
	{ "lw-01",         0x080000, 0x0318f298, 2 | BRF_GRA },
	{ "lw-08",         0x080000, 0x25a8e43c, 2 | BRF_GRA },
	{ "lw-05",         0x080000, 0xe4552fd7, 2 | BRF_GRA },
	{ "lw-12",         0x080000, 0x8e6a832b, 2 | BRF_GRA },

	{ "lwu00",         0x010000, 0x59df2a63, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "lw-03u",        0x020000, 0x807d051f, 4 | BRF_SND },
	{ "lw-04u",        0x020000, 0xe6cd098e, 4 | BRF_SND },
};

STD_ROM_PICK(Forgottn) STD_ROM_FN(Forgottn)

void load_stars()
{
	CpsLoadStars(CpsStar, 5);
}

int forgottnInit()
{
	pCpsInitCallback = load_stars;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsForgottn = {
	"forgottn", NULL, NULL, "1988",
	"Forgotten Worlds (US)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1_GENERIC,
	NULL,ForgottnRomInfo,ForgottnRomName,forgottnInputInfo, forgottnDIPInfo,
	forgottnInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Lost Worlds (Japan)

static struct BurnRomInfo LostwrldRomDesc[] = {
	{ "lw-11c.14f", 0x020000, 0x67e42546, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "lw-15c.14g", 0x020000, 0x402e2a46, 1 | BRF_ESS | BRF_PRG },
	{ "lw-10c.13f", 0x020000, 0xc46479d7, 1 | BRF_ESS | BRF_PRG },
	{ "lw-14c.13g", 0x020000, 0x97670f4a, 1 | BRF_ESS | BRF_PRG },
	{ "lw-07",      0x080000, 0xfd252a26, 1 | BRF_ESS | BRF_PRG },

	{ "lw-02",      0x080000, 0x43e6c5c8, 2 | BRF_GRA },
	{ "lw-09",      0x080000, 0x899cb4ad, 2 | BRF_GRA },
	{ "lw-06",      0x080000, 0x5b9edffc, 2 | BRF_GRA },
	{ "lw-13",      0x080000, 0x8e058ef5, 2 | BRF_GRA },
	{ "lw-01",      0x080000, 0x0318f298, 2 | BRF_GRA },
	{ "lw-08",      0x080000, 0x25a8e43c, 2 | BRF_GRA },
	{ "lw-05",      0x080000, 0xe4552fd7, 2 | BRF_GRA },
	{ "lw-12",      0x080000, 0x8e6a832b, 2 | BRF_GRA },

	{ "lwu00",      0x010000, 0x59df2a63, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "lw-03.14c",  0x020000, 0xce2159e7, 4 | BRF_SND },
	{ "lw-04.13c",  0x020000, 0x39305536, 4 | BRF_SND },
};

STD_ROM_PICK(Lostwrld) STD_ROM_FN(Lostwrld)

struct BurnDriver BurnDrvCpsLostwrld = {
	"lostwrld", "forgottn", NULL, "1988",
	"Lost Worlds (Japan)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1_GENERIC,
	NULL,LostwrldRomInfo,LostwrldRomName,forgottnInputInfo, forgottnDIPInfo,
	forgottnInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Ghouls'n Ghosts (World)

static struct BurnRomInfo GhoulsRomDesc[] = {
	{ "ghl29.bin", 0x020000, 0x166a58a2, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "ghl30.bin", 0x020000, 0x7ac8407a, 1 | BRF_ESS | BRF_PRG },
	{ "ghl27.bin", 0x020000, 0xf734b2be, 1 | BRF_ESS | BRF_PRG },
	{ "ghl28.bin", 0x020000, 0x03d3e714, 1 | BRF_ESS | BRF_PRG },
	{ "ghl17.bin", 0x080000, 0x3ea1b0f2, 1 | BRF_ESS | BRF_PRG },

	{ "ghl5.bin",  0x080000, 0x0ba9c0b0, 2 | BRF_GRA },
	{ "ghl7.bin",  0x080000, 0x5d760ab9, 2 | BRF_GRA },
	{ "ghl6.bin",  0x080000, 0x4ba90b59, 2 | BRF_GRA },
	{ "ghl8.bin",  0x080000, 0x4bdee9de, 2 | BRF_GRA },
	{ "ghl09.bin", 0x010000, 0xae24bb19, 2 | BRF_GRA },
	{ "ghl18.bin", 0x010000, 0xd34e271a, 2 | BRF_GRA },
	{ "ghl13.bin", 0x010000, 0x3f70dd37, 2 | BRF_GRA },
	{ "ghl22.bin", 0x010000, 0x7e69e2e6, 2 | BRF_GRA },
	{ "ghl11.bin", 0x010000, 0x37c9b6c6, 2 | BRF_GRA },
	{ "ghl20.bin", 0x010000, 0x2f1345b4, 2 | BRF_GRA },
	{ "ghl15.bin", 0x010000, 0x3c2a212a, 2 | BRF_GRA },
	{ "ghl24.bin", 0x010000, 0x889aac05, 2 | BRF_GRA },
	{ "ghl10.bin", 0x010000, 0xbcc0f28c, 2 | BRF_GRA },
	{ "ghl19.bin", 0x010000, 0x2a40166a, 2 | BRF_GRA },
	{ "ghl14.bin", 0x010000, 0x20f85c03, 2 | BRF_GRA },
	{ "ghl23.bin", 0x010000, 0x8426144b, 2 | BRF_GRA },
	{ "ghl12.bin", 0x010000, 0xda088d61, 2 | BRF_GRA },
	{ "ghl21.bin", 0x010000, 0x17e11df0, 2 | BRF_GRA },
	{ "ghl16.bin", 0x010000, 0xf187ba1c, 2 | BRF_GRA },
	{ "ghl25.bin", 0x010000, 0x29f79c78, 2 | BRF_GRA },

	{ "ghl26.bin", 0x010000, 0x3692f6e5, 3 | BRF_ESS | BRF_PRG }, // z80 program
};

STD_ROM_PICK(Ghouls); STD_ROM_FN(Ghouls);

void ghouls_program_fix()
{
	if (!strcmp(BurnDrvGetTextA(DRV_NAME), "ghouls"))
	{
		// Patch out self-test... it takes forever
		*((unsigned short*)(CpsRom + 0x61964)) = 0x4ef9;
		*((unsigned short*)(CpsRom + 0x61966)) = 0x0000;
		*((unsigned short*)(CpsRom + 0x61968)) = 0x0400;
	}

	BurnByteswap(CpsRom + 0x080000, 0x080000);
}

int ghoulsInit()
{
	pCpsInitCallback = ghouls_program_fix;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsGhouls = {
	"ghouls", NULL, NULL, "1988",
	"Ghouls'n Ghosts (World)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,GhoulsRomInfo,GhoulsRomName,ghoulsInputInfo, ghoulsDIPInfo,
	ghoulsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Ghouls'n Ghosts (US)

static struct BurnRomInfo GhoulsuRomDesc[] = {
	{ "dmu29",     0x020000, 0x334d85b2, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "dmu30",     0x020000, 0xcee8ceb5, 1 | BRF_ESS | BRF_PRG },
	{ "dmu27",     0x020000, 0x4a524140, 1 | BRF_ESS | BRF_PRG },
	{ "dmu28",     0x020000, 0x94aae205, 1 | BRF_ESS | BRF_PRG },
	{ "ghl17.bin", 0x080000, 0x3ea1b0f2, 1 | BRF_ESS | BRF_PRG },

	{ "ghl5.bin",  0x080000, 0x0ba9c0b0, 2 | BRF_GRA },
	{ "ghl7.bin",  0x080000, 0x5d760ab9, 2 | BRF_GRA },
	{ "ghl6.bin",  0x080000, 0x4ba90b59, 2 | BRF_GRA },
	{ "ghl8.bin",  0x080000, 0x4bdee9de, 2 | BRF_GRA },
	{ "ghl09.bin", 0x010000, 0xae24bb19, 2 | BRF_GRA },
	{ "ghl18.bin", 0x010000, 0xd34e271a, 2 | BRF_GRA },
	{ "ghl13.bin", 0x010000, 0x3f70dd37, 2 | BRF_GRA },
	{ "ghl22.bin", 0x010000, 0x7e69e2e6, 2 | BRF_GRA },
	{ "ghl11.bin", 0x010000, 0x37c9b6c6, 2 | BRF_GRA },
	{ "ghl20.bin", 0x010000, 0x2f1345b4, 2 | BRF_GRA },
	{ "ghl15.bin", 0x010000, 0x3c2a212a, 2 | BRF_GRA },
	{ "ghl24.bin", 0x010000, 0x889aac05, 2 | BRF_GRA },
	{ "ghl10.bin", 0x010000, 0xbcc0f28c, 2 | BRF_GRA },
	{ "ghl19.bin", 0x010000, 0x2a40166a, 2 | BRF_GRA },
	{ "ghl14.bin", 0x010000, 0x20f85c03, 2 | BRF_GRA },
	{ "ghl23.bin", 0x010000, 0x8426144b, 2 | BRF_GRA },
	{ "ghl12.bin", 0x010000, 0xda088d61, 2 | BRF_GRA },
	{ "ghl21.bin", 0x010000, 0x17e11df0, 2 | BRF_GRA },
	{ "ghl16.bin", 0x010000, 0xf187ba1c, 2 | BRF_GRA },
	{ "ghl25.bin", 0x010000, 0x29f79c78, 2 | BRF_GRA },

	{ "ghl26.bin", 0x010000, 0x3692f6e5, 3 | BRF_ESS | BRF_PRG }, // z80 program
};

STD_ROM_PICK(Ghoulsu); STD_ROM_FN(Ghoulsu);

struct BurnDriver BurnDrvCpsGhoulsu = {
	"ghoulsu", "ghouls", NULL, "1988",
	"Ghouls'n Ghosts (US)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,GhoulsuRomInfo,GhoulsuRomName,ghoulsInputInfo, ghoulsDIPInfo,
	ghoulsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Dai Makai-Mura (Japan)

static struct BurnRomInfo daimakaiRomDesc[] = {
	{ "dmj_38.bin", 0x020000, 0x82fd1798, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "dmj_39.bin", 0x020000, 0x35366ccc, 1 | BRF_ESS | BRF_PRG },
	{ "dmj_40.bin", 0x020000, 0xa17c170a, 1 | BRF_ESS | BRF_PRG },
	{ "dmj_41.bin", 0x020000, 0x6af0b391, 1 | BRF_ESS | BRF_PRG },
	{ "ghl17.bin",  0x080000, 0x3ea1b0f2, 1 | BRF_ESS | BRF_PRG },

	{ "ghl5.bin",   0x080000, 0x0ba9c0b0, 2 | BRF_GRA },
	{ "ghl7.bin",   0x080000, 0x5d760ab9, 2 | BRF_GRA },
	{ "ghl6.bin",   0x080000, 0x4ba90b59, 2 | BRF_GRA },
	{ "ghl8.bin",   0x080000, 0x4bdee9de, 2 | BRF_GRA },
	{ "ghl09.bin",  0x010000, 0xae24bb19, 2 | BRF_GRA },
	{ "ghl18.bin",  0x010000, 0xd34e271a, 2 | BRF_GRA },
	{ "ghl13.bin",  0x010000, 0x3f70dd37, 2 | BRF_GRA },
	{ "ghl22.bin",  0x010000, 0x7e69e2e6, 2 | BRF_GRA },
	{ "ghl11.bin",  0x010000, 0x37c9b6c6, 2 | BRF_GRA },
	{ "ghl20.bin",  0x010000, 0x2f1345b4, 2 | BRF_GRA },
	{ "ghl15.bin",  0x010000, 0x3c2a212a, 2 | BRF_GRA },
	{ "ghl24.bin",  0x010000, 0x889aac05, 2 | BRF_GRA },
	{ "ghl10.bin",  0x010000, 0xbcc0f28c, 2 | BRF_GRA },
	{ "ghl19.bin",  0x010000, 0x2a40166a, 2 | BRF_GRA },
	{ "ghl14.bin",  0x010000, 0x20f85c03, 2 | BRF_GRA },
	{ "ghl23.bin",  0x010000, 0x8426144b, 2 | BRF_GRA },
	{ "ghl12.bin",  0x010000, 0xda088d61, 2 | BRF_GRA },
	{ "ghl21.bin",  0x010000, 0x17e11df0, 2 | BRF_GRA },
	{ "ghl16.bin",  0x010000, 0xf187ba1c, 2 | BRF_GRA },
	{ "ghl25.bin",  0x010000, 0x29f79c78, 2 | BRF_GRA },

	{ "ghl26.bin",  0x010000, 0x3692f6e5, 3 | BRF_ESS | BRF_PRG }, // z80 program
};

STD_ROM_PICK(daimakai); STD_ROM_FN(daimakai);

struct BurnDriver BurnDrvCpsDaimakai = {
	"daimakai", "ghouls", NULL, "1988",
	"Dai Makai-Mura (Japan)\0", NULL, "Capcom", "CPS1",
	L"\u5927\u9B54\u754C\u6751\0Dai Makai-Mura\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,daimakaiRomInfo,daimakaiRomName,ghoulsInputInfo, ghoulsDIPInfo,
	ghoulsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Strider (US set 1)

static struct BurnRomInfo StriderRomDesc[] = {
	{ "strider.30", 0x020000, 0xda997474, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "strider.35", 0x020000, 0x5463aaa3, 1 | BRF_ESS | BRF_PRG },
	{ "strider.31", 0x020000, 0xd20786db, 1 | BRF_ESS | BRF_PRG },
	{ "strider.36", 0x020000, 0x21aa2863, 1 | BRF_ESS | BRF_PRG },
	{ "strider.32", 0x080000, 0x9b3cfc08, 1 | BRF_ESS | BRF_PRG },

	{ "strider.06", 0x080000, 0x4eee9aea, 2 | BRF_GRA },
	{ "strider.08", 0x080000, 0x2d7f21e4, 2 | BRF_GRA },
	{ "strider.02", 0x080000, 0x7705aa46, 2 | BRF_GRA },
	{ "strider.04", 0x080000, 0x5b18b722, 2 | BRF_GRA },
	{ "strider.05", 0x080000, 0x005f000b, 2 | BRF_GRA },
	{ "strider.07", 0x080000, 0xb9441519, 2 | BRF_GRA },
	{ "strider.01", 0x080000, 0xb7d04e8b, 2 | BRF_GRA },
	{ "strider.03", 0x080000, 0x6b4713b4, 2 | BRF_GRA },

	{ "strider.09", 0x010000, 0x2ed403bc, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "strider.18", 0x020000, 0x4386bc80, 4 | BRF_SND },
	{ "strider.19", 0x020000, 0x444536d7, 4 | BRF_SND },
};

STD_ROM_PICK(Strider); STD_ROM_FN(Strider);

struct BurnDriver BurnDrvCpsStrider = {
	"strider", NULL, NULL, "1989",
	"Strider (US set 1)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,StriderRomInfo,StriderRomName,striderInputInfo, striderDIPInfo,
	forgottnInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Strider (US set 2)

static struct BurnRomInfo StridruaRomDesc[] = {
	{ "strid.30",   0x020000, 0x66aec273, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "strid.35",   0x020000, 0x50e0e865, 1 | BRF_ESS | BRF_PRG },
	{ "strid.31",   0x020000, 0xeae93bd1, 1 | BRF_ESS | BRF_PRG },
	{ "strid.36",   0x020000, 0xb904a31d, 1 | BRF_ESS | BRF_PRG }, 
	{ "strider.32", 0x080000, 0x9b3cfc08, 1 | BRF_ESS | BRF_PRG },

	{ "strider.06", 0x080000, 0x4eee9aea, 2 | BRF_GRA },
	{ "strider.08", 0x080000, 0x2d7f21e4, 2 | BRF_GRA },
	{ "strider.02", 0x080000, 0x7705aa46, 2 | BRF_GRA },
	{ "strider.04", 0x080000, 0x5b18b722, 2 | BRF_GRA },
	{ "strider.05", 0x080000, 0x005f000b, 2 | BRF_GRA },
	{ "strider.07", 0x080000, 0xb9441519, 2 | BRF_GRA },
	{ "strider.01", 0x080000, 0xb7d04e8b, 2 | BRF_GRA },
	{ "strider.03", 0x080000, 0x6b4713b4, 2 | BRF_GRA },

	{ "strid.09",   0x010000, 0x08d63519, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "strider.18", 0x020000, 0x4386bc80, 4 | BRF_SND },
	{ "strider.19", 0x020000, 0x444536d7, 4 | BRF_SND },
};

STD_ROM_PICK(Stridrua); STD_ROM_FN(Stridrua);

struct BurnDriver BurnDrvCpsStridrua = {
	"stridrua", "strider", NULL, "1989",
	"Strider (US set 2)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,StridruaRomInfo,StridruaRomName,striderInputInfo, striderDIPInfo,
	forgottnInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Strider Hiryu (Japan set 1)

static struct BurnRomInfo StriderjRomDesc[] = {
	{ "sthj23.bin", 0x080000, 0x046e7b12, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "strider.32", 0x080000, 0x9b3cfc08, 1 | BRF_ESS | BRF_PRG },

	{ "strider.06", 0x080000, 0x4eee9aea, 2 | BRF_GRA },
	{ "strider.08", 0x080000, 0x2d7f21e4, 2 | BRF_GRA },
	{ "strider.02", 0x080000, 0x7705aa46, 2 | BRF_GRA },
	{ "strider.04", 0x080000, 0x5b18b722, 2 | BRF_GRA },
	{ "strider.05", 0x080000, 0x005f000b, 2 | BRF_GRA },
	{ "strider.07", 0x080000, 0xb9441519, 2 | BRF_GRA },
	{ "strider.01", 0x080000, 0xb7d04e8b, 2 | BRF_GRA },
	{ "strider.03", 0x080000, 0x6b4713b4, 2 | BRF_GRA },

	{ "strider.09", 0x010000, 0x2ed403bc, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "strider.18", 0x020000, 0x4386bc80, 4 | BRF_SND },
	{ "strider.19", 0x020000, 0x444536d7, 4 | BRF_SND },
};

STD_ROM_PICK(Striderj) STD_ROM_FN(Striderj)

void striderj_stars()
{
	CpsLoadStars(CpsStar, 2);
}

int striderjInit()
{
	pCpsInitCallback = striderj_stars;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsStriderj = {
	"striderj", "strider", NULL, "1989",
	"Strider Hiryu (Japan set 1)\0", NULL, "Capcom", "CPS1",
	L"\u30B9\u30C8\u30E9\u30A4\u30C0\u30FC\u98DB\u7ADC\0Strider Hiryu\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,StriderjRomInfo,StriderjRomName,striderInputInfo, striderDIPInfo,
	striderjInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Strider Hiryu (Japan set 2)

static struct BurnRomInfo StridrjaRomDesc[] = {
	{ "sth36.bin",  0x020000, 0x53c7b006, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sth42.bin",  0x020000, 0x4037f65f, 1 | BRF_ESS | BRF_PRG },
	{ "sth37.bin",  0x020000, 0x80e8877d, 1 | BRF_ESS | BRF_PRG },
	{ "sth43.bin",  0x020000, 0x6b3fa466, 1 | BRF_ESS | BRF_PRG },
	{ "strider.32", 0x080000, 0x9b3cfc08, 1 | BRF_ESS | BRF_PRG },

	{ "strider.06", 0x080000, 0x4eee9aea, 2 | BRF_GRA },
	{ "strider.08", 0x080000, 0x2d7f21e4, 2 | BRF_GRA },
	{ "strider.02", 0x080000, 0x7705aa46, 2 | BRF_GRA },
	{ "strider.04", 0x080000, 0x5b18b722, 2 | BRF_GRA },
	{ "strider.05", 0x080000, 0x005f000b, 2 | BRF_GRA },
	{ "strider.07", 0x080000, 0xb9441519, 2 | BRF_GRA },
	{ "strider.01", 0x080000, 0xb7d04e8b, 2 | BRF_GRA },
	{ "strider.03", 0x080000, 0x6b4713b4, 2 | BRF_GRA },

	{ "strider.09", 0x010000, 0x2ed403bc, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "strider.18", 0x020000, 0x4386bc80, 4 | BRF_SND },
	{ "strider.19", 0x020000, 0x444536d7, 4 | BRF_SND },
};

STD_ROM_PICK(Stridrja); STD_ROM_FN(Stridrja);

struct BurnDriver BurnDrvCpsStridrja = {
	"stridrja", "strider", NULL, "1989",
	"Strider Hiryu (Japan set 2)\0", NULL, "Capcom", "CPS1",
	L"\u30B9\u30C8\u30E9\u30A4\u30C0\u30FC\u98DB\u7ADC\0Strider Hiryu\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,StridrjaRomInfo,StridrjaRomName,striderInputInfo, striderDIPInfo,
	forgottnInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Dynasty Wars (World)

static struct BurnRomInfo DwRomDesc[] = {
	{ "30.11f",    0x020000, 0xf9ec6d68, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "35.11h",    0x020000, 0xe41fff2f, 1 | BRF_ESS | BRF_PRG },
	{ "31.12f",    0x020000, 0xe3de76ff, 1 | BRF_ESS | BRF_PRG },
	{ "36.12h",    0x020000, 0x7a13cfbf, 1 | BRF_ESS | BRF_PRG },
	{ "tkm-9.8h",  0x080000, 0x93654bcf, 1 | BRF_ESS | BRF_PRG },

	{ "tkm-5.7a",  0x080000, 0xf64bb6a0, 2 | BRF_GRA },
	{ "tkm-8.9a",  0x080000, 0x21fe6274, 2 | BRF_GRA },
	{ "tkm-6.3a",  0x080000, 0x0bf228cb, 2 | BRF_GRA },
	{ "tkm-7.5a",  0x080000, 0x1255dfb1, 2 | BRF_GRA },
	{ "tkm-1.8a",  0x080000, 0x44f7661e, 2 | BRF_GRA },
	{ "tkm-4.10a", 0x080000, 0xa54c515d, 2 | BRF_GRA },
	{ "tkm-2.4a",  0x080000, 0xca5c687c, 2 | BRF_GRA },
	{ "tkm-3.6a",  0x080000, 0xf9fe6591, 2 | BRF_GRA },

	{ "tke17.12b", 0x010000, 0xb3b79d4f, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "tke18.11c", 0x020000, 0xac6e307d, 4 | BRF_SND },
	{ "tke19.12c", 0x020000, 0x068741db, 4 | BRF_SND },
};

STD_ROM_PICK(Dw) STD_ROM_FN(Dw)

struct BurnDriver BurnDrvCpsDw = {
	"dynwar", NULL, NULL, "1989",
	"Dynasty Wars (World)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,DwRomInfo,DwRomName,dwInputInfo, dwDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Tenchi wo Kurau (Japan)

static struct BurnRomInfo DwjRomDesc[] = {
	{ "36.bin", 0x020000, 0x1a516657, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "42.bin", 0x020000, 0x12a290a0, 1 | BRF_ESS | BRF_PRG },
	{ "37.bin", 0x020000, 0x932fc943, 1 | BRF_ESS | BRF_PRG },
	{ "43.bin", 0x020000, 0x872ad76d, 1 | BRF_ESS | BRF_PRG },
	{ "34.bin", 0x020000, 0x8f663d00, 1 | BRF_ESS | BRF_PRG },
	{ "40.bin", 0x020000, 0x1586dbf3, 1 | BRF_ESS | BRF_PRG },
	{ "35.bin", 0x020000, 0x9db93d7a, 1 | BRF_ESS | BRF_PRG },
	{ "41.bin", 0x020000, 0x1aae69a4, 1 | BRF_ESS | BRF_PRG },

	{ "09.bin", 0x020000, 0xc3e83c69, 2 | BRF_GRA },
	{ "01.bin", 0x020000, 0x187b2886, 2 | BRF_GRA },
	{ "13.bin", 0x020000, 0x0273d87d, 2 | BRF_GRA },
	{ "05.bin", 0x020000, 0x339378b8, 2 | BRF_GRA },
	{ "24.bin", 0x020000, 0xc6909b6f, 2 | BRF_GRA },
	{ "17.bin", 0x020000, 0x2e2f8320, 2 | BRF_GRA },
	{ "38.bin", 0x020000, 0xcd7923ed, 2 | BRF_GRA },
	{ "32.bin", 0x020000, 0x21a0a453, 2 | BRF_GRA },
	{ "10.bin", 0x020000, 0xff28f8d0, 2 | BRF_GRA },
	{ "02.bin", 0x020000, 0xcc83c02f, 2 | BRF_GRA },
	{ "14.bin", 0x020000, 0x18fb232c, 2 | BRF_GRA },
	{ "06.bin", 0x020000, 0x6f9edd75, 2 | BRF_GRA },
	{ "25.bin", 0x020000, 0x152ea74a, 2 | BRF_GRA },
	{ "18.bin", 0x020000, 0x1833f932, 2 | BRF_GRA },
	{ "39.bin", 0x020000, 0xbc09b360, 2 | BRF_GRA },
	{ "33.bin", 0x020000, 0x89de1533, 2 | BRF_GRA },
	{ "11.bin", 0x020000, 0x29eaf490, 2 | BRF_GRA },
	{ "03.bin", 0x020000, 0x7bf51337, 2 | BRF_GRA },
	{ "15.bin", 0x020000, 0xd36cdb91, 2 | BRF_GRA },
	{ "07.bin", 0x020000, 0xe04af054, 2 | BRF_GRA },
	{ "26.bin", 0x020000, 0x07fc714b, 2 | BRF_GRA },
	{ "19.bin", 0x020000, 0x7114e5c6, 2 | BRF_GRA },
	{ "28.bin", 0x020000, 0xaf62bf07, 2 | BRF_GRA },
	{ "21.bin", 0x020000, 0x523f462a, 2 | BRF_GRA },
	{ "12.bin", 0x020000, 0x38652339, 2 | BRF_GRA },
	{ "04.bin", 0x020000, 0x4951bc0f, 2 | BRF_GRA },
	{ "16.bin", 0x020000, 0x381608ae, 2 | BRF_GRA },
	{ "08.bin", 0x020000, 0xb475d4e9, 2 | BRF_GRA },
	{ "27.bin", 0x020000, 0xa27e81fa, 2 | BRF_GRA },
	{ "20.bin", 0x020000, 0x002796dc, 2 | BRF_GRA },
	{ "29.bin", 0x020000, 0x6b41f82d, 2 | BRF_GRA },
	{ "22.bin", 0x020000, 0x52145369, 2 | BRF_GRA },

	{ "23.bin", 0x010000, 0xb3b79d4f, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "30.bin", 0x020000, 0x7e5f6cb4, 4 | BRF_SND },
	{ "31.bin", 0x020000, 0x4a30c737, 4 | BRF_SND },
};

STD_ROM_PICK(Dwj) STD_ROM_FN(Dwj)

struct BurnDriver BurnDrvCpsDwj = {
	"dynwarj", "dynwar", NULL, "1989",
	"Tenchi wo Kurau (Japan)\0", NULL, "Capcom", "CPS1",
	L"\u8EE2\u5730\u3092\u55B0\u3089\u3046\0Tenchi wo Kurau\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,DwjRomInfo,DwjRomName,dwInputInfo, dwDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Willow (US)

static struct BurnRomInfo WillowRomDesc[] = {
	{ "wlu_30.rom",  0x020000, 0xd604dbb1, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "willow-u.35", 0x020000, 0x7a791e77, 1 | BRF_ESS | BRF_PRG },
	{ "wlu_31.rom",  0x020000, 0x0eb48a83, 1 | BRF_ESS | BRF_PRG },
	{ "wlu_36.rom",  0x020000, 0x36100209, 1 | BRF_ESS | BRF_PRG },
	{ "wl_32.rom",   0x080000, 0xdfd9f643, 1 | BRF_ESS | BRF_PRG },

	{ "wl_gfx5.rom", 0x080000, 0xafa74b73, 2 | BRF_GRA },
	{ "wl_gfx7.rom", 0x080000, 0x12a0dc0b, 2 | BRF_GRA },
	{ "wl_gfx1.rom", 0x080000, 0xc6f2abce, 2 | BRF_GRA },
	{ "wl_gfx3.rom", 0x080000, 0x4aa4c6d3, 2 | BRF_GRA },
	{ "wl_24.rom",   0x020000, 0x6f0adee5, 2 | BRF_GRA },
	{ "wl_14.rom",   0x020000, 0x9cf3027d, 2 | BRF_GRA },
	{ "wl_26.rom",   0x020000, 0xf09c8ecf, 2 | BRF_GRA },
	{ "wl_16.rom",   0x020000, 0xe35407aa, 2 | BRF_GRA },
	{ "wl_20.rom",   0x020000, 0x84992350, 2 | BRF_GRA },
	{ "wl_10.rom",   0x020000, 0xb87b5a36, 2 | BRF_GRA },
	{ "wl_22.rom",   0x020000, 0xfd3f89f0, 2 | BRF_GRA },
	{ "wl_12.rom",   0x020000, 0x7da49d69, 2 | BRF_GRA },

	{ "wl_09.rom",   0x010000, 0xf6b3d060, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "wl_18.rom",   0x020000, 0xbde23d4d, 4 | BRF_SND },
	{ "wl_19.rom",   0x020000, 0x683898f5, 4 | BRF_SND },
};

STD_ROM_PICK(Willow); STD_ROM_FN(Willow);

struct BurnDriver BurnDrvCpsWillow = {
	"willow", NULL, NULL, "1989",
	"Willow (US)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,WillowRomInfo,WillowRomName,willowInputInfo, willowDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Willow (Japan, Japanese)

static struct BurnRomInfo WillowjRomDesc[] = {
	{ "wl36.bin",    0x020000, 0x2b0d7cbc, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "wl42.bin",    0x020000, 0x1ac39615, 1 | BRF_ESS | BRF_PRG },
	{ "wl37.bin",    0x020000, 0x30a717fa, 1 | BRF_ESS | BRF_PRG },
	{ "wl43.bin",    0x020000, 0xd0dddc9e, 1 | BRF_ESS | BRF_PRG },
	{ "wl_32.rom",   0x080000, 0xdfd9f643, 1 | BRF_ESS | BRF_PRG }, 

	{ "wl_gfx5.rom", 0x080000, 0xafa74b73, 2 | BRF_GRA },
	{ "wl_gfx7.rom", 0x080000, 0x12a0dc0b, 2 | BRF_GRA },
	{ "wl_gfx1.rom", 0x080000, 0xc6f2abce, 2 | BRF_GRA },
	{ "wl_gfx3.rom", 0x080000, 0x4aa4c6d3, 2 | BRF_GRA },
	{ "wl_24.rom",   0x020000, 0x6f0adee5, 2 | BRF_GRA },
	{ "wl_14.rom",   0x020000, 0x9cf3027d, 2 | BRF_GRA },
	{ "wl_26.rom",   0x020000, 0xf09c8ecf, 2 | BRF_GRA },
	{ "wl_16.rom",   0x020000, 0xe35407aa, 2 | BRF_GRA },
	{ "wl_20.rom",   0x020000, 0x84992350, 2 | BRF_GRA },
	{ "wl_10.rom",   0x020000, 0xb87b5a36, 2 | BRF_GRA },
	{ "wl_22.rom",   0x020000, 0xfd3f89f0, 2 | BRF_GRA },
	{ "wl_12.rom",   0x020000, 0x7da49d69, 2 | BRF_GRA },

	{ "wl_09.rom",   0x010000, 0xf6b3d060, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "wl_18.rom",   0x020000, 0xbde23d4d, 4 | BRF_SND },
	{ "wl_19.rom",   0x020000, 0x683898f5, 4 | BRF_SND },
};

STD_ROM_PICK(Willowj); STD_ROM_FN(Willowj);

struct BurnDriver BurnDrvCpsWillowj = {
	"willowj", "willow", NULL, "1989",
	"Willow (Japan, Japanese)\0", "Japanese language", "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,WillowjRomInfo,WillowjRomName,willowInputInfo, willowDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Willow (Japan, English)

static struct BurnRomInfo WillowjeRomDesc[] = {
	{ "wlu_30.rom",  0x020000, 0xd604dbb1, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "wlu_35.rom",  0x020000, 0xdaee72fe, 1 | BRF_ESS | BRF_PRG },
	{ "wlu_31.rom",  0x020000, 0x0eb48a83, 1 | BRF_ESS | BRF_PRG },
	{ "wlu_36.rom",  0x020000, 0x36100209, 1 | BRF_ESS | BRF_PRG },
	{ "wl_32.rom",   0x080000, 0xdfd9f643, 1 | BRF_ESS | BRF_PRG },

	{ "wl_gfx5.rom", 0x080000, 0xafa74b73, 2 | BRF_GRA },
	{ "wl_gfx7.rom", 0x080000, 0x12a0dc0b, 2 | BRF_GRA },
	{ "wl_gfx1.rom", 0x080000, 0xc6f2abce, 2 | BRF_GRA },
	{ "wl_gfx3.rom", 0x080000, 0x4aa4c6d3, 2 | BRF_GRA },
	{ "wl_24.rom",   0x020000, 0x6f0adee5, 2 | BRF_GRA },
	{ "wl_14.rom",   0x020000, 0x9cf3027d, 2 | BRF_GRA },
	{ "wl_26.rom",   0x020000, 0xf09c8ecf, 2 | BRF_GRA },
	{ "wl_16.rom",   0x020000, 0xe35407aa, 2 | BRF_GRA },
	{ "wl_20.rom",   0x020000, 0x84992350, 2 | BRF_GRA },
	{ "wl_10.rom",   0x020000, 0xb87b5a36, 2 | BRF_GRA },
	{ "wl_22.rom",   0x020000, 0xfd3f89f0, 2 | BRF_GRA },
	{ "wl_12.rom",   0x020000, 0x7da49d69, 2 | BRF_GRA },

	{ "wl_09.rom",   0x010000, 0xf6b3d060, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "wl_18.rom",   0x020000, 0xbde23d4d, 4 | BRF_SND },
	{ "wl_19.rom",   0x020000, 0x683898f5, 4 | BRF_SND },
};

STD_ROM_PICK(Willowje); STD_ROM_FN(Willowje);

struct BurnDriver BurnDrvCpsWillowje = {
	"willowje", "willow", NULL, "1989",
	"Willow (Japan, English)\0", "English language", "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,WillowjeRomInfo,WillowjeRomName,willowInputInfo, willowDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// U.N. Squadron (US)

static struct BurnRomInfo UnsquadRomDesc[] = {
	{ "unsquad.30", 0x020000, 0x24d8f88d, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "unsquad.35", 0x020000, 0x8b954b59, 1 | BRF_ESS | BRF_PRG },
	{ "unsquad.31", 0x020000, 0x33e9694b, 1 | BRF_ESS | BRF_PRG },
	{ "unsquad.36", 0x020000, 0x7cc8fb9e, 1 | BRF_ESS | BRF_PRG },
	{ "unsquad.32", 0x080000, 0xae1d7fb0, 1 | BRF_ESS | BRF_PRG },

	{ "unsquad.05", 0x080000, 0xbf4575d8, 2 | BRF_GRA },
	{ "unsquad.07", 0x080000, 0xa02945f4, 2 | BRF_GRA },
	{ "unsquad.01", 0x080000, 0x5965ca8d, 2 | BRF_GRA },
	{ "unsquad.03", 0x080000, 0xac6db17d, 2 | BRF_GRA },

	{ "unsquad.09", 0x010000, 0xf3dd1367, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "unsquad.18", 0x020000, 0x584b43a9, 4 | BRF_SND },
};

STD_ROM_PICK(Unsquad); STD_ROM_FN(Unsquad);

struct BurnDriver BurnDrvCpsUnsquad = {
	"unsquad", NULL, NULL, "1989",
	"U.N. Squadron (US)\0", NULL, "Daipro / Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,UnsquadRomInfo,UnsquadRomName,unsquadInputInfo, unsquadDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Area 88 (Japan)

static struct BurnRomInfo Area88RomDesc[] = {
	{ "ar36.bin",   0x020000, 0x65030392, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "ar42.bin",   0x020000, 0xc48170de, 1 | BRF_ESS | BRF_PRG },
	{ "unsquad.31", 0x020000, 0x33e9694b, 1 | BRF_ESS | BRF_PRG },
	{ "unsquad.36", 0x020000, 0x7cc8fb9e, 1 | BRF_ESS | BRF_PRG },
	{ "unsquad.32", 0x080000, 0xae1d7fb0, 1 | BRF_ESS | BRF_PRG },

	{ "unsquad.05", 0x080000, 0xbf4575d8, 2 | BRF_GRA },
	{ "unsquad.07", 0x080000, 0xa02945f4, 2 | BRF_GRA },
	{ "unsquad.01", 0x080000, 0x5965ca8d, 2 | BRF_GRA },
	{ "unsquad.03", 0x080000, 0xac6db17d, 2 | BRF_GRA },

	{ "unsquad.09", 0x010000, 0xf3dd1367, 3 | BRF_ESS | BRF_PRG }, // z80 program
 
	{ "unsquad.18", 0x020000, 0x584b43a9, 4 | BRF_SND },
};

STD_ROM_PICK(Area88); STD_ROM_FN(Area88);

struct BurnDriver BurnDrvCpsArea88 = {
	"area88", "unsquad", NULL, "1989",
	"Area 88 (Japan)\0", NULL, "Daipro / Capcom", "CPS1",
	L"\u30A8\u30EA\u30A2\uFF18\uFF18\0Area 88\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Area88RomInfo,Area88RomName,unsquadInputInfo, unsquadDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Final Fight (World)

static struct BurnRomInfo FfightRomDesc[] = {
	{ "ff30-36.bin",  0x020000, 0xf9a5ce83, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "ff35-42.bin",  0x020000, 0x65f11215, 1 | BRF_ESS | BRF_PRG },
	{ "ff31-37.bin",  0x020000, 0xe1033784, 1 | BRF_ESS | BRF_PRG },
	{ "ff36-43.bin",  0x020000, 0x995e968a, 1 | BRF_ESS | BRF_PRG },
	{ "ff32-32m.bin", 0x080000, 0xc747696e, 1 | BRF_ESS | BRF_PRG },

	{ "ff05-05m.bin", 0x080000, 0x9c284108, 2 | BRF_GRA },
	{ "ff07-07m.bin", 0x080000, 0xa7584dfb, 2 | BRF_GRA },
	{ "ff01-01m.bin", 0x080000, 0x0b605e44, 2 | BRF_GRA },
	{ "ff03-03m.bin", 0x080000, 0x52291cd2, 2 | BRF_GRA },

	{ "ff09-09.bin",  0x010000, 0xb8367eb5, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ff18-18.bin",  0x020000, 0x375c66e7, 4 | BRF_SND },
	{ "ff19-19.bin",  0x020000, 0x1ef137f9, 4 | BRF_SND },
};

STD_ROM_PICK(Ffight) STD_ROM_FN(Ffight)

struct BurnDriver BurnDrvCpsFfight = {
	"ffight", NULL, NULL, "1989",
	"Final Fight (World)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,FfightRomInfo,FfightRomName,ffightInputInfo, ffightDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Final Fight (US)

static struct BurnRomInfo FfightuRomDesc[] = {
	{ "ff30-36.bin",   0x020000, 0xe2a48af9, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "ff35-42.bin",   0x020000, 0xf4bb480e, 1 | BRF_ESS | BRF_PRG },
	{ "ff31-37.bin",   0x020000, 0xc371c667, 1 | BRF_ESS | BRF_PRG },
	{ "ff43.rom",      0x020000, 0x2f5771f9, 1 | BRF_ESS | BRF_PRG },
	{ "ff32-32m.bin",  0x080000, 0xc747696e, 1 | BRF_ESS | BRF_PRG },

	{ "ff05-05m.bin",  0x080000, 0x9c284108, 2 | BRF_GRA },
	{ "ff07-07m.bin",  0x080000, 0xa7584dfb, 2 | BRF_GRA },
	{ "ff01-01m.bin",  0x080000, 0x0b605e44, 2 | BRF_GRA },
	{ "ff03-03m.bin",  0x080000, 0x52291cd2, 2 | BRF_GRA },

	{ "ff09-09.bin",   0x010000, 0xb8367eb5, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ff18-18.bin",   0x020000, 0x375c66e7, 4 | BRF_SND },
	{ "ff19-19.bin",   0x020000, 0x1ef137f9, 4 | BRF_SND },
};

STD_ROM_PICK(Ffightu) STD_ROM_FN(Ffightu)

struct BurnDriver BurnDrvCpsFfightu = {
	"ffightu", "ffight", NULL, "1989",
	"Final Fight (US)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,FfightuRomInfo,FfightuRomName,ffightInputInfo, ffightDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Final Fight (US 900112)

static struct BurnRomInfo FfightuaRomDesc[] = {
	{ "36",            0x020000, 0xe2a48af9, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "42",            0x020000, 0xf4bb480e, 1 | BRF_ESS | BRF_PRG },
	{ "37",            0x020000, 0xc371c667, 1 | BRF_ESS | BRF_PRG },
	{ "43",            0x020000, 0x2f5771f9, 1 | BRF_ESS | BRF_PRG },
	{ "ff32-32m.bin",  0x080000, 0xc747696e, 1 | BRF_ESS | BRF_PRG },

	{ "ff05-05m.bin",  0x080000, 0x9c284108, 2 | BRF_GRA },
	{ "ff07-07m.bin",  0x080000, 0xa7584dfb, 2 | BRF_GRA },
	{ "ff01-01m.bin",  0x080000, 0x0b605e44, 2 | BRF_GRA },
	{ "ff03-03m.bin",  0x080000, 0x52291cd2, 2 | BRF_GRA },

	{ "ff09-09.bin",   0x010000, 0xb8367eb5, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ff18-18.bin",   0x020000, 0x375c66e7, 4 | BRF_SND },
	{ "ff19-19.bin",   0x020000, 0x1ef137f9, 4 | BRF_SND },
};

STD_ROM_PICK(Ffightua) STD_ROM_FN(Ffightua)

struct BurnDriver BurnDrvCpsFfightua = {
	"ffightua", "ffight", NULL, "1989",
	"Final Fight (US 900112)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,FfightuaRomInfo,FfightuaRomName,ffightInputInfo, ffightDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Final Fight (US 900613)

static struct BurnRomInfo FfightubRomDesc[] = {
	{ "30_2da0.010",  0x020000, 0xed988977, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "35_46ff.010",  0x020000, 0x07bf1c21, 1 | BRF_ESS | BRF_PRG },
	{ "31_c933.010",  0x020000, 0xdba5a476, 1 | BRF_ESS | BRF_PRG },
	{ "36_0ace.010",  0x020000, 0x4d89f542, 1 | BRF_ESS | BRF_PRG },
	{ "ff32-32m.bin", 0x080000, 0xc747696e, 1 | BRF_ESS | BRF_PRG },

	{ "ff05-05m.bin", 0x080000, 0x9c284108, 2 | BRF_GRA },
	{ "ff07-07m.bin", 0x080000, 0xa7584dfb, 2 | BRF_GRA },
	{ "ff01-01m.bin", 0x080000, 0x0b605e44, 2 | BRF_GRA },
	{ "ff03-03m.bin", 0x080000, 0x52291cd2, 2 | BRF_GRA },

	{ "23_ceb1.512",  0x010000, 0xb8367eb5, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "18_ccb5.010",  0x020000, 0x375c66e7, 4 | BRF_SND },
	{ "19_476d.010",  0x020000, 0x1ef137f9, 4 | BRF_SND },
};

STD_ROM_PICK(Ffightub) STD_ROM_FN(Ffightub)

struct BurnDriver BurnDrvCpsFfightub = {
	"ffightub", "ffight",NULL,"1989",
	"Final Fight (US 900613)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,FfightubRomInfo,FfightubRomName,ffightInputInfo, ffightDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Final Fight (Japan)

static struct BurnRomInfo FfightjRomDesc[] = {
	{ "ff30-36.bin",  0x020000, 0xf9a5ce83, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "ff35-42.bin",  0x020000, 0x65f11215, 1 | BRF_ESS | BRF_PRG },
	{ "ff31-37.bin",  0x020000, 0xe1033784, 1 | BRF_ESS | BRF_PRG },
	{ "ff43.bin",     0x020000, 0xb6dee1c3, 1 | BRF_ESS | BRF_PRG },
	{ "ff32-32m.bin", 0x080000, 0xc747696e, 1 | BRF_ESS | BRF_PRG }, 

	{ "ff09.bin",     0x020000, 0x5b116d0d, 2 | BRF_GRA },
	{ "ff01.bin",     0x020000, 0x815b1797, 2 | BRF_GRA },
	{ "ff13.bin",     0x020000, 0x8721a7da, 2 | BRF_GRA },
	{ "ff05.bin",     0x020000, 0xd0fcd4b5, 2 | BRF_GRA },
	{ "ff24.bin",     0x020000, 0xa1ab607a, 2 | BRF_GRA },
	{ "ff17.bin",     0x020000, 0x2dc18cf4, 2 | BRF_GRA },
	{ "ff38.bin",     0x020000, 0x6535a57f, 2 | BRF_GRA },
	{ "ff32.bin",     0x020000, 0xc8bc4a57, 2 | BRF_GRA },
	{ "ff10.bin",     0x020000, 0x624a924a, 2 | BRF_GRA },
	{ "ff02.bin",     0x020000, 0x5d91f694, 2 | BRF_GRA },
	{ "ff14.bin",     0x020000, 0x0a2e9101, 2 | BRF_GRA },
	{ "ff06.bin",     0x020000, 0x1c18f042, 2 | BRF_GRA },
	{ "ff25.bin",     0x020000, 0x6e8181ea, 2 | BRF_GRA },
	{ "ff18.bin",     0x020000, 0xb19ede59, 2 | BRF_GRA },
	{ "ff39.bin",     0x020000, 0x9416b477, 2 | BRF_GRA },
	{ "ff33.bin",     0x020000, 0x7369fa07, 2 | BRF_GRA },

	{ "ff09-09.bin",  0x010000, 0xb8367eb5, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ff18-18.bin",  0x020000, 0x375c66e7, 4 | BRF_SND },
	{ "ff19-19.bin",  0x020000, 0x1ef137f9, 4 | BRF_SND },
};

STD_ROM_PICK(Ffightj) STD_ROM_FN(Ffightj)

struct BurnDriver BurnDrvCpsFfightj = {
	"ffightj", "ffight", NULL, "1989",
	"Final Fight (Japan)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,FfightjRomInfo,FfightjRomName,ffightInputInfo, ffightDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Final Fight (Japan 900305)

static struct BurnRomInfo Ffightj1RomDesc[] = {
	{ "ff30-36.rom",  0x020000, 0x088ed1c9, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "ff35-42.rom",  0x020000, 0xc4c491e6, 1 | BRF_ESS | BRF_PRG },
	{ "ff31-37.rom",  0x020000, 0x708557ff, 1 | BRF_ESS | BRF_PRG },
	{ "ff36-43.rom",  0x020000, 0xc004004a, 1 | BRF_ESS | BRF_PRG },
	{ "ff32-32m.bin", 0x080000, 0xc747696e, 1 | BRF_ESS | BRF_PRG },

	{ "ff09.bin",     0x020000, 0x5b116d0d, 2 | BRF_GRA },
	{ "ff01.bin",     0x020000, 0x815b1797, 2 | BRF_GRA },
	{ "ff13.bin",     0x020000, 0x8721a7da, 2 | BRF_GRA },
	{ "ff05.bin",     0x020000, 0xd0fcd4b5, 2 | BRF_GRA },
	{ "ff24.bin",     0x020000, 0xa1ab607a, 2 | BRF_GRA },
	{ "ff17.bin",     0x020000, 0x2dc18cf4, 2 | BRF_GRA },
	{ "ff38.bin",     0x020000, 0x6535a57f, 2 | BRF_GRA },
	{ "ff32.bin",     0x020000, 0xc8bc4a57, 2 | BRF_GRA },
	{ "ff10.bin",     0x020000, 0x624a924a, 2 | BRF_GRA },
	{ "ff02.bin",     0x020000, 0x5d91f694, 2 | BRF_GRA },
	{ "ff14.bin",     0x020000, 0x0a2e9101, 2 | BRF_GRA },
	{ "ff06.bin",     0x020000, 0x1c18f042, 2 | BRF_GRA },
	{ "ff25.bin",     0x020000, 0x6e8181ea, 2 | BRF_GRA },
	{ "ff18.bin",     0x020000, 0xb19ede59, 2 | BRF_GRA },
	{ "ff39.bin",     0x020000, 0x9416b477, 2 | BRF_GRA },
	{ "ff33.bin",     0x020000, 0x7369fa07, 2 | BRF_GRA },

	{ "ff09-09.bin",  0x010000, 0xb8367eb5, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ff18-18.bin",  0x020000, 0x375c66e7, 4 | BRF_SND },
	{ "ff19-19.bin",  0x020000, 0x1ef137f9, 4 | BRF_SND },
};

STD_ROM_PICK(Ffightj1) STD_ROM_FN(Ffightj1)

struct BurnDriver BurnDrvCpsFfightj1 = {
	"ffightj1", "ffight", NULL, "1989",
	"Final Fight (Japan 900305)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Ffightj1RomInfo,Ffightj1RomName,ffightInputInfo, ffightDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// On CPS-1-based hardware (bootleg clone)

// Final Crash (World, bootleg) 

static struct BurnRomInfo fcrashRomDesc[] = {
	{ "9.bin",  0x020000, 0xc6854c91, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "5.bin",  0x020000, 0x77f7c2b3, 1 | BRF_ESS | BRF_PRG },
	{ "8.bin",  0x020000, 0x1895b3df, 1 | BRF_ESS | BRF_PRG },
	{ "4.bin",  0x020000, 0xbbd411ee, 1 | BRF_ESS | BRF_PRG },
	{ "7.bin",  0x020000, 0x5b23ebf2, 1 | BRF_ESS | BRF_PRG },
	{ "3.bin",  0x020000, 0xaba2aebe, 1 | BRF_ESS | BRF_PRG },
	{ "6.bin",  0x020000, 0xd4bf37f6, 1 | BRF_ESS | BRF_PRG },
	{ "2.bin",  0x020000, 0x07ac8f43, 1 | BRF_ESS | BRF_PRG },

	{ "18.bin", 0x020000, 0xf1eee6d9, 2 | BRF_GRA },
	{ "20.bin", 0x020000, 0x675f4537, 2 | BRF_GRA },
	{ "22.bin", 0x020000, 0xdb8a32ac, 2 | BRF_GRA },
	{ "24.bin", 0x020000, 0xf4113e57, 2 | BRF_GRA },
	{ "10.bin", 0x020000, 0xd478853e, 2 | BRF_GRA },
	{ "12.bin", 0x020000, 0x25055642, 2 | BRF_GRA },
	{ "14.bin", 0x020000, 0xb77d0328, 2 | BRF_GRA },
	{ "16.bin", 0x020000, 0xea111a79, 2 | BRF_GRA },
	{ "19.bin", 0x020000, 0xb3aa1f48, 2 | BRF_GRA },
	{ "21.bin", 0x020000, 0x04d175c9, 2 | BRF_GRA },
	{ "23.bin", 0x020000, 0xe592ba4f, 2 | BRF_GRA },
	{ "25.bin", 0x020000, 0xb89a740f, 2 | BRF_GRA },
	{ "11.bin", 0x020000, 0xd4457a60, 2 | BRF_GRA },
	{ "13.bin", 0x020000, 0x3b26a37d, 2 | BRF_GRA },
	{ "15.bin", 0x020000, 0x6d837e09, 2 | BRF_GRA },
	{ "17.bin", 0x020000, 0xc59a4d6c, 2 | BRF_GRA },

	{ "1.bin",  0x020000, 0x5b276c14, 3 | BRF_ESS | BRF_PRG }, // Audio CPU + Sample Data
};

STD_ROM_PICK(fcrash) STD_ROM_FN(fcrash)

void fcrash_load()
{
	unsigned char *tmp = (unsigned char*)malloc(0x020000);
	for (int i = 0; i < 16; i++) {
		BurnLoadRom(tmp, 8 + i, 1);

		for (int j = 0; j < 0x020000; j++) {
			CpsGfx[(i & ~3) * 0x020000 + (j * 4) + (i & 3)] = tmp[j];
		}
	}
	free (tmp);

	unsigned int *mem32 = (unsigned int*)CpsGfx;
	for (int i = 0; i < 0x200000 / 4; i++)
	{
		mem32[i] = BITSWAP32(~mem32[i], 31, 23, 15, 7, 
						30, 22, 14, 6,
						29, 21, 13, 5, 
						28, 20, 12, 4,
						27, 19, 11, 3, 
						26, 18, 10, 2,
						25, 17,  9, 1, 
						24, 16,  8, 0);
	}
}

int fcrashInit()
{
	pCpsInitCallback = fcrash_load;

	return CpsInit();
}

struct BurnDriverX BurnDrvCpsfcrash = {
	"fcrash", NULL, NULL, "1989",
	"Final Crash (World, bootleg)\0", NULL, "Playmark [Capcom]", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_BOOTLEG,2,HARDWARE_CAPCOM_CPS1,
	NULL,fcrashRomInfo,fcrashRomName,ffightInputInfo, ffightDIPInfo,
	fcrashInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// 1941 - Counter Attack (World)

static struct BurnRomInfo NTFORomDesc[] = {
	{ "41e_30.rom",  0x020000, 0x9deb1e75, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "41e_35.rom",  0x020000, 0xd63942b3, 1 | BRF_ESS | BRF_PRG },
	{ "41e_31.rom",  0x020000, 0xdf201112, 1 | BRF_ESS | BRF_PRG },
	{ "41e_36.rom",  0x020000, 0x816a818f, 1 | BRF_ESS | BRF_PRG },
	{ "41_32.rom",   0x080000, 0x4e9648ca, 1 | BRF_ESS | BRF_PRG },

	{ "41_gfx5.rom", 0x080000, 0x01d1cb11, 2 | BRF_GRA },
	{ "41_gfx7.rom", 0x080000, 0xaeaa3509, 2 | BRF_GRA },
	{ "41_gfx1.rom", 0x080000, 0xff77985a, 2 | BRF_GRA },
	{ "41_gfx3.rom", 0x080000, 0x983be58f, 2 | BRF_GRA },

	{ "41_09.rom",   0x010000, 0x0f9d8527, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "41_18.rom",   0x020000, 0xd1f15aeb, 4 | BRF_SND },
	{ "41_19.rom",   0x020000, 0x15aec3a6, 4 | BRF_SND },
};

STD_ROM_PICK(NTFO) STD_ROM_FN(NTFO)

struct BurnDriver BurnDrvCps1941 = {
	"1941", NULL, NULL, "1990",
	"1941 - Counter Attack (World)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_ORIENTATION_VERTICAL, 2, HARDWARE_CAPCOM_CPS1,
	NULL,NTFORomInfo,NTFORomName,NTFOInputInfo, NTFODIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// 1941 - Counter Attack (Japan)

static struct BurnRomInfo NTFOJRomDesc[] = {
	{ "4136.bin",    0x020000, 0x7fbd42ab, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "4142.bin",    0x020000, 0xc7781f89, 1 | BRF_ESS | BRF_PRG },
	{ "4137.bin",    0x020000, 0xc6464b0b, 1 | BRF_ESS | BRF_PRG },
	{ "4143.bin",    0x020000, 0x440fc0b5, 1 | BRF_ESS | BRF_PRG },
	{ "41_32.rom",   0x080000, 0x4e9648ca, 1 | BRF_ESS | BRF_PRG },

	{ "41_gfx5.rom", 0x080000, 0x01d1cb11, 2 | BRF_GRA },
	{ "41_gfx7.rom", 0x080000, 0xaeaa3509, 2 | BRF_GRA },
	{ "41_gfx1.rom", 0x080000, 0xff77985a, 2 | BRF_GRA },
	{ "41_gfx3.rom", 0x080000, 0x983be58f, 2 | BRF_GRA },

	{ "41_09.rom",   0x010000, 0x0f9d8527, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "41_18.rom",   0x020000, 0xd1f15aeb, 4 | BRF_SND },
	{ "41_19.rom",   0x020000, 0x15aec3a6, 4 | BRF_SND },
};

STD_ROM_PICK(NTFOJ) STD_ROM_FN(NTFOJ)

struct BurnDriver BurnDrvCps1941j = {
	"1941j", "1941", NULL, "1990",
	"1941 - Counter Attack (Japan)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL, 2, HARDWARE_CAPCOM_CPS1,
	NULL,NTFOJRomInfo,NTFOJRomName,NTFOInputInfo, NTFODIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// Mercs (World 900302)

static struct BurnRomInfo MercsRomDesc[] = {
	{ "so2_30e.rom",  0x020000, 0xe17f9bf7, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "so2_35e.rom",  0x020000, 0x78e63575, 1 | BRF_ESS | BRF_PRG },
	{ "so2_31e.rom",  0x020000, 0x51204d36, 1 | BRF_ESS | BRF_PRG },
	{ "so2_36e.rom",  0x020000, 0x9cfba8b4, 1 | BRF_ESS | BRF_PRG },
	{ "so2_32.rom",   0x080000, 0x2eb5cf0c, 1 | BRF_ESS | BRF_PRG },

	{ "so2_gfx6.rom", 0x080000, 0xaa6102af, 2 |  BRF_GRA },
	{ "so2_gfx8.rom", 0x080000, 0x839e6869, 2 |  BRF_GRA },
	{ "so2_gfx2.rom", 0x080000, 0x597c2875, 2 |  BRF_GRA },
	{ "so2_gfx4.rom", 0x080000, 0x912a9ca0, 2 |  BRF_GRA },
	{ "so2_24.rom",   0x020000, 0x3f254efe, 2 |  BRF_GRA },
	{ "so2_14.rom",   0x020000, 0xf5a8905e, 2 |  BRF_GRA },
	{ "so2_26.rom",   0x020000, 0xf3aa5a4a, 2 |  BRF_GRA },
	{ "so2_16.rom",   0x020000, 0xb43cd1a8, 2 |  BRF_GRA },
	{ "so2_20.rom",   0x020000, 0x8ca751a3, 2 |  BRF_GRA },
	{ "so2_10.rom",   0x020000, 0xe9f569fd, 2 |  BRF_GRA },
	{ "so2_22.rom",   0x020000, 0xfce9a377, 2 |  BRF_GRA },
	{ "so2_12.rom",   0x020000, 0xb7df8a06, 2 |  BRF_GRA },

	{ "so2_09.rom",   0x010000, 0xd09d7c7a, 3 |  BRF_ESS | BRF_PRG },

	{ "so2_18.rom",   0x020000, 0xbbea1643, 4 |  BRF_SND },
	{ "so2_19.rom",   0x020000, 0xac58aa71, 4 |  BRF_SND },
};

STD_ROM_PICK(Mercs) STD_ROM_FN(Mercs)

struct BurnDriver BurnDrvCpsMercs = {
	"mercs", NULL, NULL, "1990",
	"Mercs (World 900302)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_ORIENTATION_VERTICAL, 3, HARDWARE_CAPCOM_CPS1,
	NULL,MercsRomInfo,MercsRomName,mercsInputInfo, mercsDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// Mercs (US 900302)

static struct BurnRomInfo MercsuRomDesc[] = {
	{ "so2_30e.rom",  0x020000, 0xe17f9bf7, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "s02-35",       0x020000, 0x4477df61, 1 | BRF_ESS | BRF_PRG },
	{ "so2_31e.rom",  0x020000, 0x51204d36, 1 | BRF_ESS | BRF_PRG },
	{ "so2_36e.rom",  0x020000, 0x9cfba8b4, 1 | BRF_ESS | BRF_PRG }, 
	{ "so2_32.rom",   0x080000, 0x2eb5cf0c, 1 | BRF_ESS | BRF_PRG },

	{ "so2_gfx6.rom", 0x080000, 0xaa6102af, 2 |  BRF_GRA },
	{ "so2_gfx8.rom", 0x080000, 0x839e6869, 2 |  BRF_GRA },
	{ "so2_gfx2.rom", 0x080000, 0x597c2875, 2 |  BRF_GRA },
	{ "so2_gfx4.rom", 0x080000, 0x912a9ca0, 2 |  BRF_GRA },
	{ "so2_24.rom",   0x020000, 0x3f254efe, 2 |  BRF_GRA },
	{ "so2_14.rom",   0x020000, 0xf5a8905e, 2 |  BRF_GRA },
	{ "so2_26.rom",   0x020000, 0xf3aa5a4a, 2 |  BRF_GRA },
	{ "so2_16.rom",   0x020000, 0xb43cd1a8, 2 |  BRF_GRA },
	{ "so2_20.rom",   0x020000, 0x8ca751a3, 2 |  BRF_GRA },
	{ "so2_10.rom",   0x020000, 0xe9f569fd, 2 |  BRF_GRA },
	{ "so2_22.rom",   0x020000, 0xfce9a377, 2 |  BRF_GRA },
	{ "so2_12.rom",   0x020000, 0xb7df8a06, 2 |  BRF_GRA },

	{ "so2_09.rom",   0x010000, 0xd09d7c7a, 3 |  BRF_ESS | BRF_PRG },

	{ "so2_18.rom",   0x020000, 0xbbea1643, 4 |  BRF_SND },
	{ "so2_19.rom",   0x020000, 0xac58aa71, 4 |  BRF_SND },
};

STD_ROM_PICK(Mercsu) STD_ROM_FN(Mercsu)

struct BurnDriver BurnDrvCpsMercsu = {
	"mercsu", "mercs", NULL, "1990",
	"Mercs (US 900302)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL, 3, HARDWARE_CAPCOM_CPS1,
	NULL,MercsuRomInfo,MercsuRomName,mercsInputInfo, mercsDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// Mercs (US 900608)

static struct BurnRomInfo MercsuaRomDesc[] = {
	{ "so2.30a",      0x020000, 0xe4e725d7, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "so2.35a",      0x020000, 0xe7843445, 1 | BRF_ESS | BRF_PRG },
	{ "so2.31a",      0x020000, 0xc0b91dea, 1 | BRF_ESS | BRF_PRG },
	{ "so2.36a",      0x020000, 0x591edf6c, 1 | BRF_ESS | BRF_PRG },
	{ "so2_32.rom",   0x080000, 0x2eb5cf0c, 1 | BRF_ESS | BRF_PRG },

	{ "so2_gfx6.rom", 0x080000, 0xaa6102af, 2 |  BRF_GRA },
	{ "so2_gfx8.rom", 0x080000, 0x839e6869, 2 |  BRF_GRA },
	{ "so2_gfx2.rom", 0x080000, 0x597c2875, 2 |  BRF_GRA },
	{ "so2_gfx4.rom", 0x080000, 0x912a9ca0, 2 |  BRF_GRA },
	{ "so2_24.rom",   0x020000, 0x3f254efe, 2 |  BRF_GRA },
	{ "so2_14.rom",   0x020000, 0xf5a8905e, 2 |  BRF_GRA },
	{ "so2_26.rom",   0x020000, 0xf3aa5a4a, 2 |  BRF_GRA },
	{ "so2_16.rom",   0x020000, 0xb43cd1a8, 2 |  BRF_GRA },
	{ "so2_20.rom",   0x020000, 0x8ca751a3, 2 |  BRF_GRA },
	{ "so2_10.rom",   0x020000, 0xe9f569fd, 2 |  BRF_GRA },
	{ "so2_22.rom",   0x020000, 0xfce9a377, 2 |  BRF_GRA },
	{ "so2_12.rom",   0x020000, 0xb7df8a06, 2 |  BRF_GRA },

	{ "so2_09.rom",   0x010000, 0xd09d7c7a, 3 |  BRF_ESS | BRF_PRG },

	{ "so2_18.rom",   0x020000, 0xbbea1643, 4 |  BRF_SND },
	{ "so2_19.rom",   0x020000, 0xac58aa71, 4 |  BRF_SND },
};

STD_ROM_PICK(Mercsua) STD_ROM_FN(Mercsua)

struct BurnDriver BurnDrvCpsMercsua = {
	"mercsua", "mercs", NULL, "1990",
	"Mercs (US 900608)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL, 3, HARDWARE_CAPCOM_CPS1,
	NULL,MercsuaRomInfo,MercsuaRomName,mercsInputInfo, mercsDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// Senjo no Ookami II (Japan 900302)

static struct BurnRomInfo MercsjRomDesc[] = {
	{ "so2_30e.rom",  0x020000, 0xe17f9bf7, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "so2_42.bin",   0x020000, 0x2c3884c6, 1 | BRF_ESS | BRF_PRG },
	{ "so2_31e.rom",  0x020000, 0x51204d36, 1 | BRF_ESS | BRF_PRG },
	{ "so2_36e.rom",  0x020000, 0x9cfba8b4, 1 | BRF_ESS | BRF_PRG },
	{ "so2_32.rom",   0x080000, 0x2eb5cf0c, 1 | BRF_ESS | BRF_PRG },

	{ "so2_gfx6.rom", 0x080000, 0xaa6102af, 2 |  BRF_GRA },
	{ "so2_gfx8.rom", 0x080000, 0x839e6869, 2 |  BRF_GRA },
	{ "so2_gfx2.rom", 0x080000, 0x597c2875, 2 |  BRF_GRA },
	{ "so2_gfx4.rom", 0x080000, 0x912a9ca0, 2 |  BRF_GRA },
	{ "so2_24.rom",   0x020000, 0x3f254efe, 2 |  BRF_GRA },
	{ "so2_14.rom",   0x020000, 0xf5a8905e, 2 |  BRF_GRA },
	{ "so2_26.rom",   0x020000, 0xf3aa5a4a, 2 |  BRF_GRA },
	{ "so2_16.rom",   0x020000, 0xb43cd1a8, 2 |  BRF_GRA },
	{ "so2_20.rom",   0x020000, 0x8ca751a3, 2 |  BRF_GRA },
	{ "so2_10.rom",   0x020000, 0xe9f569fd, 2 |  BRF_GRA },
	{ "so2_22.rom",   0x020000, 0xfce9a377, 2 |  BRF_GRA },
	{ "so2_12.rom",   0x020000, 0xb7df8a06, 2 |  BRF_GRA },

	{ "so2_09.rom",   0x010000, 0xd09d7c7a, 3 |  BRF_ESS | BRF_PRG },

	{ "so2_18.rom",   0x020000, 0xbbea1643, 4 |  BRF_SND },
	{ "so2_19.rom",   0x020000, 0xac58aa71, 4 |  BRF_SND },
};

STD_ROM_PICK(Mercsj) STD_ROM_FN(Mercsj)

struct BurnDriver BurnDrvCpsMercsj = {
	"mercsj", "mercs", NULL, "1990",
	"Senjo no Ookami II (Japan 900302)\0", NULL, "Capcom", "CPS1",
	L"\u6226\u5834\u306E\u72FC II (Ookami 2 Japan 900302)\0Senjo no Ookami II\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL, 3, HARDWARE_CAPCOM_CPS1,
	NULL,MercsjRomInfo,MercsjRomName,mercsInputInfo, mercsDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// Mega Twins (World 900619)

static struct BurnRomInfo MtwinsRomDesc[] = {
	{ "che_30.rom",  0x020000, 0x9a2a2db1, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "che_35.rom",  0x020000, 0xa7f96b02, 1 | BRF_ESS | BRF_PRG },
	{ "che_31.rom",  0x020000, 0xbbff8a99, 1 | BRF_ESS | BRF_PRG },
	{ "che_36.rom",  0x020000, 0x0fa00c39, 1 | BRF_ESS | BRF_PRG },
	{ "ch_32.rom",   0x080000, 0x9b70bd41, 1 | BRF_ESS | BRF_PRG },

	{ "ch_gfx5.rom", 0x080000, 0x4ec75f15, 2 | BRF_GRA },
	{ "ch_gfx7.rom", 0x080000, 0xd85d00d6, 2 | BRF_GRA },
	{ "ch_gfx1.rom", 0x080000, 0xf33ca9d4, 2 | BRF_GRA },
	{ "ch_gfx3.rom", 0x080000, 0x0ba2047f, 2 | BRF_GRA },

	{ "ch_09.rom",   0x010000, 0x4d4255b7, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ch_18.rom",   0x020000, 0xf909e8de, 4 | BRF_SND },
	{ "ch_19.rom",   0x020000, 0xfc158cf7, 4 | BRF_SND },
};

STD_ROM_PICK(Mtwins) STD_ROM_FN(Mtwins)

struct BurnDriver BurnDrvCpsMtwins = {
	"mtwins", NULL, NULL, "1990",
	"Mega Twins (World 900619)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,MtwinsRomInfo,MtwinsRomName,mtwinsInputInfo, mtwinsDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Chiki Chiki Boys (Japan 900619)

static struct BurnRomInfo ChikijRomDesc[] = {
	{ "chj36a.bin",  0x020000, 0xec1328d8, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "chj42a.bin",  0x020000, 0x4ae13503, 1 | BRF_ESS | BRF_PRG },
	{ "chj37a.bin",  0x020000, 0x46d2cf7b, 1 | BRF_ESS | BRF_PRG },
	{ "chj43a.bin",  0x020000, 0x8d387fe8, 1 | BRF_ESS | BRF_PRG },
	{ "ch_32.rom",   0x080000, 0x9b70bd41, 1 | BRF_ESS | BRF_PRG },

	{ "ch_gfx5.rom", 0x080000, 0x4ec75f15, 2 | BRF_GRA },
	{ "ch_gfx7.rom", 0x080000, 0xd85d00d6, 2 | BRF_GRA },
	{ "ch_gfx1.rom", 0x080000, 0xf33ca9d4, 2 | BRF_GRA },
	{ "ch_gfx3.rom", 0x080000, 0x0ba2047f, 2 | BRF_GRA },

	{ "ch_09.rom",   0x010000, 0x4d4255b7, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ch_18.rom",   0x020000, 0xf909e8de, 4 | BRF_SND },
	{ "ch_19.rom",   0x020000, 0xfc158cf7, 4 | BRF_SND },
};

STD_ROM_PICK(Chikij) STD_ROM_FN(Chikij)

struct BurnDriver BurnDrvCpsChikij = {
	"chikij", "mtwins", NULL, "1990",
	"Chiki Chiki Boys (Japan 900619)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,ChikijRomInfo,ChikijRomName,mtwinsInputInfo, mtwinsDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Magic Sword - Heroic Fantasy (World 900725)

static struct BurnRomInfo MswordRomDesc[] = {
	{ "mse_30.rom",  0x020000, 0x03fc8dbc, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "mse_35.rom",  0x020000, 0xd5bf66cd, 1 | BRF_ESS | BRF_PRG },
	{ "mse_31.rom",  0x020000, 0x30332bcf, 1 | BRF_ESS | BRF_PRG },
	{ "mse_36.rom",  0x020000, 0x8f7d6ce9, 1 | BRF_ESS | BRF_PRG },
	{ "ms_32.rom",   0x080000, 0x2475ddfc, 1 | BRF_ESS | BRF_PRG },

	{ "ms_gfx5.rom", 0x080000, 0xc00fe7e2, 2 | BRF_GRA },
	{ "ms_gfx7.rom", 0x080000, 0x4ccacac5, 2 | BRF_GRA },
	{ "ms_gfx1.rom", 0x080000, 0x0d2bbe00, 2 | BRF_GRA },
	{ "ms_gfx3.rom", 0x080000, 0x3a1a5bf4, 2 | BRF_GRA },

	{ "ms_9.rom",    0x010000, 0x57b29519, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ms_18.rom",   0x020000, 0xfb64e90d, 4 | BRF_SND },
	{ "ms_19.rom",   0x020000, 0x74f892b9, 4 | BRF_SND },
};

STD_ROM_PICK(Msword) STD_ROM_FN(Msword)

struct BurnDriver BurnDrvCpsMsword = {
	"msword", NULL, NULL, "1990",
	"Magic Sword - Heroic Fantasy (World 900725)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,MswordRomInfo,MswordRomName,mswordInputInfo, mswordDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Magic Sword - Heroic Fantasy (World 900623)

static struct BurnRomInfo Mswordr1RomDesc[] = {
	{ "ms30.11f",    0x020000, 0x21c1f078, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "ms35.11h",    0x020000, 0xa540a73a, 1 | BRF_ESS | BRF_PRG },
	{ "ms31.12f",    0x020000, 0xd7e762b5, 1 | BRF_ESS | BRF_PRG },
	{ "ms36.12h",    0x020000, 0x66f2dcdb, 1 | BRF_ESS | BRF_PRG },
	{ "ms_32.rom",   0x080000, 0x2475ddfc, 1 | BRF_ESS | BRF_PRG },

	{ "ms_gfx5.rom", 0x080000, 0xc00fe7e2, 2 | BRF_GRA },
	{ "ms_gfx7.rom", 0x080000, 0x4ccacac5, 2 | BRF_GRA },
	{ "ms_gfx1.rom", 0x080000, 0x0d2bbe00, 2 | BRF_GRA },
	{ "ms_gfx3.rom", 0x080000, 0x3a1a5bf4, 2 | BRF_GRA },

	{ "ms_9.rom",    0x010000, 0x57b29519, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ms_18.rom",   0x020000, 0xfb64e90d, 4 | BRF_SND },
	{ "ms_19.rom",   0x020000, 0x74f892b9, 4 | BRF_SND },
};

STD_ROM_PICK(Mswordr1) STD_ROM_FN(Mswordr1)

struct BurnDriver BurnDrvCpsMswordr1 = {
	"mswordr1", "msword", NULL, "1990",
	"Magic Sword - Heroic Fantasy (World 900623)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Mswordr1RomInfo,Mswordr1RomName,mswordInputInfo, mswordDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Magic Sword - Heroic Fantasy (US 900725)

static struct BurnRomInfo MsworduRomDesc[] = {
	{ "msu30",       0x020000, 0xd963c816, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "msu35",       0x020000, 0x72f179b3, 1 | BRF_ESS | BRF_PRG },
	{ "msu31",       0x020000, 0x20cd7904, 1 | BRF_ESS | BRF_PRG },
	{ "msu36",       0x020000, 0xbf88c080, 1 | BRF_ESS | BRF_PRG },
	{ "ms_32.rom",   0x080000, 0x2475ddfc, 1 | BRF_ESS | BRF_PRG },

	{ "ms_gfx5.rom", 0x080000, 0xc00fe7e2, 2 | BRF_GRA },
	{ "ms_gfx7.rom", 0x080000, 0x4ccacac5, 2 | BRF_GRA },
	{ "ms_gfx1.rom", 0x080000, 0x0d2bbe00, 2 | BRF_GRA },
	{ "ms_gfx3.rom", 0x080000, 0x3a1a5bf4, 2 | BRF_GRA },

	{ "ms_9.rom",    0x010000, 0x57b29519, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ms_18.rom",   0x020000, 0xfb64e90d, 4 | BRF_SND },
	{ "ms_19.rom",   0x020000, 0x74f892b9, 4 | BRF_SND },
};

STD_ROM_PICK(Mswordu) STD_ROM_FN(Mswordu)

struct BurnDriver BurnDrvCpsMswordu = {
	"mswordu", "msword", NULL, "1990",
	"Magic Sword - Heroic Fantasy (US 900725)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,MsworduRomInfo,MsworduRomName,mswordInputInfo, mswordDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Magic Sword (Japan 900623)

static struct BurnRomInfo MswordjRomDesc[] = {
	{ "msj_36.bin",  0x020000, 0x04f0ef50, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "msj_42.bin",  0x020000, 0x9fcbb9cd, 1 | BRF_ESS | BRF_PRG },
	{ "msj_37.bin",  0x020000, 0x6c060d70, 1 | BRF_ESS | BRF_PRG },
	{ "msj_43.bin",  0x020000, 0xaec77787, 1 | BRF_ESS | BRF_PRG },
	{ "ms_32.rom",   0x080000, 0x2475ddfc, 1 | BRF_ESS | BRF_PRG },

	{ "ms_gfx5.rom", 0x080000, 0xc00fe7e2, 2 | BRF_GRA },
	{ "ms_gfx7.rom", 0x080000, 0x4ccacac5, 2 | BRF_GRA },
	{ "ms_gfx1.rom", 0x080000, 0x0d2bbe00, 2 | BRF_GRA },
	{ "ms_gfx3.rom", 0x080000, 0x3a1a5bf4, 2 | BRF_GRA },

	{ "ms_9.rom",    0x010000, 0x57b29519, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ms_18.rom",   0x020000, 0xfb64e90d, 4 | BRF_SND },
	{ "ms_19.rom",   0x020000, 0x74f892b9, 4 | BRF_SND },
};

STD_ROM_PICK(Mswordj) STD_ROM_FN(Mswordj)

struct BurnDriver BurnDrvCpsMswordj = {
	"mswordj", "msword", NULL, "1990",
	"Magic Sword (Japan 900623)\0", NULL, "Capcom", "CPS1",
	L"Magic Sword - \u30DE\u30B8\u30C3\u30AF\uFF65\u30BD\u30FC\u30C9 (23.06.1990 Japan)\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,MswordjRomInfo,MswordjRomName,mswordInputInfo, mswordDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Carrier Air Wing (World 901012)

static struct BurnRomInfo CawingRomDesc[] = {
	{ "cae_30a.rom", 0x020000, 0x91fceacd, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "cae_35a.rom", 0x020000, 0x3ef03083, 1 | BRF_ESS | BRF_PRG },
	{ "cae_31a.rom", 0x020000, 0xe5b75caf, 1 | BRF_ESS | BRF_PRG },
	{ "cae_36a.rom", 0x020000, 0xc73fd713, 1 | BRF_ESS | BRF_PRG },
	{ "ca_32.rom",   0x080000, 0x0c4837d4, 1 | BRF_ESS | BRF_PRG },

	{ "ca_gfx5.rom", 0x080000, 0x66d4cc37, 2 | BRF_GRA },
	{ "ca_gfx7.rom", 0x080000, 0xb6f896f2, 2 | BRF_GRA },
	{ "ca_gfx1.rom", 0x080000, 0x4d0620fd, 2 | BRF_GRA },
	{ "ca_gfx3.rom", 0x080000, 0x0b0341c3, 2 | BRF_GRA },

	{ "ca_9.rom",    0x010000, 0x96fe7485, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ca_18.rom",   0x020000, 0x4a613a2c, 4 | BRF_SND },
	{ "ca_19.rom",   0x020000, 0x74584493, 4 | BRF_SND },
};

STD_ROM_PICK(Cawing) STD_ROM_FN(Cawing)

struct BurnDriver BurnDrvCpsCawing = {
	"cawing", NULL, NULL, "1990",
	"Carrier Air Wing (World 901012)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,CawingRomInfo,CawingRomName,cawingInputInfo, cawingDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Carrier Air Wing (World 901009)

static struct BurnRomInfo cawingr1RomDesc[] = {
	{ "cae_30.31.f", 0x020000, 0x23305cd5, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "cau_35.11.h", 0x020000, 0x69419113, 1 | BRF_ESS | BRF_PRG },
	{ "cae_31.12.f", 0x020000, 0x9008dfb3, 1 | BRF_ESS | BRF_PRG },
	{ "cae_36.12.h", 0x020000, 0x4dbf6f8e, 1 | BRF_ESS | BRF_PRG },
	{ "ca_32.rom",   0x080000, 0x0c4837d4, 1 | BRF_ESS | BRF_PRG },

	{ "ca_gfx5.rom", 0x080000, 0x66d4cc37, 2 | BRF_GRA },
	{ "ca_gfx7.rom", 0x080000, 0xb6f896f2, 2 | BRF_GRA },
	{ "ca_gfx1.rom", 0x080000, 0x4d0620fd, 2 | BRF_GRA },
	{ "ca_gfx3.rom", 0x080000, 0x0b0341c3, 2 | BRF_GRA },

	{ "ca_9.rom",    0x010000, 0x96fe7485, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ca_18.rom",   0x020000, 0x4a613a2c, 4 | BRF_SND },
	{ "ca_19.rom",   0x020000, 0x74584493, 4 | BRF_SND },
};

STD_ROM_PICK(cawingr1) STD_ROM_FN(cawingr1)

struct BurnDriver BurnDrvCpscawingr1 = {
	"cawingr1", "cawing", NULL, "1990",
	"Carrier Air Wing (World 901009)\0", "", "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,cawingr1RomInfo,cawingr1RomName,cawingInputInfo, cawingDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Carrier Air Wing (US 901012)

static struct BurnRomInfo CawinguRomDesc[] = {
	{ "cae_30a.rom", 0x020000, 0x91fceacd, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "cau_35a.rom", 0x020000, 0xf090d9b2, 1 | BRF_ESS | BRF_PRG },
	{ "cae_31a.rom", 0x020000, 0xe5b75caf, 1 | BRF_ESS | BRF_PRG },
	{ "cae_36a.rom", 0x020000, 0xc73fd713, 1 | BRF_ESS | BRF_PRG },
	{ "ca_32.rom",   0x080000, 0x0c4837d4, 1 | BRF_ESS | BRF_PRG },

	{ "ca_gfx5.rom", 0x080000, 0x66d4cc37, 2 | BRF_GRA },
	{ "ca_gfx7.rom", 0x080000, 0xb6f896f2, 2 | BRF_GRA },
	{ "ca_gfx1.rom", 0x080000, 0x4d0620fd, 2 | BRF_GRA },
	{ "ca_gfx3.rom", 0x080000, 0x0b0341c3, 2 | BRF_GRA },

	{ "ca_9.rom",    0x010000, 0x96fe7485, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ca_18.rom",   0x020000, 0x4a613a2c, 4 | BRF_SND },
	{ "ca_19.rom",   0x020000, 0x74584493, 4 | BRF_SND },
};

STD_ROM_PICK(Cawingu) STD_ROM_FN(Cawingu)

struct BurnDriver BurnDrvCpsCawingu = {
	"cawingu", "cawing", NULL, "1990",
	"Carrier Air Wing (US 901012)\0", "", "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,CawinguRomInfo,CawinguRomName,cawingInputInfo, cawingDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// U.S. Navy (Japan 901012)

static struct BurnRomInfo CawingjRomDesc[] = {
	{ "cae_30a.rom", 0x020000, 0x91fceacd, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "caj42a.bin",  0x020000, 0x039f8362, 1 | BRF_ESS | BRF_PRG },
	{ "cae_31a.rom", 0x020000, 0xe5b75caf, 1 | BRF_ESS | BRF_PRG },
	{ "cae_36a.rom", 0x020000, 0xc73fd713, 1 | BRF_ESS | BRF_PRG },
	{ "caj34.bin",   0x020000, 0x51ea57f4, 1 | BRF_ESS | BRF_PRG },
	{ "caj40.bin",   0x020000, 0x2ab71ae1, 1 | BRF_ESS | BRF_PRG },
	{ "caj35.bin",   0x020000, 0x01d71973, 1 | BRF_ESS | BRF_PRG },
	{ "caj41.bin",   0x020000, 0x3a43b538, 1 | BRF_ESS | BRF_PRG },

	{ "caj09.bin",   0x020000, 0x41b0f9a6, 2 | BRF_GRA },
	{ "caj01.bin",   0x020000, 0x1002d0b8, 2 | BRF_GRA },
	{ "caj13.bin",   0x020000, 0x6f3948b2, 2 | BRF_GRA },
	{ "caj05.bin",   0x020000, 0x207373d7, 2 | BRF_GRA },
	{ "caj24.bin",   0x020000, 0xe356aad7, 2 | BRF_GRA },
	{ "caj17.bin",   0x020000, 0x540f2fd8, 2 | BRF_GRA },
	{ "caj38.bin",   0x020000, 0x2464d4ab, 2 | BRF_GRA },
	{ "caj32.bin",   0x020000, 0x9b5836b3, 2 | BRF_GRA },
	{ "caj10.bin",   0x020000, 0xbf8a5f52, 2 | BRF_GRA },
	{ "caj02.bin",   0x020000, 0x125b018d, 2 | BRF_GRA },
	{ "caj14.bin",   0x020000, 0x8458e7d7, 2 | BRF_GRA },
	{ "caj06.bin",   0x020000, 0xcf80e164, 2 | BRF_GRA },
	{ "caj25.bin",   0x020000, 0xcdd0204d, 2 | BRF_GRA },
	{ "caj18.bin",   0x020000, 0x29c1d4b1, 2 | BRF_GRA },
	{ "caj39.bin",   0x020000, 0xeea23b67, 2 | BRF_GRA },
	{ "caj33.bin",   0x020000, 0xdde3891f, 2 | BRF_GRA },

	{ "ca_9.rom",    0x010000, 0x96fe7485, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ca_18.rom",   0x020000, 0x4a613a2c, 4 | BRF_SND },
	{ "ca_19.rom",   0x020000, 0x74584493, 4 | BRF_SND },
};

STD_ROM_PICK(Cawingj) STD_ROM_FN(Cawingj)

struct BurnDriver BurnDrvCpsCawingj = {
	"cawingj", "cawing", NULL, "1990",
	"U.S. Navy (Japan 901012)\0", "", "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,CawingjRomInfo,CawingjRomName,cawingInputInfo, cawingDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Nemo (World 901130)

static struct BurnRomInfo NemoRomDesc[] = {
	{ "nme_30a.rom", 0x020000, 0xd2c03e56, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "nme_35a.rom", 0x020000, 0x5fd31661, 1 | BRF_ESS | BRF_PRG },
	{ "nme_31a.rom", 0x020000, 0xb2bd4f6f, 1 | BRF_ESS | BRF_PRG },
	{ "nme_36a.rom", 0x020000, 0xee9450e3, 1 | BRF_ESS | BRF_PRG },
	{ "nm_32.rom",   0x080000, 0xd6d1add3, 1 | BRF_ESS | BRF_PRG },

	{ "nm_gfx5.rom", 0x080000, 0x487b8747, 2 | BRF_GRA },
	{ "nm_gfx7.rom", 0x080000, 0x203dc8c6, 2 | BRF_GRA },
	{ "nm_gfx1.rom", 0x080000, 0x9e878024, 2 | BRF_GRA },
	{ "nm_gfx3.rom", 0x080000, 0xbb01e6b6, 2 | BRF_GRA },

	{ "nm_09.rom",   0x010000, 0x0f4b0581, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "nm_18.rom",   0x020000, 0xbab333d4, 4 | BRF_SND },
	{ "nm_19.rom",   0x020000, 0x2650a0a8, 4 | BRF_SND },
};

STD_ROM_PICK(Nemo) STD_ROM_FN(Nemo)

struct BurnDriver BurnDrvCpsNemo = {
	"nemo", NULL, NULL, "1990",
	"Nemo (World 901130)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,NemoRomInfo,NemoRomName,nemoInputInfo, nemoDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Nemo (Japan 901120)

static struct BurnRomInfo NemojRomDesc[] = {
	{ "nm36.bin",    0x020000, 0xdaeceabb, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "nm42.bin",    0x020000, 0x55024740, 1 | BRF_ESS | BRF_PRG },
	{ "nm37.bin",    0x020000, 0x619068b6, 1 | BRF_ESS | BRF_PRG },
	{ "nm43.bin",    0x020000, 0xa948a53b, 1 | BRF_ESS | BRF_PRG },
	{ "nm_32.rom",   0x080000, 0xd6d1add3, 1 | BRF_ESS | BRF_PRG },

	{ "nm_gfx5.rom", 0x080000, 0x487b8747, 2 | BRF_GRA },
	{ "nm_gfx7.rom", 0x080000, 0x203dc8c6, 2 | BRF_GRA },
	{ "nm_gfx1.rom", 0x080000, 0x9e878024, 2 | BRF_GRA },
	{ "nm_gfx3.rom", 0x080000, 0xbb01e6b6, 2 | BRF_GRA },

	{ "nm_09.rom",   0x010000, 0x0f4b0581, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "nm_18.rom",   0x020000, 0xbab333d4, 4 | BRF_SND },
	{ "nm_19.rom",   0x020000, 0x2650a0a8, 4 | BRF_SND },
};

STD_ROM_PICK(Nemoj) STD_ROM_FN(Nemoj)

struct BurnDriver BurnDrvCpsNemoj = {
	"nemoj", "nemo", NULL, "1990",
	"Nemo (Japan 901120)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,NemojRomInfo,NemojRomName,nemoInputInfo, nemoDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (World 910522)

static struct BurnRomInfo Sf2RomDesc[] = {
	{ "sf2e.30g",    0x020000, 0xfe39ee33, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2e.37g",    0x020000, 0xfb92cd74, 1 | BRF_ESS | BRF_PRG },
	{ "sf2e.31g",    0x020000, 0x69a0a301, 1 | BRF_ESS | BRF_PRG },
	{ "sf2e.38g",    0x020000, 0x5e22db70, 1 | BRF_ESS | BRF_PRG },
	{ "sf2e.28g",    0x020000, 0x8bf9f1e5, 1 | BRF_ESS | BRF_PRG },
	{ "sf2e.35g",    0x020000, 0x626ef934, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin", 0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin", 0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",  0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",  0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",  0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",  0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",  0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",  0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",  0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",  0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",  0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",  0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",  0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",  0x080000, 0x0627c831, 2 | BRF_GRA },
 
	{ "sf2_09.bin",  0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2) STD_ROM_FN(Sf2)

struct BurnDriver BurnDrvCpsSf2 = {
	"sf2", NULL, NULL, "1991",
	"Street Fighter II - The World Warrior (World 910522)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2RomInfo,Sf2RomName,sf2InputInfo, sf2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (World 910214)

static struct BurnRomInfo Sf2ebRomDesc[] = {
	{ "sf2_30a.bin",  0x020000, 0x57bd7051, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2e_37b.rom", 0x020000, 0x62691cdd, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_31a.bin",  0x020000, 0xa673143d, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_38a.bin",  0x020000, 0x4c2ccef7, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_28a.bin",  0x020000, 0x4009955e, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_35a.bin",  0x020000, 0x8c1f3994, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin",  0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin",  0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",   0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",   0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",   0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",   0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",   0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",   0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",   0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",   0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",   0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",   0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",   0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",   0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",   0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",   0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",   0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2eb) STD_ROM_FN(Sf2eb)

struct BurnDriver BurnDrvCpsSf2eb = {
	"sf2eb", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (World 910214)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2ebRomInfo,Sf2ebRomName,sf2InputInfo, sf2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (US 910206)

static struct BurnRomInfo Sf2uaRomDesc[] = {
	{ "sf2u.30a",    0x020000, 0x08beb861, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2u.37a",    0x020000, 0xb7638d69, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.31a",    0x020000, 0x0d5394e0, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.38a",    0x020000, 0x42d6a79e, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.28a",    0x020000, 0x387a175c, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.35a",    0x020000, 0xa1a5adcc, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin", 0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin", 0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",  0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",  0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",  0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",  0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",  0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",  0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",  0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",  0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",  0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",  0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",  0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",  0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",  0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ua) STD_ROM_FN(Sf2ua)

struct BurnDriver BurnDrvCpsSf2ua = {
	"sf2ua", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (US 910206)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2uaRomInfo,Sf2uaRomName,sf2InputInfo, sf2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (US 910214)

static struct BurnRomInfo Sf2ubRomDesc[] = {
	{ "sf2_30a.bin", 0x020000, 0x57bd7051, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2u.37b",    0x020000, 0x4a54d479, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_31a.bin", 0x020000, 0xa673143d, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_38a.bin", 0x020000, 0x4c2ccef7, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_28a.bin", 0x020000, 0x4009955e, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_35a.bin", 0x020000, 0x8c1f3994, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin", 0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin", 0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",  0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",  0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",  0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",  0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",  0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",  0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",  0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",  0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",  0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",  0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",  0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",  0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",  0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ub) STD_ROM_FN(Sf2ub)

struct BurnDriver BurnDrvCpsSf2ub = {
	"sf2ub", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (US 910214)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2ubRomInfo,Sf2ubRomName,sf2InputInfo, sf2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (US 910318)

static struct BurnRomInfo Sf2udRomDesc[] = {
	{ "sf2u.30d",    0x020000, 0x4bb2657c, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2u.37d",    0x020000, 0xb33b42f2, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.31d",    0x020000, 0xd57b67d7, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.38d",    0x020000, 0x9c8916ef, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.28d",    0x020000, 0x175819d1, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.35d",    0x020000, 0x82060da4, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin", 0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin", 0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",  0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",  0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",  0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",  0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",  0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",  0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",  0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",  0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",  0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",  0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",  0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",  0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",  0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ud) STD_ROM_FN(Sf2ud)

struct BurnDriver BurnDrvCpsSf2ud = {
	"sf2ud", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (US 910318)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2udRomInfo,Sf2udRomName,sf2InputInfo, sf2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (US 910228)

static struct BurnRomInfo Sf2ueRomDesc[] = {
	{ "sf2u.30e",      0x020000, 0xf37cd088, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2u.37e",      0x020000, 0x6c61a513, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.31e",      0x020000, 0x7c4771b4, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.38e",      0x020000, 0xa4bd0cd9, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.28e",      0x020000, 0xe3b95625, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.35e",      0x020000, 0x3648769a, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin",   0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin",   0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",    0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",    0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",    0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",    0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",    0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",    0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",    0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",    0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",    0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",    0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",    0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",    0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",    0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",    0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",    0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ue) STD_ROM_FN(Sf2ue)

struct BurnDriver BurnDrvCpsSf2ue = {
	"sf2ue", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (US 910228)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2ueRomInfo,Sf2ueRomName,Sf2ueInputInfo, sf2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (US 910411)

static struct BurnRomInfo Sf2ufRomDesc[] = {
	{ "sf2u.30f",    0x020000, 0xfe39ee33, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2u.37f",    0x020000, 0x169e7388, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.31f",    0x020000, 0x69a0a301, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.38f",    0x020000, 0x1510e4e2, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.28f",    0x020000, 0xacd8175b, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.35f",    0x020000, 0xc0a80bd1, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin", 0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin", 0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",  0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",  0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",  0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",  0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",  0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",  0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",  0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",  0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",  0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",  0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",  0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",  0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",  0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2uf) STD_ROM_FN(Sf2uf)

struct BurnDriver BurnDrvCpsSf2uf = {
	"sf2uf", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (US 910411)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2ufRomInfo,Sf2ufRomName,sf2InputInfo, sf2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (US 910522)

static struct BurnRomInfo Sf2uiRomDesc[] = {
	{ "sf2u.30f",    0x020000, 0xfe39ee33, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2u.37i",    0x020000, 0x9df707dd, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.31f",    0x020000, 0x69a0a301, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.38i",    0x020000, 0x4cb46daf, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.28i",    0x020000, 0x1580be4c, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.35i",    0x020000, 0x1468d185, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin", 0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin", 0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",  0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",  0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",  0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",  0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",  0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",  0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",  0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",  0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",  0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",  0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",  0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",  0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",  0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ui) STD_ROM_FN(Sf2ui)

struct BurnDriver BurnDrvCpsSf2ui = {
	"sf2ui", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (US 910522)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2uiRomInfo,Sf2uiRomName,sf2InputInfo, sf2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (US 911101)

static struct BurnRomInfo Sf2ukRomDesc[] = {
	{ "sf2u.30k",    0x020000, 0x8f66076c, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2u.37k",    0x020000, 0x4e1f6a83, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.31k",    0x020000, 0xf9f89f60, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.38k",    0x020000, 0x6ce0a85a, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.28k",    0x020000, 0x8e958f31, 1 | BRF_ESS | BRF_PRG },
	{ "sf2u.35k",    0x020000, 0xfce76fad, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin", 0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin", 0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",  0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",  0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",  0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",  0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",  0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",  0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",  0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",  0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",  0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",  0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",  0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",  0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",  0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2uk) STD_ROM_FN(Sf2uk)

struct BurnDriver BurnDrvCpsSf2uk = {
	"sf2uk", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (US 911101)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2ukRomInfo,Sf2ukRomName,sf2InputInfo, sf2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (Japan 911210)

static struct BurnRomInfo Sf2jRomDesc[] = {
	{ "sf2j30.bin",  0x020000, 0x79022b31, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2j37.bin",  0x020000, 0x516776ec, 1 | BRF_ESS | BRF_PRG },
	{ "sf2j31.bin",  0x020000, 0xfe15cb39, 1 | BRF_ESS | BRF_PRG },
	{ "sf2j38.bin",  0x020000, 0x38614d70, 1 | BRF_ESS | BRF_PRG },
	{ "sf2j28.bin",  0x020000, 0xd283187a, 1 | BRF_ESS | BRF_PRG },
	{ "sf2j35.bin",  0x020000, 0xd28158e4, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin", 0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin", 0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",  0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",  0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",  0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",  0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",  0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",  0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",  0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",  0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",  0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",  0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",  0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",  0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",  0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2j) STD_ROM_FN(Sf2j)

struct BurnDriver BurnDrvCpsSf2j = {
	"sf2j", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (Japan 911210)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2jRomInfo,Sf2jRomName,sf2InputInfo, sf2jDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (Japan 910214)

static struct BurnRomInfo Sf2jaRomDesc[] = {
	{ "sf2_30a.bin",  0x020000, 0x57bd7051, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2j_37a.bin", 0x020000, 0x1e1f6844, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_31a.bin",  0x020000, 0xa673143d, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_38a.bin",  0x020000, 0x4c2ccef7, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_28a.bin",  0x020000, 0x4009955e, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_35a.bin",  0x020000, 0x8c1f3994, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin",  0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin",  0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",   0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",   0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",   0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",   0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",   0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",   0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",   0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",   0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",   0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",   0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",   0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",   0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",   0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",   0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",   0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ja) STD_ROM_FN(Sf2ja)

struct BurnDriver BurnDrvCpsSf2ja = {
	"sf2ja", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (Japan 910214)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2jaRomInfo,Sf2jaRomName,sf2InputInfo, sf2jDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II - The World Warrior (Japan 910306)

static struct BurnRomInfo Sf2jcRomDesc[] = {
	{ "sf2_30c.bin",  0x020000, 0x8add35ec, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sf2j_37c.bin", 0x020000, 0x0d74a256, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_31c.bin",  0x020000, 0xc4fff4a9, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_38c.bin",  0x020000, 0x8210fc0e, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_28c.bin",  0x020000, 0x6eddd5e8, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_35c.bin",  0x020000, 0x6bcb404c, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_29a.bin",  0x020000, 0xbb4af315, 1 | BRF_ESS | BRF_PRG },
	{ "sf2_36a.bin",  0x020000, 0xc02a13eb, 1 | BRF_ESS | BRF_PRG },

	{ "sf2_06.bin",   0x080000, 0x22c9cc8e, 2 | BRF_GRA },
	{ "sf2_08.bin",   0x080000, 0x57213be8, 2 | BRF_GRA },
	{ "sf2_05.bin",   0x080000, 0xba529b4f, 2 | BRF_GRA },
	{ "sf2_07.bin",   0x080000, 0x4b1b33a8, 2 | BRF_GRA },
	{ "sf2_15.bin",   0x080000, 0x2c7e2229, 2 | BRF_GRA },
	{ "sf2_17.bin",   0x080000, 0xb5548f17, 2 | BRF_GRA },
	{ "sf2_14.bin",   0x080000, 0x14b84312, 2 | BRF_GRA },
	{ "sf2_16.bin",   0x080000, 0x5e9cd89a, 2 | BRF_GRA },
	{ "sf2_25.bin",   0x080000, 0x994bfa58, 2 | BRF_GRA },
	{ "sf2_27.bin",   0x080000, 0x3e66ad9d, 2 | BRF_GRA },
	{ "sf2_24.bin",   0x080000, 0xc1befaa8, 2 | BRF_GRA },
	{ "sf2_26.bin",   0x080000, 0x0627c831, 2 | BRF_GRA },

	{ "sf2_09.bin",   0x010000, 0xa4823a1b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sf2_18.bin",   0x020000, 0x7f162009, 4 | BRF_SND },
	{ "sf2_19.bin",   0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2jc) STD_ROM_FN(Sf2jc)

struct BurnDriver BurnDrvCpsSf2jc = {
	"sf2jc", "sf2", NULL, "1991",
	"Street Fighter II - The World Warrior (Japan 910306)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Sf2jcRomInfo,Sf2jcRomName,sf2InputInfo, sf2jDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Three Wonders (World 910520)

static struct BurnRomInfo Wonder3wRomDesc[] = {
	{ "rte.30a",     0x020000, 0xef5b8b33, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "rte.35a",     0x020000, 0x7d705529, 1 | BRF_ESS | BRF_PRG },
	{ "rte.31a",     0x020000, 0x32835e5e, 1 | BRF_ESS | BRF_PRG },
	{ "rte.36a",     0x020000, 0x7637975f, 1 | BRF_ESS | BRF_PRG },
	{ "3wonders.28", 0x020000, 0x054137c8, 1 | BRF_ESS | BRF_PRG },
	{ "3wonders.33", 0x020000, 0x7264cb1b, 1 | BRF_ESS | BRF_PRG },
	{ "rte.29a",     0x020000, 0xcddaa919, 1 | BRF_ESS | BRF_PRG },
	{ "rte.34a",     0x020000, 0xed52e7e5, 1 | BRF_ESS | BRF_PRG },

	{ "3wonders.05", 0x080000, 0x86aef804, 2 | BRF_GRA },
	{ "3wonders.07", 0x080000, 0x4f057110, 2 | BRF_GRA },
	{ "3wonders.01", 0x080000, 0x902489d0, 2 | BRF_GRA },
	{ "3wonders.03", 0x080000, 0xe35ce720, 2 | BRF_GRA },
	{ "3wonders.06", 0x080000, 0x13cb0e7c, 2 | BRF_GRA },
	{ "3wonders.08", 0x080000, 0x1f055014, 2 | BRF_GRA },
	{ "3wonders.02", 0x080000, 0xe9a034f4, 2 | BRF_GRA },
	{ "3wonders.04", 0x080000, 0xdf0eea8b, 2 | BRF_GRA },

	{ "rt23.bin",    0x010000, 0x7d5a77a7, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "3wonders.18", 0x020000, 0x26b211ab, 4 | BRF_SND },
	{ "3wonders.19", 0x020000, 0xdbe64ad0, 4 | BRF_SND },
};

STD_ROM_PICK(Wonder3w) STD_ROM_FN(Wonder3w)

struct BurnDriver BurnDrvCps3wonders = {
	"3wonders", NULL, NULL, "1991",
	"Three Wonders (World 910520)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,Wonder3wRomInfo,Wonder3wRomName,wonders3InputInfo, threewondersDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Three Wonders (US 910520)

static struct BurnRomInfo Wonder3uRomDesc[] = {
	{ "3wonders.30", 0x020000, 0x0b156fd8, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "3wonders.35", 0x020000, 0x57350bf4, 1 | BRF_ESS | BRF_PRG },
	{ "3wonders.31", 0x020000, 0x0e723fcc, 1 | BRF_ESS | BRF_PRG },
	{ "3wonders.36", 0x020000, 0x523a45dc, 1 | BRF_ESS | BRF_PRG },
	{ "3wonders.28", 0x020000, 0x054137c8, 1 | BRF_ESS | BRF_PRG },
	{ "3wonders.33", 0x020000, 0x7264cb1b, 1 | BRF_ESS | BRF_PRG }, 
	{ "3wonders.29", 0x020000, 0x37ba3e20, 1 | BRF_ESS | BRF_PRG },
	{ "3wonders.34", 0x020000, 0xf99f46c0, 1 | BRF_ESS | BRF_PRG },

	{ "3wonders.05", 0x080000, 0x86aef804, 2 | BRF_GRA },
	{ "3wonders.07", 0x080000, 0x4f057110, 2 | BRF_GRA },
	{ "3wonders.01", 0x080000, 0x902489d0, 2 | BRF_GRA },
	{ "3wonders.03", 0x080000, 0xe35ce720, 2 | BRF_GRA },
	{ "3wonders.06", 0x080000, 0x13cb0e7c, 2 | BRF_GRA },
	{ "3wonders.08", 0x080000, 0x1f055014, 2 | BRF_GRA },
	{ "3wonders.02", 0x080000, 0xe9a034f4, 2 | BRF_GRA },
	{ "3wonders.04", 0x080000, 0xdf0eea8b, 2 | BRF_GRA },

	{ "3wonders.09", 0x010000, 0xabfca165, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "3wonders.18", 0x020000, 0x26b211ab, 4 | BRF_SND },
	{ "3wonders.19", 0x020000, 0xdbe64ad0, 4 | BRF_SND },
};

STD_ROM_PICK(Wonder3u) STD_ROM_FN(Wonder3u)

struct BurnDriver BurnDrvCps3wonderu = {
	"3wonderu", "3wonders", NULL, "1991",
	"Three Wonders (US 910520)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Wonder3uRomInfo,Wonder3uRomName,wonders3InputInfo, threewondersDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Wonder 3 (Japan 910520)

static struct BurnRomInfo Wonder3RomDesc[] = {
	{ "rtj36.bin",   0x020000, 0xe3741247, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "rtj42.bin",   0x020000, 0xb4baa117, 1 | BRF_ESS | BRF_PRG },
	{ "rtj37.bin",   0x020000, 0xa1f677b0, 1 | BRF_ESS | BRF_PRG },
	{ "rtj43.bin",   0x020000, 0x85337a47, 1 | BRF_ESS | BRF_PRG },
	{ "3wonders.28", 0x020000, 0x054137c8, 1 | BRF_ESS | BRF_PRG },
	{ "3wonders.33", 0x020000, 0x7264cb1b, 1 | BRF_ESS | BRF_PRG },
	{ "rtj35.bin",   0x020000, 0xe72f9ea3, 1 | BRF_ESS | BRF_PRG },
	{ "rtj41.bin",   0x020000, 0xa11ee998, 1 | BRF_ESS | BRF_PRG },

 	{ "3wonders.05", 0x080000, 0x86aef804, 2 | BRF_GRA },
	{ "3wonders.07", 0x080000, 0x4f057110, 2 | BRF_GRA },
	{ "3wonders.01", 0x080000, 0x902489d0, 2 | BRF_GRA },
	{ "3wonders.03", 0x080000, 0xe35ce720, 2 | BRF_GRA },
	{ "3wonders.06", 0x080000, 0x13cb0e7c, 2 | BRF_GRA },
	{ "3wonders.08", 0x080000, 0x1f055014, 2 | BRF_GRA },
	{ "3wonders.02", 0x080000, 0xe9a034f4, 2 | BRF_GRA },
	{ "3wonders.04", 0x080000, 0xdf0eea8b, 2 | BRF_GRA },

	{ "3wonders.09", 0x010000, 0xabfca165, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "3wonders.18", 0x020000, 0x26b211ab, 4 | BRF_SND },
	{ "3wonders.19", 0x020000, 0xdbe64ad0, 4 | BRF_SND },
};

STD_ROM_PICK(Wonder3) STD_ROM_FN(Wonder3)

struct BurnDriver BurnDrvCpsWonder3 = {
	"wonder3", "3wonders", NULL, "1991",
	"Wonder 3 (Japan 910520)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,Wonder3RomInfo,Wonder3RomName,wonders3InputInfo, threewondersDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// The King of Dragons (World 910711)

static struct BurnRomInfo KodRomDesc[] = {
	{ "kod30.rom", 0x020000, 0xc7414fd4, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "kod37.rom", 0x020000, 0xa5bf40d2, 1 | BRF_ESS | BRF_PRG },
	{ "kod31.rom", 0x020000, 0x1fffc7bd, 1 | BRF_ESS | BRF_PRG },
	{ "kod38.rom", 0x020000, 0x89e57a82, 1 | BRF_ESS | BRF_PRG },
	{ "kod28.rom", 0x020000, 0x9367bcd9, 1 | BRF_ESS | BRF_PRG },
	{ "kod35.rom", 0x020000, 0x4ca6a48a, 1 | BRF_ESS | BRF_PRG },
	{ "kod29.rom", 0x020000, 0x6a0ba878, 1 | BRF_ESS | BRF_PRG },
	{ "kod36.rom", 0x020000, 0xb509b39d, 1 | BRF_ESS | BRF_PRG },

	{ "kod02.rom", 0x080000, 0xe45b8701, 2 | BRF_GRA },
	{ "kod04.rom", 0x080000, 0xa7750322, 2 | BRF_GRA },
	{ "kod01.rom", 0x080000, 0x5f74bf78, 2 | BRF_GRA },
	{ "kod03.rom", 0x080000, 0x5e5303bf, 2 | BRF_GRA },
	{ "kod11.rom", 0x080000, 0x113358f3, 2 | BRF_GRA },
	{ "kod13.rom", 0x080000, 0x38853c44, 2 | BRF_GRA },
	{ "kod10.rom", 0x080000, 0x9ef36604, 2 | BRF_GRA },
	{ "kod12.rom", 0x080000, 0x402b9b4f, 2 | BRF_GRA },

	{ "kod09.rom", 0x010000, 0xf5514510, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "kod18.rom", 0x020000, 0x69ecb2c8, 4 | BRF_SND },
	{ "kod19.rom", 0x020000, 0x02d851c1, 4 | BRF_SND },
};

STD_ROM_PICK(Kod) STD_ROM_FN(Kod)

struct BurnDriver BurnDrvCpsKod = {
	"kod", NULL, NULL, "1991",
	"The King of Dragons (World 910711)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,3,HARDWARE_CAPCOM_CPS1,
	NULL,KodRomInfo,KodRomName,kodInputInfo, kodDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// The King of Dragons (US 910910)

static struct BurnRomInfo KoduRomDesc[] = {
	{ "kdu-30b.bin", 0x020000, 0x825817f9, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "kdu-37b.bin", 0x020000, 0xd2422dfb, 1 | BRF_ESS | BRF_PRG },
	{ "kdu-31b.bin", 0x020000, 0x9af36039, 1 | BRF_ESS | BRF_PRG },
	{ "kdu-38b.bin", 0x020000, 0xbe8405a1, 1 | BRF_ESS | BRF_PRG },
	{ "kod28.rom",   0x020000, 0x9367bcd9, 1 | BRF_ESS | BRF_PRG },
	{ "kod35.rom",   0x020000, 0x4ca6a48a, 1 | BRF_ESS | BRF_PRG },
	{ "kd-29.bin",   0x020000, 0x0360fa72, 1 | BRF_ESS | BRF_PRG },
	{ "kd-36a.bin",  0x020000, 0x95a3cef8, 1 | BRF_ESS | BRF_PRG },

	{ "kod02.rom",   0x080000, 0xe45b8701, 2 | BRF_GRA },
	{ "kod04.rom",   0x080000, 0xa7750322, 2 | BRF_GRA },
	{ "kod01.rom",   0x080000, 0x5f74bf78, 2 | BRF_GRA },
	{ "kod03.rom",   0x080000, 0x5e5303bf, 2 | BRF_GRA },
	{ "kod11.rom",   0x080000, 0x113358f3, 2 | BRF_GRA },
	{ "kod13.rom",   0x080000, 0x38853c44, 2 | BRF_GRA },
	{ "kod10.rom",   0x080000, 0x9ef36604, 2 | BRF_GRA },
	{ "kod12.rom",   0x080000, 0x402b9b4f, 2 | BRF_GRA },

	{ "kd09.bin",    0x010000, 0xbac6ec26, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "kd18.bin",    0x020000, 0x4c63181d, 4 | BRF_SND },
	{ "kd19.bin",    0x020000, 0x92941b80, 4 | BRF_SND },
};

STD_ROM_PICK(Kodu) STD_ROM_FN(Kodu)

struct BurnDriver BurnDrvCpsKodu = {
	"kodu", "kod", NULL, "1991",
	"The King of Dragons (US 910910)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,3,HARDWARE_CAPCOM_CPS1,
	NULL,KoduRomInfo,KoduRomName,kodInputInfo, kodDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// The King of Dragons (Japan 910805)

static struct BurnRomInfo KodjRomDesc[] = {
	{ "kd30.bin",  0x020000, 0xebc788ad, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "kd37.bin",  0x020000, 0xe55c3529, 1 | BRF_ESS | BRF_PRG },
	{ "kd31.bin",  0x020000, 0xc710d722, 1 | BRF_ESS | BRF_PRG },
	{ "kd38.bin",  0x020000, 0x57d6ed3a, 1 | BRF_ESS | BRF_PRG },
	{ "kd33.bin",  0x080000, 0x9bd7ad4b, 1 | BRF_ESS | BRF_PRG },

	{ "kod02.rom", 0x080000, 0xe45b8701, 2 | BRF_GRA },
	{ "kod04.rom", 0x080000, 0xa7750322, 2 | BRF_GRA },
	{ "kod01.rom", 0x080000, 0x5f74bf78, 2 | BRF_GRA },
	{ "kod03.rom", 0x080000, 0x5e5303bf, 2 | BRF_GRA },
	{ "kod11.rom", 0x080000, 0x113358f3, 2 | BRF_GRA },
	{ "kod13.rom", 0x080000, 0x38853c44, 2 | BRF_GRA },
	{ "kod10.rom", 0x080000, 0x9ef36604, 2 | BRF_GRA },
	{ "kod12.rom", 0x080000, 0x402b9b4f, 2 | BRF_GRA },

	{ "kd09.bin",  0x010000, 0xbac6ec26, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "kd18.bin",  0x020000, 0x4c63181d, 4 | BRF_SND },
	{ "kd19.bin",  0x020000, 0x92941b80, 4 | BRF_SND },
};

STD_ROM_PICK(Kodj) STD_ROM_FN(Kodj)

struct BurnDriver BurnDrvCpsKodj = {
	"kodj", "kod", NULL, "1991",
	"The King of Dragons (Japan 910805)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,3,HARDWARE_CAPCOM_CPS1,
	NULL,KodjRomInfo,KodjRomName,kodInputInfo, kodDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// The King of Dragons (bootleg)

static struct BurnRomInfo KodbRomDesc[] = {
	{ "kod.17",    0x080000, 0x036dd74c, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "kod.18",    0x080000, 0x3e4b7295, 1 | BRF_ESS | BRF_PRG },

	{ "kod02.rom", 0x080000, 0xe45b8701, 2 | BRF_GRA },
	{ "kod04.rom", 0x080000, 0xa7750322, 2 | BRF_GRA },
	{ "kod01.rom", 0x080000, 0x5f74bf78, 2 | BRF_GRA },
	{ "kod03.rom", 0x080000, 0x5e5303bf, 2 | BRF_GRA },
	{ "kod11.rom", 0x080000, 0x113358f3, 2 | BRF_GRA },
	{ "kod13.rom", 0x080000, 0x38853c44, 2 | BRF_GRA },
	{ "kod10.rom", 0x080000, 0x9ef36604, 2 | BRF_GRA },
	{ "kod12.rom", 0x080000, 0x402b9b4f, 2 | BRF_GRA },

	{ "kod.15",    0x010000, 0x01cae60c, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "kd18.bin",  0x020000, 0x4c63181d, 4 | BRF_SND },
	{ "kd19.bin",  0x020000, 0x92941b80, 4 | BRF_SND },
};

STD_ROM_PICK(Kodb) STD_ROM_FN(Kodb)

void bootleg_progload()
{
	// First two program roms are 512kb and
	// should be interleaved
	BurnLoadRom(CpsRom + 0x000001, 0, 2);
	BurnLoadRom(CpsRom + 0x000000, 1, 2);
}

int kodbInit()
{
	Kodb = 1;
	pCpsInitCallback = bootleg_progload;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsKodb = {
	"kodb", "kod", NULL, "1991",
	"The King of Dragons (bootleg)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_CLONE | BDF_BOOTLEG,3,HARDWARE_CAPCOM_CPS1,
	NULL,KodbRomInfo,KodbRomName,kodInputInfo, kodDIPInfo,
	kodbInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Captain Commando (World 911014)

static struct BurnRomInfo CaptcommRomDesc[] = {
	{ "cce_23d.rom", 0x080000, 0x19c58ece, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "cc_22d.rom",  0x080000, 0xa91949b7, 1 | BRF_ESS | BRF_PRG },
	{ "cc_24d.rom",  0x020000, 0x680e543f, 1 | BRF_ESS | BRF_PRG },
	{ "cc_28d.rom",  0x020000, 0x8820039f, 1 | BRF_ESS | BRF_PRG },

	{ "gfx_01.rom",  0x080000, 0x7261d8ba, 2 | BRF_GRA },
	{ "gfx_03.rom",  0x080000, 0x6a60f949, 2 | BRF_GRA },
	{ "gfx_02.rom",  0x080000, 0x00637302, 2 | BRF_GRA },
	{ "gfx_04.rom",  0x080000, 0xcc87cf61, 2 | BRF_GRA },
	{ "gfx_05.rom",  0x080000, 0x28718bed, 2 | BRF_GRA },
	{ "gfx_07.rom",  0x080000, 0xd4acc53a, 2 | BRF_GRA },
	{ "gfx_06.rom",  0x080000, 0x0c69f151, 2 | BRF_GRA },
	{ "gfx_08.rom",  0x080000, 0x1f9ebb97, 2 | BRF_GRA },

	{ "cc_09.rom",   0x010000, 0x698e8b58, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "cc_18.rom",   0x020000, 0x6de2c2db, 4 | BRF_SND },
	{ "cc_19.rom",   0x020000, 0xb99091ae, 4 | BRF_SND },
};

STD_ROM_PICK(Captcomm) STD_ROM_FN(Captcomm)

struct BurnDriver BurnDrvCpsCaptcomm = {
	"captcomm", NULL, NULL, "1991",
	"Captain Commando (World 911014)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,4,HARDWARE_CAPCOM_CPS1,
	NULL,CaptcommRomInfo,CaptcommRomName,captcommInputInfo, captcommDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Captain Commando (US 910928)

static struct BurnRomInfo CaptcomuRomDesc[] = {
	{ "23b",        0x080000, 0x03da44fd, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "22c",        0x080000, 0x9b82a052, 1 | BRF_ESS | BRF_PRG },
	{ "24b",        0x020000, 0x84ff99b2, 1 | BRF_ESS | BRF_PRG },
	{ "28b",        0x020000, 0xfbcec223, 1 | BRF_ESS | BRF_PRG },

	{ "gfx_01.rom", 0x080000, 0x7261d8ba, 2 | BRF_GRA },
	{ "gfx_03.rom", 0x080000, 0x6a60f949, 2 | BRF_GRA },
	{ "gfx_02.rom", 0x080000, 0x00637302, 2 | BRF_GRA },
	{ "gfx_04.rom", 0x080000, 0xcc87cf61, 2 | BRF_GRA },
	{ "gfx_05.rom", 0x080000, 0x28718bed, 2 | BRF_GRA },
	{ "gfx_07.rom", 0x080000, 0xd4acc53a, 2 | BRF_GRA },
	{ "gfx_06.rom", 0x080000, 0x0c69f151, 2 | BRF_GRA },
	{ "gfx_08.rom", 0x080000, 0x1f9ebb97, 2 | BRF_GRA },

	{ "cc_09.rom",  0x010000, 0x698e8b58, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "cc_18.rom",  0x020000, 0x6de2c2db, 4 | BRF_SND },
	{ "cc_19.rom",  0x020000, 0xb99091ae, 4 | BRF_SND },
};

STD_ROM_PICK(Captcomu) STD_ROM_FN(Captcomu)

struct BurnDriver BurnDrvCpsCaptcomu = {
	"captcomu", "captcomm", NULL, "1991",
	"Captain Commando (US 910928)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,4,HARDWARE_CAPCOM_CPS1,
	NULL,CaptcomuRomInfo,CaptcomuRomName,captcommInputInfo, captcommDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Captain Commando (Japan 911202)

static struct BurnRomInfo CaptcomjRomDesc[] = {
	{ "cc23.bin",   0x080000, 0x5b482b62, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "cc22.bin",   0x080000, 0x0fd34195, 1 | BRF_ESS | BRF_PRG },
	{ "cc24.bin",   0x020000, 0x3a794f25, 1 | BRF_ESS | BRF_PRG },
	{ "cc28.bin",   0x020000, 0xfc3c2906, 1 | BRF_ESS | BRF_PRG },

	{ "gfx_01.rom", 0x080000, 0x7261d8ba, 2 | BRF_GRA },
	{ "gfx_03.rom", 0x080000, 0x6a60f949, 2 | BRF_GRA },
	{ "gfx_02.rom", 0x080000, 0x00637302, 2 | BRF_GRA },
	{ "gfx_04.rom", 0x080000, 0xcc87cf61, 2 | BRF_GRA },
	{ "gfx_05.rom", 0x080000, 0x28718bed, 2 | BRF_GRA },
	{ "gfx_07.rom", 0x080000, 0xd4acc53a, 2 | BRF_GRA },
	{ "gfx_06.rom", 0x080000, 0x0c69f151, 2 | BRF_GRA },
	{ "gfx_08.rom", 0x080000, 0x1f9ebb97, 2 | BRF_GRA },

	{ "cc_09.rom",  0x010000, 0x698e8b58, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "cc_18.rom",  0x020000, 0x6de2c2db, 4 | BRF_SND },
	{ "cc_19.rom",  0x020000, 0xb99091ae, 4 | BRF_SND },
};

STD_ROM_PICK(Captcomj) STD_ROM_FN(Captcomj)

struct BurnDriver BurnDrvCpsCaptcomj = {
	"captcomj", "captcomm", NULL, "1991",
	"Captain Commando (Japan 911202)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,4,HARDWARE_CAPCOM_CPS1,
	NULL,CaptcomjRomInfo,CaptcomjRomName,captcommInputInfo, captcommDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Knights of the Round (World 911127)

static struct BurnRomInfo KnightsRomDesc[] = {
	{ "kr_23e.rom",  0x080000, 0x1b3997eb, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "kr_22.rom",   0x080000, 0xd0b671a9, 1 | BRF_ESS | BRF_PRG },

	{ "kr_gfx1.rom", 0x080000, 0x9e36c1a4, 2 | BRF_GRA },
	{ "kr_gfx3.rom", 0x080000, 0xc5832cae, 2 | BRF_GRA },
	{ "kr_gfx2.rom", 0x080000, 0xf095be2d, 2 | BRF_GRA },
	{ "kr_gfx4.rom", 0x080000, 0x179dfd96, 2 | BRF_GRA },
	{ "kr_gfx5.rom", 0x080000, 0x1f4298d2, 2 | BRF_GRA },
	{ "kr_gfx7.rom", 0x080000, 0x37fa8751, 2 | BRF_GRA },
	{ "kr_gfx6.rom", 0x080000, 0x0200bc3d, 2 | BRF_GRA },
	{ "kr_gfx8.rom", 0x080000, 0x0bb2b4e7, 2 | BRF_GRA },

	{ "kr_09.rom",   0x010000, 0x5e44d9ee, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "kr_18.rom",   0x020000, 0xda69d15f, 4 | BRF_SND },
	{ "kr_19.rom",   0x020000, 0xbfc654e9, 4 | BRF_SND },
};

STD_ROM_PICK(Knights) STD_ROM_FN(Knights)

struct BurnDriver BurnDrvCpsKnights = {
	"knights", NULL, NULL, "1991",
	"Knights of the Round (World 911127)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,3,HARDWARE_CAPCOM_CPS1,
	NULL,KnightsRomInfo,KnightsRomName,knightsInputInfo, knightsDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Knights of the Round (US 911127)

static struct BurnRomInfo KnightsuRomDesc[] = {
	{ "kru23.rom",   0x080000, 0x252bc2ba, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "kr_22.rom",   0x080000, 0xd0b671a9, 1 | BRF_ESS | BRF_PRG },

	{ "kr_gfx1.rom", 0x080000, 0x9e36c1a4, 2 | BRF_GRA },
	{ "kr_gfx3.rom", 0x080000, 0xc5832cae, 2 | BRF_GRA },
	{ "kr_gfx2.rom", 0x080000, 0xf095be2d, 2 | BRF_GRA },
	{ "kr_gfx4.rom", 0x080000, 0x179dfd96, 2 | BRF_GRA },
	{ "kr_gfx5.rom", 0x080000, 0x1f4298d2, 2 | BRF_GRA },
	{ "kr_gfx7.rom", 0x080000, 0x37fa8751, 2 | BRF_GRA },
	{ "kr_gfx6.rom", 0x080000, 0x0200bc3d, 2 | BRF_GRA },
	{ "kr_gfx8.rom", 0x080000, 0x0bb2b4e7, 2 | BRF_GRA },

	{ "kr_09.rom",   0x010000, 0x5e44d9ee, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "kr_18.rom",   0x020000, 0xda69d15f, 4 | BRF_SND },
	{ "kr_19.rom",   0x020000, 0xbfc654e9, 4 | BRF_SND },
};

STD_ROM_PICK(Knightsu) STD_ROM_FN(Knightsu)

struct BurnDriver BurnDrvCpsKnightsu = {
	"knightsu", "knights", NULL, "1991",
	"Knights of the Round (US 911127)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,3,HARDWARE_CAPCOM_CPS1,
	NULL,KnightsuRomInfo,KnightsuRomName,knightsInputInfo, knightsDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Knights of the Round (Japan 911127)

static struct BurnRomInfo KnightsjRomDesc[] = {
	{ "krj30.bin",   0x020000, 0xad3d1a8e, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "krj37.bin",   0x020000, 0xe694a491, 1 | BRF_ESS | BRF_PRG },
	{ "krj31.bin",   0x020000, 0x85596094, 1 | BRF_ESS | BRF_PRG },
	{ "krj38.bin",   0x020000, 0x9198bf8f, 1 | BRF_ESS | BRF_PRG },
	{ "kr_22.rom",   0x080000, 0xd0b671a9, 1 | BRF_ESS | BRF_PRG },

	{ "kr_gfx1.rom", 0x080000, 0x9e36c1a4, 2 | BRF_GRA },
	{ "kr_gfx3.rom", 0x080000, 0xc5832cae, 2 | BRF_GRA },
	{ "kr_gfx2.rom", 0x080000, 0xf095be2d, 2 | BRF_GRA },
	{ "kr_gfx4.rom", 0x080000, 0x179dfd96, 2 | BRF_GRA },
	{ "kr_gfx5.rom", 0x080000, 0x1f4298d2, 2 | BRF_GRA },
	{ "kr_gfx7.rom", 0x080000, 0x37fa8751, 2 | BRF_GRA },
	{ "kr_gfx6.rom", 0x080000, 0x0200bc3d, 2 | BRF_GRA },
	{ "kr_gfx8.rom", 0x080000, 0x0bb2b4e7, 2 | BRF_GRA },

	{ "kr_09.rom",   0x010000, 0x5e44d9ee, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "kr_18.rom",   0x020000, 0xda69d15f, 4 | BRF_SND },
	{ "kr_19.rom",   0x020000, 0xbfc654e9, 4 | BRF_SND },
};

STD_ROM_PICK(Knightsj) STD_ROM_FN(Knightsj)

struct BurnDriver BurnDrvCpsKnightsj = {
	"knightsj", "knights", NULL, "1991",
	"Knights of the Round (Japan 911127)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,3,HARDWARE_CAPCOM_CPS1,
	NULL,KnightsjRomInfo,KnightsjRomName,knightsInputInfo, knightsDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II': Champion Edition (World 920313)

static struct BurnRomInfo Sf2ceRomDesc[] = {
	{"sf2ce.23",    0x080000, 0x3f846b74, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"sf2ce.22",    0x080000, 0x99f1cca4, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ce) STD_ROM_FN(Sf2ce)

struct BurnDriver BurnDrvCpsSf2ce = {
	"sf2ce", NULL, NULL, "1992",
	"Street Fighter II': Champion Edition (World 920313)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2ceRomInfo, Sf2ceRomName, sf2InputInfo, Sf2ceDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (US 920313)

static struct BurnRomInfo Sf2ceuaRomDesc[] = {
	{"s92u-23a",    0x080000, 0xac44415b, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"sf2ce.22",    0x080000, 0x99f1cca4, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ceua) STD_ROM_FN(Sf2ceua)

struct BurnDriver BurnDrvCpsSf2ceua = {
	"sf2ceua", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (US 920313)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2ceuaRomInfo, Sf2ceuaRomName, sf2InputInfo, Sf2ceDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (US 920513)

static struct BurnRomInfo Sf2ceubRomDesc[] = {
	{"s92-23b",    0x080000, 0x996a3015, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"s92-22b",    0x080000, 0x2bbe15ed, 1 | BRF_ESS | BRF_PRG },
	{"s92-21b",    0x080000, 0xb383cb1c, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin", 0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin", 0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin", 0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin", 0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin", 0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin", 0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin", 0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin", 0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin", 0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin", 0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin", 0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin", 0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin", 0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin", 0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin", 0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ceub) STD_ROM_FN(Sf2ceub)

struct BurnDriver BurnDrvCpsSf2ceub={
	"sf2ceub", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (US 920513)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2ceubRomInfo, Sf2ceubRomName, sf2InputInfo, Sf2ceDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (US 920803)

static struct BurnRomInfo Sf2ceucRomDesc[] = {
	{"s92-23c",    0x080000, 0x0a8b6aa2, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"s92-22c",    0x080000, 0x5fd8630b, 1 | BRF_ESS | BRF_PRG },
	{"s92-21b",    0x080000, 0xb383cb1c, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin", 0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin", 0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin", 0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin", 0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin", 0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin", 0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin", 0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin", 0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin", 0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin", 0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin", 0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin", 0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin", 0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin", 0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin", 0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2ceuc) STD_ROM_FN(Sf2ceuc)

struct BurnDriver BurnDrvCpsSf2ceuc={
	"sf2ceuc", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (US 920803)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2ceucRomInfo, Sf2ceucRomName, sf2InputInfo, Sf2ceDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (Japan 920513)

static struct BurnRomInfo Sf2cejRomDesc[] = {
	{"s92j_23b.bin", 0x080000, 0x140876c5, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"s92j_22b.bin", 0x080000, 0x2fbb3bfe, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin",  0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",   0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",   0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",   0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",   0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",   0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",   0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",   0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",   0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",   0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",   0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",   0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",   0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",   0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",   0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",   0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2cej) STD_ROM_FN(Sf2cej)

struct BurnDriver BurnDrvCpsSf2cej={
	"sf2cej", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (Japan 920513)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2cejRomInfo, Sf2cejRomName, sf2InputInfo, Sf2cejDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (Rainbow set 1)

static struct BurnRomInfo Sf2rbRomDesc[] = {
	{"sf2d__23.rom", 0x080000, 0x450532b0, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"sf2d__22.rom", 0x080000, 0xfe9d9cf5, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin",  0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",   0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",   0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",   0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",   0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",   0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",   0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",   0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",   0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",   0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",   0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",   0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",   0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",   0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",   0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",   0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2rb) STD_ROM_FN(Sf2rb)

void sf2rb_patch()
{
	// byteswap first two roms
	BurnByteswap(CpsRom, 0x100000);

	*((unsigned short*)(CpsRom + 0xe5464)) = 0x6012;
}

int sf2rbInit()
{
	pCpsInitCallback = sf2rb_patch;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsSf2rb={
	"sf2rb", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (Rainbow set 1)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2rbRomInfo, Sf2rbRomName, sf2InputInfo, Sf2ceDIPInfo, 
	sf2rbInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (Rainbow set 2)

static struct BurnRomInfo Sf2rb2RomDesc[] = {
	{"27.bin",      0x020000, 0x40296ecd, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"31.bin",      0x020000, 0x87954a41, 1 | BRF_ESS | BRF_PRG },
	{"26.bin",      0x020000, 0xa6974195, 1 | BRF_ESS | BRF_PRG },
	{"30.bin",      0x020000, 0x8141fe32, 1 | BRF_ESS | BRF_PRG },
	{"25.bin",      0x020000, 0x9ef8f772, 1 | BRF_ESS | BRF_PRG },
	{"29.bin",      0x020000, 0x7d9c479c, 1 | BRF_ESS | BRF_PRG },
	{"24.bin",      0x020000, 0x93579684, 1 | BRF_ESS | BRF_PRG },
	{"28.bin",      0x020000, 0xff728865, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },
 
	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2rb2) STD_ROM_FN(Sf2rb2)

void sf2rb2_patch()
{
	*((unsigned short*)(CpsRom + 0xe5332)) = 0x6014;
}

int sf2rb2Init()
{
	pCpsInitCallback = sf2rb2_patch;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsSf2rb2={
	"sf2rb2", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (Rainbow set 2)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2rb2RomInfo, Sf2rb2RomName, sf2InputInfo, Sf2ceDIPInfo, 
	sf2rb2Init, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (Red Wave)

static struct BurnRomInfo Sf2redRomDesc[] = {
	{"sf2red.23",  0x080000, 0x40276abb, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"sf2red.22",  0x080000, 0x18daf387, 1 | BRF_ESS | BRF_PRG },
	{"sf2red.21",  0x080000, 0x52c486bb, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin", 0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin", 0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin", 0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin", 0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin", 0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin", 0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin", 0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin", 0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin", 0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin", 0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin", 0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin", 0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin", 0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin", 0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin", 0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2red) STD_ROM_FN(Sf2red)

struct BurnDriver BurnDrvCpsSf2red={
	"sf2red", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (Red Wave)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2redRomInfo, Sf2redRomName, sf2InputInfo, Sf2ceDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (V004)

static struct BurnRomInfo Sf2v004RomDesc[] = {
	{"sf2v004.23", 0x080000, 0x52d19f2c, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"sf2v004.22", 0x080000, 0x4b26fde7, 1 | BRF_ESS | BRF_PRG },
	{"sf2red.21",  0x080000, 0x52c486bb, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin", 0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin", 0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin", 0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin", 0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin", 0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin", 0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin", 0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin", 0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin", 0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin", 0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin", 0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin", 0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin", 0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin", 0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin", 0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2v004) STD_ROM_FN(Sf2v004)

struct BurnDriver BurnDrvCpsSf2v004={
	"sf2v004", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (V004)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2v004RomInfo, Sf2v004RomName, sf2InputInfo, Sf2ceDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (Accelerator Pt.II)

static struct BurnRomInfo Sf2accp2RomDesc[] = {
	{"sf2ca-23.bin", 0x080000, 0x36c3ba2f, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"sf2ca-22.bin", 0x080000, 0x0550453d, 1 | BRF_ESS | BRF_PRG },
	{"sf2ca-21.bin", 0x040000, 0x4c1c43ba, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",   0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",   0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",   0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",   0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",   0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",   0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",   0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",   0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",   0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",   0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",   0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",   0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",   0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",   0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",   0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2accp2) STD_ROM_FN(Sf2accp2)

void Sf2accp2_patch()
{
	*((unsigned short*)(CpsRom + 0x11756)) = 0x4e71;
}

int Sf2accp2Init()
{
	pCpsInitCallback = Sf2accp2_patch;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsSf2accp2={
	"sf2accp2", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (Accelerator Pt.II)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2accp2RomInfo, Sf2accp2RomName, sf2InputInfo, Sf2ceDIPInfo, 
	Sf2accp2Init, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (M1)

static struct BurnRomInfo Sf2m1RomDesc[] = {
	{"222e", 	0x080000, 0x1e20d0a3, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"196e", 	0x080000, 0x88cc38a3, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2m1) STD_ROM_FN(Sf2m1)

void sf2m1_fix()
{
	bootleg_progload();

#if 0
	// hack the program rom to make it run
	unsigned char patch_data[0x10C] = {
		0x3C, 0x03, 0xED, 0x33, 0x26, 0x00, 0x80, 0x00,	0x00, 0x01, 0xED, 0x33, 0x2E, 0x00, 0x80, 0x00,
		0x08, 0x01, 0xED, 0x33, 0x2A, 0x00, 0x80, 0x00,	0x04, 0x01, 0xED, 0x33, 0x28, 0x00, 0x80, 0x00,
		0x02, 0x01, 0xED, 0x33, 0x2C, 0x00, 0x80, 0x00,	0x06, 0x01, 0xED, 0x33, 0x64, 0x00, 0x80, 0x00,
		0x14, 0x01, 0xED, 0x33, 0x5C, 0x00, 0x80, 0x00,	0x0C, 0x01, 0xED, 0x33, 0x60, 0x00, 0x80, 0x00,
		0x10, 0x01, 0xED, 0x33, 0x62, 0x00, 0x80, 0x00,	0x12, 0x01, 0xED, 0x33, 0x5E, 0x00, 0x80, 0x00,
		0x0E, 0x01, 0xED, 0x33, 0x66, 0x00, 0x80, 0x00,	0x16, 0x01, 0xED, 0x33, 0x70, 0x00, 0x80, 0x00,
		0x20, 0x01, 0x2D, 0x10, 0xCE, 0x02, 0x2D, 0x12,	0xDB, 0x02, 0x00, 0xB3, 0x40, 0x1B, 0x89, 0x00,
		0x48, 0xE1, 0x6D, 0x80, 0x48, 0x00, 0xC0, 0x33,	0x80, 0x00, 0x22, 0x01, 0xED, 0x33, 0x4E, 0x00,
		0x80, 0x00, 0x66, 0x01, 0xED, 0x13, 0xD7, 0x02,	0x80, 0x00, 0x30, 0x00, 0xED, 0x33, 0x50, 0x00,
		0x80, 0x00, 0x68, 0x01, 0xED, 0x33, 0x52, 0x00,	0x80, 0x00, 0x6A, 0x01, 0xED, 0x33, 0x54, 0x00,
		0x80, 0x00, 0x6C, 0x01, 0xED, 0x33, 0x56, 0x00,	0x80, 0x00, 0x6E, 0x01, 0x00, 0x61, 0x78, 0x0E,
		0xED, 0x33, 0x58, 0x00, 0x80, 0x00, 0x70, 0x01,	0xED, 0x33, 0x30, 0x00, 0x80, 0x00, 0x0A, 0x01,
		0xB8, 0x4E, 0x68, 0x14, 0xB8, 0x4E, 0x02, 0x63,	0x00, 0x61, 0x32, 0x11, 0x00, 0x61, 0x40, 0x12,
		0x00, 0x61, 0xBC, 0x0F, 0x2D, 0x52, 0x1C, 0x00,	0xED, 0x50, 0x0E, 0x82, 0x2D, 0x42, 0xDE, 0x02,
		0xED, 0x41, 0x00, 0x80, 0x3C, 0x30, 0x0F, 0x00,	0x28, 0x0C, 0x01, 0x00, 0x00, 0x00, 0x0C, 0x66,
		0x28, 0x53, 0x01, 0x00, 0x06, 0x66, 0x7C, 0x11,	0x04, 0x00, 0x00, 0x00, 0xE8, 0x41, 0x20, 0x00,
		0xC8, 0x51, 0xE6, 0xFF, 0x3C, 0x30, 0x50, 0x00, 0xC8, 0x51, 0xFE, 0xFF
	};

	memcpy (CpsRom + 0x5d8, patch_data, 0x10C);

	*((unsigned short*)(CpsRom + 0x15A4)) = 0xFFC0; // 0x0000 - BG Alignment - Wofh kludge
	*((unsigned short*)(CpsRom + 0x6322)) = 0x0181; // 0x0006 - SFX		 - Cawingb kludge
#endif
}

int sf2m1Init()
{
	pCpsInitCallback = sf2m1_fix;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsSf2m1={
	"sf2m1", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (M1)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
        NULL, Sf2m1RomInfo, Sf2m1RomName, sf2InputInfo, Sf2cehacksDIPInfo, 
	sf2m1Init, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (M2)

static struct BurnRomInfo Sf2m2RomDesc[] = {
	{"ch222esp",    0x080000, 0x9e6d058a, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"ch196esp",    0x080000, 0xed2ff437, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2m2) STD_ROM_FN(Sf2m2)

void sf2m2_patch()
{
	bootleg_progload();

	*((unsigned short*)(CpsRom + 0xc0670)) = 0x4e71;
}

int sf2m2Init()
{
	pCpsInitCallback = sf2m2_patch;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsSf2m2={
	"sf2m2", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (M2)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
        NULL, Sf2m2RomInfo, Sf2m2RomName, Sf2yycInputInfo, Sf2cehacksDIPInfo, 
	sf2m2Init, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (M3)

static struct BurnRomInfo Sf2m3RomDesc[] = {
	{"u222chp",     0x080000, 0xdb567b66, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"u196chp",     0x080000, 0x95ea597e, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2m3) STD_ROM_FN(Sf2m3)

int sf2m3Init()
{
	pCpsInitCallback = bootleg_progload;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsSf2m3={
	"sf2m3", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (M3)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
        NULL, Sf2m3RomInfo, Sf2m3RomName, Sf2yycInputInfo, Sf2cehacksDIPInfo, 
	sf2m3Init, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (M4)

static struct BurnRomInfo Sf2m4RomDesc[] = {
	{"u222ne",      0x080000, 0x7133489e, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"u196ne",      0x080000, 0xb07a4f90, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2m4) STD_ROM_FN(Sf2m4)

struct BurnDriver BurnDrvCpsSf2m4={
	"sf2m4", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (M4)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
        NULL, Sf2m4RomInfo, Sf2m4RomName, Sf2yycInputInfo, Sf2cehacksDIPInfo, 
	sf2m3Init, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (M5)

static struct BurnRomInfo Sf2m5RomDesc[] = {
	{"u222",        0x080000, 0x03991fba, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"u196",        0x080000, 0x39f15a1e, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2m5) STD_ROM_FN(Sf2m5)

struct BurnDriver BurnDrvCpsSf2m5={
	"sf2m5", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (M5)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
        NULL, Sf2m5RomInfo, Sf2m5RomName, Sf2yycInputInfo, Sf2cehacksDIPInfo, 
	sf2m3Init, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (M6)

static struct BurnRomInfo Sf2m6RomDesc[] = {
	{"u222-6b",     0x080000, 0x0a3692be, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"u196-6b",     0x080000, 0x80454da7, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2m6) STD_ROM_FN(Sf2m6)

struct BurnDriver BurnDrvCpsSf2m6={
	"sf2m6", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (M6)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
        NULL, Sf2m6RomInfo, Sf2m6RomName, Sf2yycInputInfo, Sf2cehacksDIPInfo, 
	sf2m3Init, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (M7)

static struct BurnRomInfo Sf2m7RomDesc[] = {
	{"u222-2i",     0x040000, 0x1ca7adbd, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"u196-2i",     0x040000, 0xf758408c, 1 | BRF_ESS | BRF_PRG },
	{"u222-2s",     0x040000, 0x720cea3e, 1 | BRF_ESS | BRF_PRG },
	{"u196-2s",     0x040000, 0x9932832c, 1 | BRF_ESS | BRF_PRG },
	{"s92_21a.bin", 0x080000, 0x925a7877, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2m7) STD_ROM_FN(Sf2m7)

struct BurnDriver BurnDrvCpsSf2m7={
	"sf2m7", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (M7)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
        NULL, Sf2m7RomInfo, Sf2m7RomName, Sf2yycInputInfo, Sf2cehacksDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (YYC)

static struct BurnRomInfo Sf2yycRomDesc[] = {
	{"b12.rom",     0x080000, 0x8f742fd5, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"b14.rom",     0x080000, 0x8831ec7f, 1 | BRF_ESS | BRF_PRG },
	{"b11.rom",     0x020000, 0x94a46525, 1 | BRF_ESS | BRF_PRG },
	{"b13.rom",     0x020000, 0x8fb3dd47, 1 | BRF_ESS | BRF_PRG },
	{"b11.rom",     0x020000, 0x94a46525, 1 | BRF_ESS | BRF_PRG },
	{"b13.rom",     0x020000, 0x8fb3dd47, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2yyc) STD_ROM_FN(Sf2yyc)

void sf2yyc_patch()
{
	bootleg_progload();

        // Patch invalid opcodes (protection?)
	*((unsigned short*)(CpsRom + 0xe55be)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55ca)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55cc)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55ce)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55d0)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55d2)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55d4)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55d6)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55d8)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55da)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55de)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55e2)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55e4)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55e6)) = 0x4e71;
	*((unsigned short*)(CpsRom + 0xe55e8)) = 0x4e71;
}

int sf2yycInit()
{
	pCpsInitCallback = sf2yyc_patch;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsSf2yyc={
	"sf2yyc", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (YYC)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
        NULL, Sf2yycRomInfo, Sf2yycRomName, Sf2yycInputInfo, Sf2cehacksDIPInfo, 
	sf2yycInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Champion Edition (Kouryu)

static struct BurnRomInfo KouryuRomDesc[] = {
	{"u222.rom",    0x080000, 0x9236a79a, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"u196.rom",    0x080000, 0xb23a869d, 1 | BRF_ESS | BRF_PRG },
	{"u221.rom",    0x020000, 0x64e6e091, 1 | BRF_ESS | BRF_PRG },
	{"u195.rom",    0x020000, 0xc95e4443, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s92_10.bin",  0x080000, 0x960687d5, 2 | BRF_GRA },
	{"s92_11.bin",  0x080000, 0x978ecd18, 2 | BRF_GRA },
	{"s92_12.bin",  0x080000, 0xd6ec9a0a, 2 | BRF_GRA },
	{"s92_13.bin",  0x080000, 0xed2c67f6, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Kouryu) STD_ROM_FN(Kouryu)

struct BurnDriver BurnDrvCpsKouryu={
	"sf2koryu", "sf2ce", NULL, "1992",
	"Street Fighter II': Champion Edition (Kouryu)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_BOOTLEG, 2, HARDWARE_CAPCOM_CPS1, 
        NULL, KouryuRomInfo, KouryuRomName, Sf2yycInputInfo, Sf2cehacksDIPInfo, 
	sf2m3Init, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Varth - Operation Thunderstorm (World 920714)

static struct BurnRomInfo VarthRomDesc[] = {
	{ "vae_30b.rom", 0x020000, 0xadb8d391, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "vae_35b.rom", 0x020000, 0x44e5548f, 1 | BRF_ESS | BRF_PRG },
	{ "vae_31b.rom", 0x020000, 0x1749a71c, 1 | BRF_ESS | BRF_PRG },
	{ "vae_36b.rom", 0x020000, 0x5f2e2450, 1 | BRF_ESS | BRF_PRG },
	{ "vae_28b.rom", 0x020000, 0xe524ca50, 1 | BRF_ESS | BRF_PRG },
	{ "vae_33b.rom", 0x020000, 0xc0bbf8c9, 1 | BRF_ESS | BRF_PRG },
	{ "vae_29b.rom", 0x020000, 0x6640996a, 1 | BRF_ESS | BRF_PRG },
	{ "vae_34b.rom", 0x020000, 0xfa59be8a, 1 | BRF_ESS | BRF_PRG },

	{ "va_gfx5.rom", 0x080000, 0xb1fb726e, 2 | BRF_GRA },
	{ "va_gfx7.rom", 0x080000, 0x4c6588cd, 2 | BRF_GRA },
	{ "va_gfx1.rom", 0x080000, 0x0b1ace37, 2 | BRF_GRA },
	{ "va_gfx3.rom", 0x080000, 0x44dfe706, 2 | BRF_GRA },

	{ "va_09.rom",   0x010000, 0x7a99446e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "va_18.rom",   0x020000, 0xde30510e, 4 | BRF_SND },
	{ "va_19.rom",   0x020000, 0x0610a4ac, 4 | BRF_SND },
};

STD_ROM_PICK(Varth) STD_ROM_FN(Varth)

struct BurnDriver BurnDrvCpsVarth = {
	"varth", NULL, NULL, "1992",
	"Varth - Operation Thunderstorm (World 920714)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_ORIENTATION_VERTICAL, 2, HARDWARE_CAPCOM_CPS1,
	NULL,VarthRomInfo,VarthRomName,varthInputInfo, varthDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// Varth - Operation Thunderstorm (World 920612)

static struct BurnRomInfo varthr1RomDesc[] = {
	{ "vae_30a.rom", 0x020000, 0x7fcd0091, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "vae_35a.rom", 0x020000, 0x35cf9509, 1 | BRF_ESS | BRF_PRG },
	{ "vae_31a.rom", 0x020000, 0x15e5ee81, 1 | BRF_ESS | BRF_PRG },
	{ "vae_36a.rom", 0x020000, 0x153a201e, 1 | BRF_ESS | BRF_PRG },
	{ "vae_28a.rom", 0x020000, 0x7a0e0d25, 1 | BRF_ESS | BRF_PRG },
	{ "vae_33a.rom", 0x020000, 0xf2365922, 1 | BRF_ESS | BRF_PRG },
	{ "vae_29a.rom", 0x020000, 0x5e2cd2c3, 1 | BRF_ESS | BRF_PRG },
	{ "vae_34a.rom", 0x020000, 0x3d9bdf83, 1 | BRF_ESS | BRF_PRG },

	{ "va_gfx5.rom", 0x080000, 0xb1fb726e, 2 | BRF_GRA },
	{ "va_gfx7.rom", 0x080000, 0x4c6588cd, 2 | BRF_GRA },
	{ "va_gfx1.rom", 0x080000, 0x0b1ace37, 2 | BRF_GRA },
	{ "va_gfx3.rom", 0x080000, 0x44dfe706, 2 | BRF_GRA },

	{ "va_09.rom",   0x010000, 0x7a99446e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "va_18.rom",   0x020000, 0xde30510e, 4 | BRF_SND },
	{ "va_19.rom",   0x020000, 0x0610a4ac, 4 | BRF_SND },
};

STD_ROM_PICK(varthr1) STD_ROM_FN(varthr1)

struct BurnDriver BurnDrvCpsvarthr1 = {
	"varthr1", "varth", NULL, "1992",
	"Varth - Operation Thunderstorm (World 920612)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL, 2, HARDWARE_CAPCOM_CPS1,
	NULL,varthr1RomInfo,varthr1RomName,varthInputInfo, varthDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// Varth - Operation Thunderstorm (US 920612)

static struct BurnRomInfo VarthuRomDesc[] = {
	{ "vau23a.bin",  0x080000, 0xfbe68726, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "vae_28a.rom", 0x020000, 0x7a0e0d25, 1 | BRF_ESS | BRF_PRG },
	{ "vae_33a.rom", 0x020000, 0xf2365922, 1 | BRF_ESS | BRF_PRG },
	{ "vae_29a.rom", 0x020000, 0x5e2cd2c3, 1 | BRF_ESS | BRF_PRG },
	{ "vae_34a.rom", 0x020000, 0x3d9bdf83, 1 | BRF_ESS | BRF_PRG },

	{ "va_gfx5.rom", 0x080000, 0xb1fb726e, 2 | BRF_GRA },
	{ "va_gfx7.rom", 0x080000, 0x4c6588cd, 2 | BRF_GRA },
	{ "va_gfx1.rom", 0x080000, 0x0b1ace37, 2 | BRF_GRA },
	{ "va_gfx3.rom", 0x080000, 0x44dfe706, 2 | BRF_GRA },

	{ "va_09.rom",   0x010000, 0x7a99446e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "va_18.rom",   0x020000, 0xde30510e, 4 | BRF_SND },
	{ "va_19.rom",   0x020000, 0x0610a4ac, 4 | BRF_SND },
};

STD_ROM_PICK(Varthu) STD_ROM_FN(Varthu)

struct BurnDriver BurnDrvCpsVarthu = {
	"varthu", "varth", NULL, "1992",
	"Varth - Operation Thunderstorm (US 920612)\0", NULL, "Capcom (Romstar license)", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL, 2, HARDWARE_CAPCOM_CPS1,
	NULL,VarthuRomInfo,VarthuRomName,varthInputInfo, varthDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// Varth - Operation Thunderstorm (Japan 920714)

static struct BurnRomInfo VarthjRomDesc[] = {
	{ "vaj36b.bin",  0x020000, 0x1d798d6a, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "vaj42b.bin",  0x020000, 0x0f720233, 1 | BRF_ESS | BRF_PRG },
	{ "vaj37b.bin",  0x020000, 0x24414b17, 1 | BRF_ESS | BRF_PRG },
	{ "vaj43b.bin",  0x020000, 0x34b4b06c, 1 | BRF_ESS | BRF_PRG },
	{ "vaj34b.bin",  0x020000, 0x87c79aed, 1 | BRF_ESS | BRF_PRG },
	{ "vaj40b.bin",  0x020000, 0x210b4bd0, 1 | BRF_ESS | BRF_PRG },
	{ "vaj35b.bin",  0x020000, 0x6b0da69f, 1 | BRF_ESS | BRF_PRG },
	{ "vaj41b.bin",  0x020000, 0x6542c8a4, 1 | BRF_ESS | BRF_PRG },

	{ "va_gfx5.rom", 0x080000, 0xb1fb726e, 2 | BRF_GRA },
	{ "va_gfx7.rom", 0x080000, 0x4c6588cd, 2 | BRF_GRA },
	{ "va_gfx1.rom", 0x080000, 0x0b1ace37, 2 | BRF_GRA },
	{ "va_gfx3.rom", 0x080000, 0x44dfe706, 2 | BRF_GRA },

	{ "va_09.rom",   0x010000, 0x7a99446e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "va_18.rom",   0x020000, 0xde30510e, 4 | BRF_SND },
	{ "va_19.rom",   0x020000, 0x0610a4ac, 4 | BRF_SND },
};

STD_ROM_PICK(Varthj) STD_ROM_FN(Varthj)

struct BurnDriver BurnDrvCpsVarthj = {
	"varthj", "varth", NULL, "1992",
	"Varth - Operation Thunderstorm (Japan 920714)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL, 2 ,HARDWARE_CAPCOM_CPS1,
	NULL,VarthjRomInfo,VarthjRomName,varthInputInfo, varthDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,224,384,3,4
};


// Capcom World 2 (Japan 920611)

static struct BurnRomInfo Cworld2jRomDesc[] = {
	{ "q536.bin", 0x020000, 0x38a08099, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "q542.bin", 0x020000, 0x4d29b3a4, 1 | BRF_ESS | BRF_PRG },
	{ "q537.bin", 0x020000, 0xeb547ebc, 1 | BRF_ESS | BRF_PRG },
	{ "q543.bin", 0x020000, 0x3ef65ea8, 1 | BRF_ESS | BRF_PRG },
	{ "q534.bin", 0x020000, 0x7fcc1317, 1 | BRF_ESS | BRF_PRG },
	{ "q540.bin", 0x020000, 0x7f14b7b4, 1 | BRF_ESS | BRF_PRG },
	{ "q535.bin", 0x020000, 0xabacee26, 1 | BRF_ESS | BRF_PRG },
	{ "q541.bin", 0x020000, 0xd3654067, 1 | BRF_ESS | BRF_PRG },

	{ "q509.bin", 0x020000, 0x48496d80, 2 | BRF_GRA },
	{ "q501.bin", 0x020000, 0xc5453f56, 2 | BRF_GRA },
	{ "q513.bin", 0x020000, 0xc741ac52, 2 | BRF_GRA },
	{ "q505.bin", 0x020000, 0x143e068f, 2 | BRF_GRA },
	{ "q524.bin", 0x020000, 0xb419d139, 2 | BRF_GRA },
	{ "q517.bin", 0x020000, 0xbd3b4d11, 2 | BRF_GRA },
	{ "q538.bin", 0x020000, 0x9c24670c, 2 | BRF_GRA },
	{ "q532.bin", 0x020000, 0x3ef9c7c2, 2 | BRF_GRA },
	{ "q510.bin", 0x020000, 0x119e5e93, 2 | BRF_GRA },
	{ "q502.bin", 0x020000, 0xa2cadcbe, 2 | BRF_GRA },
	{ "q514.bin", 0x020000, 0xa8755f82, 2 | BRF_GRA },
	{ "q506.bin", 0x020000, 0xc92a91fc, 2 | BRF_GRA },
	{ "q525.bin", 0x020000, 0x979237cb, 2 | BRF_GRA },
	{ "q518.bin", 0x020000, 0xc57da03c, 2 | BRF_GRA },
	{ "q539.bin", 0x020000, 0xa5839b25, 2 | BRF_GRA },
	{ "q533.bin", 0x020000, 0x04d03930, 2 | BRF_GRA },

	{ "q523.bin", 0x010000, 0xe14dc524, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "q530.bin", 0x020000, 0xd10c1b68, 4 | BRF_SND },
	{ "q531.bin", 0x020000, 0x7d17e496, 4 | BRF_SND },
};

STD_ROM_PICK(Cworld2j) STD_ROM_FN(Cworld2j)

struct BurnDriver BurnDrvCpsCworld2j = {
	"cworld2j", NULL, NULL, "1992",
	"Capcom World 2 (Japan 920611)\0", NULL, "Capcom", "CPS1",
	L"Capcom World 2 (\u30AF\u30A4\u30BA\uFF15 \u3042\u3069\u3079\u3093\u3061\u3083\u30FC\u304F\u3044\u305A \u304B\u3077\u3053\u3093\u308F\u30FC\u308B\u3069\uFF12 Japan 920611)\0", NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,Cworld2jRomInfo,Cworld2jRomName,cworld2jInputInfo, cworld2jDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Warriors of Fate (World 921002)

static struct BurnRomInfo WofRomDesc[] = {
	{ "tk2e_23b.rom", 0x080000, 0x11fb2ed1, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "tk2e_22b.rom", 0x080000, 0x479b3f24, 1 | BRF_ESS | BRF_PRG },

	{ "tk2_gfx1.rom", 0x080000, 0x0d9cb9bf, 2 | BRF_GRA },
	{ "tk2_gfx3.rom", 0x080000, 0x45227027, 2 | BRF_GRA },
	{ "tk2_gfx2.rom", 0x080000, 0xc5ca2460, 2 | BRF_GRA },
	{ "tk2_gfx4.rom", 0x080000, 0xe349551c, 2 | BRF_GRA },
	{ "tk2_gfx5.rom", 0x080000, 0x291f0f0b, 2 | BRF_GRA },
	{ "tk2_gfx7.rom", 0x080000, 0x3edeb949, 2 | BRF_GRA },
	{ "tk2_gfx6.rom", 0x080000, 0x1abd14d6, 2 | BRF_GRA },
	{ "tk2_gfx8.rom", 0x080000, 0xb27948e3, 2 | BRF_GRA },

	{ "tk2_qa.rom",   0x020000, 0xc9183a0d, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "tk2_q1.rom",   0x080000, 0x611268cf, 5 | BRF_SND },
	{ "tk2_q2.rom",   0x080000, 0x20f55ca9, 5 | BRF_SND },
	{ "tk2_q3.rom",   0x080000, 0xbfcf6f52, 5 | BRF_SND },
	{ "tk2_q4.rom",   0x080000, 0x36642e88, 5 | BRF_SND },
};

STD_ROM_PICK(Wof) STD_ROM_FN(Wof)

int wofInit()
{
	pCpsInitCallback = wof_decode;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsWof = {
	"wof", NULL, NULL, "1992",
	"Warriors of Fate (World 921002)\0", NULL, "Capcom", "CPS1 / QSound",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,3,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,WofRomInfo,WofRomName,wofInputInfo, wofDIPInfo,
	wofInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Sangokushi II (Asia 921005)

static struct BurnRomInfo WofaRomDesc[] = {
	{ "tk2a_23b.rom", 0x080000, 0x2e024628, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "tk2a_22b.rom", 0x080000, 0x900ad4cd, 1 | BRF_ESS | BRF_PRG },

	{ "tk2_gfx1.rom", 0x080000, 0x0d9cb9bf, 2 | BRF_GRA },
	{ "tk2_gfx3.rom", 0x080000, 0x45227027, 2 | BRF_GRA },
	{ "tk2_gfx2.rom", 0x080000, 0xc5ca2460, 2 | BRF_GRA },
	{ "tk2_gfx4.rom", 0x080000, 0xe349551c, 2 | BRF_GRA },
	{ "tk2_gfx5.rom", 0x080000, 0x291f0f0b, 2 | BRF_GRA },
	{ "tk2_gfx7.rom", 0x080000, 0x3edeb949, 2 | BRF_GRA },
	{ "tk2_gfx6.rom", 0x080000, 0x1abd14d6, 2 | BRF_GRA },
	{ "tk2_gfx8.rom", 0x080000, 0xb27948e3, 2 | BRF_GRA },

	{ "tk2_qa.rom",   0x020000, 0xc9183a0d, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "tk2_q1.rom",   0x080000, 0x611268cf, 5 | BRF_SND },
	{ "tk2_q2.rom",   0x080000, 0x20f55ca9, 5 | BRF_SND },
	{ "tk2_q3.rom",   0x080000, 0xbfcf6f52, 5 | BRF_SND },
	{ "tk2_q4.rom",   0x080000, 0x36642e88, 5 | BRF_SND },
};

STD_ROM_PICK(Wofa) STD_ROM_FN(Wofa)

struct BurnDriver BurnDrvCpsWofa = {
	"wofa", "wof", NULL, "1992",
	"Sangokushi II (Asia 921005)\0", NULL, "Capcom", "CPS1 / QSound",
	L"\u4E09\u56FD\u5FD7II (Asia 921005)\0Sangokushi II\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,3,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,WofaRomInfo,WofaRomName,wofInputInfo, wofDIPInfo,
	wofInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Warriors of Fate (US 921031)

static struct BurnRomInfo WofuRomDesc[] = {
	{ "tk2u.23c",     0x080000, 0x29b89c12, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "tk2u.22c",     0x080000, 0xf5af4774, 1 | BRF_ESS | BRF_PRG },

	{ "tk2_gfx1.rom", 0x080000, 0x0d9cb9bf, 2 | BRF_GRA },
	{ "tk2_gfx3.rom", 0x080000, 0x45227027, 2 | BRF_GRA },
	{ "tk2_gfx2.rom", 0x080000, 0xc5ca2460, 2 | BRF_GRA },
	{ "tk2_gfx4.rom", 0x080000, 0xe349551c, 2 | BRF_GRA },
	{ "tk2_gfx5.rom", 0x080000, 0x291f0f0b, 2 | BRF_GRA },
	{ "tk2_gfx7.rom", 0x080000, 0x3edeb949, 2 | BRF_GRA },
	{ "tk2_gfx6.rom", 0x080000, 0x1abd14d6, 2 | BRF_GRA },
	{ "tk2_gfx8.rom", 0x080000, 0xb27948e3, 2 | BRF_GRA },

	{ "tk2_qa.rom",   0x020000, 0xc9183a0d, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "tk2_q1.rom",   0x080000, 0x611268cf, 5 | BRF_SND },
	{ "tk2_q2.rom",   0x080000, 0x20f55ca9, 5 | BRF_SND },
	{ "tk2_q3.rom",   0x080000, 0xbfcf6f52, 5 | BRF_SND },
	{ "tk2_q4.rom",   0x080000, 0x36642e88, 5 | BRF_SND },
};

STD_ROM_PICK(Wofu) STD_ROM_FN(Wofu)

struct BurnDriver BurnDrvCpsWofu = {
	"wofu", "wof", NULL, "1992",
	"Warriors of Fate (US 921031)\0", NULL, "Capcom", "CPS1 / QSound",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,3,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,WofuRomInfo,WofuRomName,wofInputInfo, wofDIPInfo,
	wofInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Tenchi wo Kurau II - Sekiheki no Tatakai (Japan 921031)

static struct BurnRomInfo WofjRomDesc[] = {
	{ "tk2j23c.bin",  0x080000, 0x9b215a68, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "tk2j22c.bin",  0x080000, 0xb74b09ac, 1 | BRF_ESS | BRF_PRG },

	{ "tk2_gfx1.rom", 0x080000, 0x0d9cb9bf, 2 | BRF_GRA },
	{ "tk2_gfx3.rom", 0x080000, 0x45227027, 2 | BRF_GRA },
	{ "tk2_gfx2.rom", 0x080000, 0xc5ca2460, 2 | BRF_GRA },
	{ "tk2_gfx4.rom", 0x080000, 0xe349551c, 2 | BRF_GRA },
	{ "tk205.bin",    0x080000, 0xe4a44d53, 2 | BRF_GRA },
	{ "tk206.bin",    0x080000, 0x58066ba8, 2 | BRF_GRA },
	{ "tk207.bin",    0x080000, 0xd706568e, 2 | BRF_GRA },
	{ "tk208.bin",    0x080000, 0xd4a19a02, 2 | BRF_GRA },

	{ "tk2_qa.rom",   0x020000, 0xc9183a0d, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "tk2_q1.rom",   0x080000, 0x611268cf, 5 | BRF_SND },
	{ "tk2_q2.rom",   0x080000, 0x20f55ca9, 5 | BRF_SND },
	{ "tk2_q3.rom",   0x080000, 0xbfcf6f52, 5 | BRF_SND },
	{ "tk2_q4.rom",   0x080000, 0x36642e88, 5 | BRF_SND },
};

STD_ROM_PICK(Wofj) STD_ROM_FN(Wofj)

struct BurnDriver BurnDrvCpsWofj = {
	"wofj", "wof", NULL, "1992",
	"Tenchi wo Kurau II - Sekiheki no Tatakai (Japan 921031)\0", NULL, "Capcom", "CPS1 / QSound",
	L"\u8EE2\u5730\u3092\u55B0\u3089\u3046II - \u8D64\u58C1\u306E\u6226\u3044 (Japan 921031)\0Tenchi wo Kurau II - Sekiheki no Tatakai\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,3,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,WofjRomInfo,WofjRomName,wofInputInfo, wofDIPInfo,
	wofInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter II': Hyper Fighting (World 921209)

static struct BurnRomInfo Sf2hfRomDesc[] = {
	{"s92e_23a.bin", 0x080000, 0x2dd72514, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"sf2_22.bin",   0x080000, 0xaea6e035, 1 | BRF_ESS | BRF_PRG },
	{"sf2_21.bin",   0x080000, 0xfd200288, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",   0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",   0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",   0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",   0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",   0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",   0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",   0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",   0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s2t_10.bin",   0x080000, 0x3c042686, 2 | BRF_GRA },
	{"s2t_11.bin",   0x080000, 0x8b7e7183, 2 | BRF_GRA },
	{"s2t_12.bin",   0x080000, 0x293c888c, 2 | BRF_GRA },
	{"s2t_13.bin",   0x080000, 0x842b35a4, 2 | BRF_GRA },

	{"s92_09.bin",   0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",   0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",   0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2hf) STD_ROM_FN(Sf2hf)

struct BurnDriver BurnDrvCpsSf2hf={
	"sf2hf", "sf2ce", NULL, "1992",
	"Street Fighter II': Hyper Fighting (World 921209)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2hfRomInfo, Sf2hfRomName, sf2InputInfo, Sf2ceDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II': Hyper Fighting (US 921209)

static struct BurnRomInfo Sf2tRomDesc[] = {
	{"sf2.23" ,    0x080000, 0x89a1fc38, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"sf2_22.bin", 0x080000, 0xaea6e035, 1 | BRF_ESS | BRF_PRG },
	{"sf2_21.bin", 0x080000, 0xfd200288, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin", 0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin", 0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin", 0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin", 0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin", 0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin", 0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin", 0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin", 0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s2t_10.bin", 0x080000, 0x3c042686, 2 | BRF_GRA },
	{"s2t_11.bin", 0x080000, 0x8b7e7183, 2 | BRF_GRA },
	{"s2t_12.bin", 0x080000, 0x293c888c, 2 | BRF_GRA },
	{"s2t_13.bin", 0x080000, 0x842b35a4, 2 | BRF_GRA },

	{"s92_09.bin", 0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin", 0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin", 0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2t) STD_ROM_FN(Sf2t)

struct BurnDriver BurnDrvCpsSf2t={
	"sf2t", "sf2hf", NULL, "1992",
	"Street Fighter II': Hyper Fighting (US 921209)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2tRomInfo, Sf2tRomName, sf2InputInfo, Sf2ceDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Street Fighter II' Turbo: Hyper Fighting (Japan 921209)

static struct BurnRomInfo Sf2tjRomDesc[] = {
	{"s2tj_23.bin", 0x080000, 0xea73b4dc, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{"s2t_22.bin",  0x080000, 0xaea6e035, 1 | BRF_ESS | BRF_PRG },
	{"s2t_21.bin",  0x080000, 0xfd200288, 1 | BRF_ESS | BRF_PRG },

	{"s92_01.bin",  0x080000, 0x03b0d852, 2 | BRF_GRA },
	{"s92_02.bin",  0x080000, 0x840289ec, 2 | BRF_GRA },
	{"s92_03.bin",  0x080000, 0xcdb5f027, 2 | BRF_GRA },
	{"s92_04.bin",  0x080000, 0xe2799472, 2 | BRF_GRA },
	{"s92_05.bin",  0x080000, 0xba8a2761, 2 | BRF_GRA },
	{"s92_06.bin",  0x080000, 0xe584bfb5, 2 | BRF_GRA },
	{"s92_07.bin",  0x080000, 0x21e3f87d, 2 | BRF_GRA },
	{"s92_08.bin",  0x080000, 0xbefc47df, 2 | BRF_GRA },
	{"s2t_10.bin",  0x080000, 0x3c042686, 2 | BRF_GRA },
	{"s2t_11.bin",  0x080000, 0x8b7e7183, 2 | BRF_GRA },
	{"s2t_12.bin",  0x080000, 0x293c888c, 2 | BRF_GRA },
	{"s2t_13.bin",  0x080000, 0x842b35a4, 2 | BRF_GRA },

	{"s92_09.bin",  0x010000, 0x08f6b60e, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{"s92_18.bin",  0x020000, 0x7f162009, 4 | BRF_SND },
	{"s92_19.bin",  0x020000, 0xbeade53f, 4 | BRF_SND },
};

STD_ROM_PICK(Sf2tj) STD_ROM_FN(Sf2tj)

struct BurnDriver BurnDrvCpsSf2tj={
	"sf2tj", "sf2hf", NULL, "1992",
	"Street Fighter II' Turbo: Hyper Fighting (Japan 921209)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE, 2, HARDWARE_CAPCOM_CPS1, 
	NULL, Sf2tjRomInfo, Sf2tjRomName, sf2InputInfo, Sf2cejDIPInfo, 
	CpsInit, CpsExit, Cps1Frame, CpsRedraw, CpsAreaScan, 
	&CpsRecalcPal, 384, 224, 19, 14
};


// Cadillacs and Dinosaurs (World 930201)

static struct BurnRomInfo DinoRomDesc[] = {
	{ "cde_23a.rom",  0x080000, 0x8f4e585e, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "cde_22a.rom",  0x080000, 0x9278aa12, 1 | BRF_ESS | BRF_PRG },
	{ "cde_21a.rom",  0x080000, 0x66d23de2, 1 | BRF_ESS | BRF_PRG },

	{ "cd_gfx01.rom", 0x080000, 0x8da4f917, 2 | BRF_GRA },
	{ "cd_gfx03.rom", 0x080000, 0x6c40f603, 2 | BRF_GRA },
	{ "cd_gfx02.rom", 0x080000, 0x09c8fc2d, 2 | BRF_GRA },
	{ "cd_gfx04.rom", 0x080000, 0x637ff38f, 2 | BRF_GRA },
	{ "cd_gfx05.rom", 0x080000, 0x470befee, 2 | BRF_GRA },
	{ "cd_gfx07.rom", 0x080000, 0x22bfb7a3, 2 | BRF_GRA },
	{ "cd_gfx06.rom", 0x080000, 0xe7599ac4, 2 | BRF_GRA },
	{ "cd_gfx08.rom", 0x080000, 0x211b4b15, 2 | BRF_GRA },

	{ "cd_q.rom",     0x020000, 0x605fdb0b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "cd_q1.rom",    0x080000, 0x60927775, 5 | BRF_SND },
	{ "cd_q2.rom",    0x080000, 0x770f4c47, 5 | BRF_SND },
	{ "cd_q3.rom",    0x080000, 0x2f273ffc, 5 | BRF_SND },
	{ "cd_q4.rom",    0x080000, 0x2c67821d, 5 | BRF_SND },
};

STD_ROM_PICK(Dino) STD_ROM_FN(Dino)

int dinoInit()
{
	pCpsInitCallback = dino_decode;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsDino = {
	"dino", NULL, NULL, "1993",
	"Cadillacs and Dinosaurs (World 930201)\0", NULL, "Capcom", "CPS1 / QSound",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,3,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,DinoRomInfo,DinoRomName,dinoInputInfo, dinoDIPInfo,
	dinoInit, CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Cadillacs and Dinosaurs (US 930201)

static struct BurnRomInfo DinouRomDesc[] = {
	{ "cdu.23a",      0x080000, 0x7c2543cd, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "cdu.22a",      0x080000, 0xfab740a9, 1 | BRF_ESS | BRF_PRG },
	{ "cde_21a.rom",  0x080000, 0x66d23de2, 1 | BRF_ESS | BRF_PRG },

	{ "cd_gfx01.rom", 0x080000, 0x8da4f917, 2 | BRF_GRA },
	{ "cd_gfx03.rom", 0x080000, 0x6c40f603, 2 | BRF_GRA },
	{ "cd_gfx02.rom", 0x080000, 0x09c8fc2d, 2 | BRF_GRA },
	{ "cd_gfx04.rom", 0x080000, 0x637ff38f, 2 | BRF_GRA },
	{ "cd_gfx05.rom", 0x080000, 0x470befee, 2 | BRF_GRA },
	{ "cd_gfx07.rom", 0x080000, 0x22bfb7a3, 2 | BRF_GRA },
	{ "cd_gfx06.rom", 0x080000, 0xe7599ac4, 2 | BRF_GRA },
	{ "cd_gfx08.rom", 0x080000, 0x211b4b15, 2 | BRF_GRA },

	{ "cd_q.rom",     0x020000, 0x605fdb0b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "cd_q1.rom",    0x080000, 0x60927775, 5 | BRF_SND },
	{ "cd_q2.rom",    0x080000, 0x770f4c47, 5 | BRF_SND },
	{ "cd_q3.rom",    0x080000, 0x2f273ffc, 5 | BRF_SND },
	{ "cd_q4.rom",    0x080000, 0x2c67821d, 5 | BRF_SND },
};

STD_ROM_PICK(Dinou) STD_ROM_FN(Dinou)

struct BurnDriver BurnDrvCpsDinou = {
	"dinou", "dino", NULL, "1993",
	"Cadillacs and Dinosaurs (US 930201)\0", NULL, "Capcom", "CPS1 / QSound",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,3,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,DinouRomInfo,DinouRomName,dinoInputInfo, dinoDIPInfo,
	dinoInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Cadillacs Kyouryuu-Shinseiki (Japan 930201)

static struct BurnRomInfo DinojRomDesc[] = {
	{ "cdj-23a.8f",   0x080000, 0x5f3ece96, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "cdj-22a.7f",   0x080000, 0xa0d8de29, 1 | BRF_ESS | BRF_PRG },
	{ "cde_21a.rom",  0x080000, 0x66d23de2, 1 | BRF_ESS | BRF_PRG },

	{ "cd_gfx01.rom", 0x080000, 0x8da4f917, 2 | BRF_GRA },
	{ "cd_gfx03.rom", 0x080000, 0x6c40f603, 2 | BRF_GRA },
	{ "cd_gfx02.rom", 0x080000, 0x09c8fc2d, 2 | BRF_GRA },
	{ "cd_gfx04.rom", 0x080000, 0x637ff38f, 2 | BRF_GRA },
	{ "cd_gfx05.rom", 0x080000, 0x470befee, 2 | BRF_GRA },
	{ "cd_gfx07.rom", 0x080000, 0x22bfb7a3, 2 | BRF_GRA },
	{ "cd_gfx06.rom", 0x080000, 0xe7599ac4, 2 | BRF_GRA },
	{ "cd_gfx08.rom", 0x080000, 0x211b4b15, 2 | BRF_GRA },

	{ "cd_q.rom",     0x020000, 0x605fdb0b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "cd_q1.rom",    0x080000, 0x60927775, 5 | BRF_SND },
	{ "cd_q2.rom",    0x080000, 0x770f4c47, 5 | BRF_SND },
	{ "cd_q3.rom",    0x080000, 0x2f273ffc, 5 | BRF_SND },
	{ "cd_q4.rom",    0x080000, 0x2c67821d, 5 | BRF_SND },
};

STD_ROM_PICK(Dinoj) STD_ROM_FN(Dinoj)

struct BurnDriver BurnDrvCpsDinoj = {
	"dinoj", "dino", NULL, "1993",
	"Cadillacs Kyouryuu-Shinseiki (Japan 930201)\0", NULL, "Capcom", "CPS1, QSound",
	L"Cadillacs \u6050\u7ADC\u65B0\u4E16\u7D00 (Japan 930201)\0Cadillacs Kyouryuu-Shinseiki\0", NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,3,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,DinojRomInfo,DinojRomName,dinoInputInfo, dinoDIPInfo,
	dinoInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// The Punisher (World 930422)

static struct BurnRomInfo PunisherRomDesc[] = {
	{ "pse_26.rom",  0x020000, 0x389a99d2, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "pse_30.rom",  0x020000, 0x68fb06ac, 1 | BRF_ESS | BRF_PRG },
	{ "pse_27.rom",  0x020000, 0x3eb181c3, 1 | BRF_ESS | BRF_PRG },
	{ "pse_31.rom",  0x020000, 0x37108e7b, 1 | BRF_ESS | BRF_PRG },
	{ "pse_24.rom",  0x020000, 0x0f434414, 1 | BRF_ESS | BRF_PRG },
	{ "pse_28.rom",  0x020000, 0xb732345d, 1 | BRF_ESS | BRF_PRG },
	{ "pse_25.rom",  0x020000, 0xb77102e2, 1 | BRF_ESS | BRF_PRG },
	{ "pse_29.rom",  0x020000, 0xec037bce, 1 | BRF_ESS | BRF_PRG },
	{ "ps_21.rom",   0x080000, 0x8affa5a9, 1 | BRF_ESS | BRF_PRG },

	{ "ps_gfx1.rom", 0x080000, 0x77b7ccab, 2 | BRF_GRA },
	{ "ps_gfx3.rom", 0x080000, 0x0122720b, 2 | BRF_GRA },
	{ "ps_gfx2.rom", 0x080000, 0x64fa58d4, 2 | BRF_GRA },
	{ "ps_gfx4.rom", 0x080000, 0x60da42c8, 2 | BRF_GRA },
	{ "ps_gfx5.rom", 0x080000, 0xc54ea839, 2 | BRF_GRA },
	{ "ps_gfx7.rom", 0x080000, 0x04c5acbd, 2 | BRF_GRA },
	{ "ps_gfx6.rom", 0x080000, 0xa544f4cc, 2 | BRF_GRA },
	{ "ps_gfx8.rom", 0x080000, 0x8f02f436, 2 | BRF_GRA },

	{ "ps_q.rom",    0x020000, 0x49ff4446, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ps_q1.rom",   0x080000, 0x31fd8726, 5 | BRF_SND },
	{ "ps_q2.rom",   0x080000, 0x980a9eef, 5 | BRF_SND },
	{ "ps_q3.rom",   0x080000, 0x0dd44491, 5 | BRF_SND },
	{ "ps_q4.rom",   0x080000, 0xbed42f03, 5 | BRF_SND },
};

STD_ROM_PICK(Punisher) STD_ROM_FN(Punisher)

int PunisherInit()
{
	pCpsInitCallback = punisher_decode;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsPunisher = {
	"punisher", NULL, NULL, "1993",
	"The Punisher (World 930422)\0", NULL, "Capcom", "CPS1 / QSound",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,PunisherRomInfo,PunisherRomName,punisherInputInfo, punishDIPInfo,
	PunisherInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// The Punisher (US 930422)

static struct BurnRomInfo PunishruRomDesc[] = {
	{ "psu26.rom",   0x020000, 0x9236d121, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "psu30.rom",   0x020000, 0x8320e501, 1 | BRF_ESS | BRF_PRG },
	{ "psu27.rom",   0x020000, 0x61c960a1, 1 | BRF_ESS | BRF_PRG },
	{ "psu31.rom",   0x020000, 0x78d4c298, 1 | BRF_ESS | BRF_PRG },
	{ "psu24.rom",   0x020000, 0x1cfecad7, 1 | BRF_ESS | BRF_PRG },
	{ "psu28.rom",   0x020000, 0xbdf921c1, 1 | BRF_ESS | BRF_PRG },
	{ "psu25.rom",   0x020000, 0xc51acc94, 1 | BRF_ESS | BRF_PRG },
	{ "psu29.rom",   0x020000, 0x52dce1ca, 1 | BRF_ESS | BRF_PRG },
	{ "ps_21.rom",   0x080000, 0x8affa5a9, 1 | BRF_ESS | BRF_PRG }, 

	{ "ps_gfx1.rom", 0x080000, 0x77b7ccab, 2 | BRF_GRA },
	{ "ps_gfx3.rom", 0x080000, 0x0122720b, 2 | BRF_GRA },
	{ "ps_gfx2.rom", 0x080000, 0x64fa58d4, 2 | BRF_GRA },
	{ "ps_gfx4.rom", 0x080000, 0x60da42c8, 2 | BRF_GRA },
	{ "ps_gfx5.rom", 0x080000, 0xc54ea839, 2 | BRF_GRA },
	{ "ps_gfx7.rom", 0x080000, 0x04c5acbd, 2 | BRF_GRA },
	{ "ps_gfx6.rom", 0x080000, 0xa544f4cc, 2 | BRF_GRA },
	{ "ps_gfx8.rom", 0x080000, 0x8f02f436, 2 | BRF_GRA },

	{ "ps_q.rom",    0x020000, 0x49ff4446, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ps_q1.rom",   0x080000, 0x31fd8726, 5 | BRF_SND },
	{ "ps_q2.rom",   0x080000, 0x980a9eef, 5 | BRF_SND },
	{ "ps_q3.rom",   0x080000, 0x0dd44491, 5 | BRF_SND },
	{ "ps_q4.rom",   0x080000, 0xbed42f03, 5 | BRF_SND },
};

STD_ROM_PICK(Punishru) STD_ROM_FN(Punishru)

struct BurnDriver BurnDrvCpsPunishru = {
	"punishru", "punisher", NULL, "1993",
	"The Punisher (US 930422)\0", NULL, "Capcom", "CPS1 / QSound",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,PunishruRomInfo,PunishruRomName,punisherInputInfo, punishDIPInfo,
	PunisherInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// The Punisher (Japan 930422)

static struct BurnRomInfo PunishrjRomDesc[] = {
	{ "psj23.bin",   0x080000, 0x6b2fda52, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "psj22.bin",   0x080000, 0xe01036bc, 1 | BRF_ESS | BRF_PRG },
	{ "ps_21.rom",   0x080000, 0x8affa5a9, 1 | BRF_ESS | BRF_PRG },

	{ "ps_gfx1.rom", 0x080000, 0x77b7ccab, 2 | BRF_GRA },
	{ "ps_gfx3.rom", 0x080000, 0x0122720b, 2 | BRF_GRA },
	{ "ps_gfx2.rom", 0x080000, 0x64fa58d4, 2 | BRF_GRA },
	{ "ps_gfx4.rom", 0x080000, 0x60da42c8, 2 | BRF_GRA },
	{ "ps_gfx5.rom", 0x080000, 0xc54ea839, 2 | BRF_GRA },
	{ "ps_gfx7.rom", 0x080000, 0x04c5acbd, 2 | BRF_GRA },
	{ "ps_gfx6.rom", 0x080000, 0xa544f4cc, 2 | BRF_GRA },
	{ "ps_gfx8.rom", 0x080000, 0x8f02f436, 2 | BRF_GRA },

	{ "ps_q.rom",    0x020000, 0x49ff4446, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "ps_q1.rom",   0x080000, 0x31fd8726, 5 | BRF_SND },
	{ "ps_q2.rom",   0x080000, 0x980a9eef, 5 | BRF_SND },
	{ "ps_q3.rom",   0x080000, 0x0dd44491, 5 | BRF_SND },
	{ "ps_q4.rom",   0x080000, 0xbed42f03, 5 | BRF_SND },
};

STD_ROM_PICK(Punishrj) STD_ROM_FN(Punishrj)

struct BurnDriver BurnDrvCpsPunishrj = {
	"punishrj", "punisher", NULL, "1993",
	"The Punisher (Japan 930422)\0", NULL, "Capcom", "CPS1 / QSound",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1_QSOUND,
	NULL,PunishrjRomInfo,PunishrjRomName,punisherInputInfo, punishDIPInfo,
	PunisherInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Saturday Night Slam Masters (World 930713)

static struct BurnRomInfo SlammastRomDesc[] = {
	{ "mbe_23e.rom",  0x080000, 0x5394057a, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "mbe_24b.rom",  0x020000, 0x95d5e729, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_28b.rom",  0x020000, 0xb1c7cbcb, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_25b.rom",  0x020000, 0xa50d3fd4, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_29b.rom",  0x020000, 0x08e32e56, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_21a.rom",  0x080000, 0xd5007b05, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_20a.rom",  0x080000, 0xaeb557b0, 1 | BRF_ESS | BRF_PRG },

	{ "mb_gfx01.rom", 0x080000, 0x41468e06, 2 | BRF_GRA },
	{ "mb_gfx03.rom", 0x080000, 0xf453aa9e, 2 | BRF_GRA },
	{ "mb_gfx02.rom", 0x080000, 0x2ffbfea8, 2 | BRF_GRA },
	{ "mb_gfx04.rom", 0x080000, 0x1eb9841d, 2 | BRF_GRA },
	{ "mb_05.bin",    0x080000, 0x506b9dc9, 2 | BRF_GRA },
	{ "mb_07.bin",    0x080000, 0xaff8c2fb, 2 | BRF_GRA },
	{ "mb_06.bin",    0x080000, 0xb76c70e9, 2 | BRF_GRA },
	{ "mb_08.bin",    0x080000, 0xe60c9556, 2 | BRF_GRA },
	{ "mb_10.bin",    0x080000, 0x97976ff5, 2 | BRF_GRA },
	{ "mb_12.bin",    0x080000, 0xb350a840, 2 | BRF_GRA },
	{ "mb_11.bin",    0x080000, 0x8fb94743, 2 | BRF_GRA },
	{ "mb_13.bin",    0x080000, 0xda810d5f, 2 | BRF_GRA },

	{ "mb_qa.rom",    0x020000, 0xe21a03c4, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "mb_q1.bin",    0x080000, 0x0630c3ce, 5 | BRF_SND },
	{ "mb_q2.bin",    0x080000, 0x354f9c21, 5 | BRF_SND },
	{ "mb_q3.bin",    0x080000, 0x7838487c, 5 | BRF_SND },
	{ "mb_q4.bin",    0x080000, 0xab66e087, 5 | BRF_SND },
	{ "mb_q5.bin",    0x080000, 0xc789fef2, 5 | BRF_SND },
	{ "mb_q6.bin",    0x080000, 0xecb81b61, 5 | BRF_SND },
	{ "mb_q7.bin",    0x080000, 0x041e49ba, 5 | BRF_SND },
	{ "mb_q8.bin",    0x080000, 0x59fe702a, 5 | BRF_SND },
};

STD_ROM_PICK(Slammast) STD_ROM_FN(Slammast)

void slampro_decode()
{
	memcpy (CpsEncZRom, CpsZRom, 0x40000);

	slammast_decode();
}

int SlammastInit()
{
	pCpsInitCallback = slampro_decode;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsSlammast = {
	"slammast", NULL, NULL, "1993",
	"Saturday Night Slam Masters (World 930713)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,4,HARDWARE_CAPCOM_CPS1,
	NULL,SlammastRomInfo,SlammastRomName,slammastInputInfo, slammastDIPInfo,
	SlammastInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Saturday Night Slam Masters (US 930713)

static struct BurnRomInfo SlammasuRomDesc[] = {
	{ "mbu-23e.rom",  0x080000, 0x224f0062, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "mbe_24b.rom",  0x020000, 0x95d5e729, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_28b.rom",  0x020000, 0xb1c7cbcb, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_25b.rom",  0x020000, 0xa50d3fd4, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_29b.rom",  0x020000, 0x08e32e56, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_21a.rom",  0x080000, 0xd5007b05, 1 | BRF_ESS | BRF_PRG },
	{ "mbu-20a.rom",  0x080000, 0xfc848af5, 1 | BRF_ESS | BRF_PRG },

	{ "mb_gfx01.rom", 0x080000, 0x41468e06, 2 | BRF_GRA },
	{ "mb_gfx03.rom", 0x080000, 0xf453aa9e, 2 | BRF_GRA },
	{ "mb_gfx02.rom", 0x080000, 0x2ffbfea8, 2 | BRF_GRA },
	{ "mb_gfx04.rom", 0x080000, 0x1eb9841d, 2 | BRF_GRA },
	{ "mb_05.bin",    0x080000, 0x506b9dc9, 2 | BRF_GRA },
	{ "mb_07.bin",    0x080000, 0xaff8c2fb, 2 | BRF_GRA },
	{ "mb_06.bin",    0x080000, 0xb76c70e9, 2 | BRF_GRA },
	{ "mb_08.bin",    0x080000, 0xe60c9556, 2 | BRF_GRA },
	{ "mb_10.bin",    0x080000, 0x97976ff5, 2 | BRF_GRA },
	{ "mb_12.bin",    0x080000, 0xb350a840, 2 | BRF_GRA },
	{ "mb_11.bin",    0x080000, 0x8fb94743, 2 | BRF_GRA },
	{ "mb_13.bin",    0x080000, 0xda810d5f, 2 | BRF_GRA },

	{ "mb_qa.rom",    0x020000, 0xe21a03c4, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "mb_q1.bin",    0x080000, 0x0630c3ce, 5 | BRF_SND },
	{ "mb_q2.bin",    0x080000, 0x354f9c21, 5 | BRF_SND },
	{ "mb_q3.bin",    0x080000, 0x7838487c, 5 | BRF_SND },
	{ "mb_q4.bin",    0x080000, 0xab66e087, 5 | BRF_SND },
	{ "mb_q5.bin",    0x080000, 0xc789fef2, 5 | BRF_SND },
	{ "mb_q6.bin",    0x080000, 0xecb81b61, 5 | BRF_SND },
	{ "mb_q7.bin",    0x080000, 0x041e49ba, 5 | BRF_SND },
	{ "mb_q8.bin",    0x080000, 0x59fe702a, 5 | BRF_SND },
};

STD_ROM_PICK(Slammasu) STD_ROM_FN(Slammasu)

struct BurnDriver BurnDrvCpsSlammasu = {
	"slammasu", "slammast", NULL, "1993",
	"Saturday Night Slam Masters (US 930713)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,4,HARDWARE_CAPCOM_CPS1,
	NULL,SlammasuRomInfo,SlammasuRomName,slammastInputInfo, slammastDIPInfo,
	SlammastInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Muscle Bomber - The Body Explosion (Japan 930713)

static struct BurnRomInfo MbomberjRomDesc[] = {
	{ "mbj23e",      0x080000, 0x0d06036a, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "mbe_24b.rom", 0x020000, 0x95d5e729, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_28b.rom", 0x020000, 0xb1c7cbcb, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_25b.rom", 0x020000, 0xa50d3fd4, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_29b.rom", 0x020000, 0x08e32e56, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_21a.rom", 0x080000, 0xd5007b05, 1 | BRF_ESS | BRF_PRG },
	{ "mbe_20a.rom", 0x080000, 0xaeb557b0, 1 | BRF_ESS | BRF_PRG },

	{ "mbj_01.bin",  0x080000, 0xa53b1c81, 2 | BRF_GRA },
	{ "mbj_03.bin",  0x080000, 0x23fe10f6, 2 | BRF_GRA },
	{ "mbj_02.bin",  0x080000, 0xcb866c2f, 2 | BRF_GRA },
	{ "mbj_04.bin",  0x080000, 0xc9143e75, 2 | BRF_GRA },
	{ "mb_05.bin",   0x080000, 0x506b9dc9, 2 | BRF_GRA },
	{ "mb_07.bin",   0x080000, 0xaff8c2fb, 2 | BRF_GRA },
	{ "mb_06.bin",   0x080000, 0xb76c70e9, 2 | BRF_GRA },
	{ "mb_08.bin",   0x080000, 0xe60c9556, 2 | BRF_GRA },
	{ "mb_10.bin",   0x080000, 0x97976ff5, 2 | BRF_GRA },
	{ "mb_12.bin",   0x080000, 0xb350a840, 2 | BRF_GRA },
	{ "mb_11.bin",   0x080000, 0x8fb94743, 2 | BRF_GRA },
	{ "mb_13.bin",   0x080000, 0xda810d5f, 2 | BRF_GRA },

	{ "mb_qa.rom",   0x020000, 0xe21a03c4, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "mb_q1.bin",   0x080000, 0x0630c3ce, 5 | BRF_SND },
	{ "mb_q2.bin",   0x080000, 0x354f9c21, 5 | BRF_SND },
	{ "mb_q3.bin",   0x080000, 0x7838487c, 5 | BRF_SND },
	{ "mb_q4.bin",   0x080000, 0xab66e087, 5 | BRF_SND },
	{ "mb_q5.bin",   0x080000, 0xc789fef2, 5 | BRF_SND },
	{ "mb_q6.bin",   0x080000, 0xecb81b61, 5 | BRF_SND },
	{ "mb_q7.bin",   0x080000, 0x041e49ba, 5 | BRF_SND },
	{ "mb_q8.bin",   0x080000, 0x59fe702a, 5 | BRF_SND },
};

STD_ROM_PICK(Mbomberj) STD_ROM_FN(Mbomberj)

struct BurnDriver BurnDrvCpsMbomberj = {
	"mbomberj", "slammast", NULL, "1993",
	"Muscle Bomber - The Body Explosion (Japan 930713)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,4,HARDWARE_CAPCOM_CPS1,
	NULL,MbomberjRomInfo,MbomberjRomName,slammastInputInfo, slammastDIPInfo,
	SlammastInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Muscle Bomber Duo - Ultimate Team Battle (World 931206)

static struct BurnRomInfo MbombrdRomDesc[] = {
	{ "mbd_26.bin",   0x020000, 0x72b7451c, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "mbde_30.rom",  0x020000, 0xa036dc16, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_27.bin",   0x020000, 0x4086f534, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_31.bin",   0x020000, 0x085f47f0, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_24.bin",   0x020000, 0xc20895a5, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_28.bin",   0x020000, 0x2618d5e1, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_25.bin",   0x020000, 0x9bdb6b11, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_29.bin",   0x020000, 0x3f52d5e5, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_21.bin",   0x080000, 0x690c026a, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_20.bin",   0x080000, 0xb8b2139b, 1 | BRF_ESS | BRF_PRG },

	{ "mb_gfx01.rom", 0x080000, 0x41468e06, 2 | BRF_GRA },
	{ "mb_gfx03.rom", 0x080000, 0xf453aa9e, 2 | BRF_GRA },
	{ "mb_gfx02.rom", 0x080000, 0x2ffbfea8, 2 | BRF_GRA },
	{ "mb_gfx04.rom", 0x080000, 0x1eb9841d, 2 | BRF_GRA },
	{ "mb_05.bin",    0x080000, 0x506b9dc9, 2 | BRF_GRA },
	{ "mb_07.bin",    0x080000, 0xaff8c2fb, 2 | BRF_GRA },
	{ "mb_06.bin",    0x080000, 0xb76c70e9, 2 | BRF_GRA },
	{ "mb_08.bin",    0x080000, 0xe60c9556, 2 | BRF_GRA },
	{ "mb_10.bin",    0x080000, 0x97976ff5, 2 | BRF_GRA },
	{ "mb_12.bin",    0x080000, 0xb350a840, 2 | BRF_GRA },
	{ "mb_11.bin",    0x080000, 0x8fb94743, 2 | BRF_GRA },
	{ "mb_13.bin",    0x080000, 0xda810d5f, 2 | BRF_GRA },

	{ "mb_q.bin",     0x020000, 0xd6fa76d1, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "mb_q1.bin",    0x080000, 0x0630c3ce, 5 | BRF_SND },
	{ "mb_q2.bin",    0x080000, 0x354f9c21, 5 | BRF_SND },
	{ "mb_q3.bin",    0x080000, 0x7838487c, 5 | BRF_SND },
	{ "mb_q4.bin",    0x080000, 0xab66e087, 5 | BRF_SND },
	{ "mb_q5.bin",    0x080000, 0xc789fef2, 5 | BRF_SND },
	{ "mb_q6.bin",    0x080000, 0xecb81b61, 5 | BRF_SND },
	{ "mb_q7.bin",    0x080000, 0x041e49ba, 5 | BRF_SND },
	{ "mb_q8.bin",    0x080000, 0x59fe702a, 5 | BRF_SND },
};

STD_ROM_PICK(Mbombrd) STD_ROM_FN(Mbombrd)

struct BurnDriver BurnDrvCpsMbombrd = {
	"mbombrd", "slammast", NULL, "1993",
	"Muscle Bomber Duo - Ultimate Team Battle (World 931206)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,4,HARDWARE_CAPCOM_CPS1,
	NULL,MbombrdRomInfo,MbombrdRomName,slammastInputInfo, slammastDIPInfo,
	SlammastInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Muscle Bomber Duo - Heat Up Warriors (Japan 931206)

static struct BurnRomInfo MbombrdjRomDesc[] = {
	{ "mbd_26.bin",   0x020000, 0x72b7451c, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "mbdj_30.bin",  0x020000, 0xbeff31cf, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_27.bin",   0x020000, 0x4086f534, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_31.bin",   0x020000, 0x085f47f0, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_24.bin",   0x020000, 0xc20895a5, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_28.bin",   0x020000, 0x2618d5e1, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_25.bin",   0x020000, 0x9bdb6b11, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_29.bin",   0x020000, 0x3f52d5e5, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_21.bin",   0x080000, 0x690c026a, 1 | BRF_ESS | BRF_PRG },
	{ "mbd_20.bin",   0x080000, 0xb8b2139b, 1 | BRF_ESS | BRF_PRG },

	{ "mbj_01.bin",   0x080000, 0xa53b1c81, 2 | BRF_GRA },
	{ "mbj_02.bin",   0x080000, 0xcb866c2f, 2 | BRF_GRA },
	{ "mbj_03.bin",   0x080000, 0x23fe10f6, 2 | BRF_GRA },  
	{ "mbj_04.bin",   0x080000, 0xc9143e75, 2 | BRF_GRA },
	{ "mb_05.bin",    0x080000, 0x506b9dc9, 2 | BRF_GRA },
	{ "mb_07.bin",    0x080000, 0xaff8c2fb, 2 | BRF_GRA },
	{ "mb_06.bin",    0x080000, 0xb76c70e9, 2 | BRF_GRA },
	{ "mb_08.bin",    0x080000, 0xe60c9556, 2 | BRF_GRA },
	{ "mb_10.bin",    0x080000, 0x97976ff5, 2 | BRF_GRA },
	{ "mb_12.bin",    0x080000, 0xb350a840, 2 | BRF_GRA },
	{ "mb_11.bin",    0x080000, 0x8fb94743, 2 | BRF_GRA },
	{ "mb_13.bin",    0x080000, 0xda810d5f, 2 | BRF_GRA },

	{ "mb_q.bin",     0x020000, 0xd6fa76d1, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "mb_q1.bin",    0x080000, 0x0630c3ce, 5 | BRF_SND },
	{ "mb_q2.bin",    0x080000, 0x354f9c21, 5 | BRF_SND },
	{ "mb_q3.bin",    0x080000, 0x7838487c, 5 | BRF_SND },
	{ "mb_q4.bin",    0x080000, 0xab66e087, 5 | BRF_SND },
	{ "mb_q5.bin",    0x080000, 0xc789fef2, 5 | BRF_SND },
	{ "mb_q6.bin",    0x080000, 0xecb81b61, 5 | BRF_SND },
	{ "mb_q7.bin",    0x080000, 0x041e49ba, 5 | BRF_SND },
	{ "mb_q8.bin",    0x080000, 0x59fe702a, 5 | BRF_SND },
};

STD_ROM_PICK(Mbombrdj) STD_ROM_FN(Mbombrdj)

struct BurnDriver BurnDrvCpsMbombrdj = {
	"mbombrdj", "slammast", NULL, "1993",
	"Muscle Bomber Duo - Heat Up Warriors (Japan 931206)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,4,HARDWARE_CAPCOM_CPS1,
	NULL,MbombrdjRomInfo,MbombrdjRomName,slammastInputInfo, slammastDIPInfo,
	SlammastInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Pnickies (Japan 940608)

static struct BurnRomInfo PnickjRomDesc[] = {
	{ "pnij36.bin", 0x020000, 0x2d4ffb2b, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "pnij42.bin", 0x020000, 0xc085dfaf, 1 | BRF_ESS | BRF_PRG },

	{ "pnij09.bin", 0x020000, 0x48177b0a, 2 | BRF_GRA },
	{ "pnij01.bin", 0x020000, 0x01a0f311, 2 | BRF_GRA },
	{ "pnij13.bin", 0x020000, 0x406451b0, 2 | BRF_GRA },
	{ "pnij05.bin", 0x020000, 0x8c515dc0, 2 | BRF_GRA },
	{ "pnij26.bin", 0x020000, 0xe2af981e, 2 | BRF_GRA },
	{ "pnij18.bin", 0x020000, 0xf17a0e56, 2 | BRF_GRA },
	{ "pnij38.bin", 0x020000, 0xeb75bd8c, 2 | BRF_GRA },
	{ "pnij32.bin", 0x020000, 0x84560bef, 2 | BRF_GRA },
	{ "pnij10.bin", 0x020000, 0xc2acc171, 2 | BRF_GRA },
	{ "pnij02.bin", 0x020000, 0x0e21fc33, 2 | BRF_GRA },
	{ "pnij14.bin", 0x020000, 0x7fe59b19, 2 | BRF_GRA },
	{ "pnij06.bin", 0x020000, 0x79f4bfe3, 2 | BRF_GRA },
	{ "pnij27.bin", 0x020000, 0x83d5cb0e, 2 | BRF_GRA },
	{ "pnij19.bin", 0x020000, 0xaf08b230, 2 | BRF_GRA },
	{ "pnij39.bin", 0x020000, 0x70fbe579, 2 | BRF_GRA },
	{ "pnij33.bin", 0x020000, 0x3ed2c680, 2 | BRF_GRA },

	{ "pnij17.bin", 0x010000, 0xe86f787a, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "pnij24.bin", 0x020000, 0x5092257d, 4 | BRF_SND },
	{ "pnij25.bin", 0x020000, 0x22109aaa, 4 | BRF_SND },
};

STD_ROM_PICK(Pnickj) STD_ROM_FN(Pnickj)

struct BurnDriver BurnDrvCpsPnickj = {
	"pnickj", NULL, NULL, "1994",
	"Pnickies (Japan 940608)\0", NULL, "Compile (Capcom license)", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,PnickjRomInfo,PnickjRomName,pnickjInputInfo, pnickjDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Quiz & Dragons (US 920701)

static struct BurnRomInfo QadRomDesc[] = {
	{ "qdu_36a.rom", 0x020000, 0xde9c24a0, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "qdu_42a.rom", 0x020000, 0xcfe36f0c, 1 | BRF_ESS | BRF_PRG },
	{ "qdu_37a.rom", 0x020000, 0x10d22320, 1 | BRF_ESS | BRF_PRG },
	{ "qdu_43a.rom", 0x020000, 0x15e6beb9, 1 | BRF_ESS | BRF_PRG },

	{ "qdu_09.rom",  0x020000, 0x8c3f9f44, 2 | BRF_GRA },
	{ "qdu_01.rom",  0x020000, 0xf688cf8f, 2 | BRF_GRA },
	{ "qdu_13.rom",  0x020000, 0xafbd551b, 2 | BRF_GRA },
	{ "qdu_05.rom",  0x020000, 0xc3db0910, 2 | BRF_GRA },
	{ "qdu_24.rom",  0x020000, 0x2f1bd0ec, 2 | BRF_GRA },
	{ "qdu_17.rom",  0x020000, 0xa812f9e2, 2 | BRF_GRA },
	{ "qdu_38.rom",  0x020000, 0xccdddd1f, 2 | BRF_GRA },
	{ "qdu_32.rom",  0x020000, 0xa8d295d3, 2 | BRF_GRA },

	{ "qdu_23.rom",  0x010000, 0xcfb5264b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "qdu_30.rom",  0x020000, 0xf190da84, 4 | BRF_SND },
	{ "qdu_31.rom",  0x020000, 0xb7583f73, 4 | BRF_SND },
};

STD_ROM_PICK(Qad) STD_ROM_FN(Qad)

struct BurnDriver BurnDrvCpsQad = {
	"qad", NULL, NULL, "1992",
	"Quiz & Dragons (US 920701)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,QadRomInfo,QadRomName,qadInputInfo, qadDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Quiz & Dragons (Japan 940921)

static struct BurnRomInfo QadjRomDesc[] = {
	{ "qad23a.bin", 0x080000, 0x4d3553de, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "qad22a.bin", 0x080000, 0x3191ddd0, 1 | BRF_ESS | BRF_PRG },

	{ "qad01.bin",  0x080000, 0x9d853b57, 2 | BRF_GRA },
	{ "qad02.bin",  0x080000, 0xb35976c4, 2 | BRF_GRA },
	{ "qad03.bin",  0x080000, 0xcea4ca8c, 2 | BRF_GRA },
	{ "qad04.bin",  0x080000, 0x41b74d1b, 2 | BRF_GRA },

	{ "qad09.bin",  0x010000, 0x733161cc, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "qad18.bin",  0x020000, 0x2bfe6f6a, 4 | BRF_SND },
	{ "qad19.bin",  0x020000, 0x13d3236b, 4 | BRF_SND },
};

STD_ROM_PICK(Qadj) STD_ROM_FN(Qadj)

struct BurnDriver BurnDrvCpsQadj = {
	"qadj", "qad", NULL, "1992",
	"Quiz & Dragons (Japan 940921)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,QadjRomInfo,QadjRomName,qadInputInfo, qadjDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Quiz Tonosama no Yabou 2 Zenkoku-ban (Japan 950123)

static struct BurnRomInfo Qtono2RomDesc[] = {
	{ "tn2j-30.11e", 0x020000, 0x9226eb5e, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "tn2j-37.11f", 0x020000, 0xd1d30da1, 1 | BRF_ESS | BRF_PRG },
	{ "tn2j-31.12e", 0x020000, 0x015e6a8a, 1 | BRF_ESS | BRF_PRG },
	{ "tn2j-38.12f", 0x020000, 0x1f139bcc, 1 | BRF_ESS | BRF_PRG },
	{ "tn2j-28.9e",  0x020000, 0x86d27f71, 1 | BRF_ESS | BRF_PRG },
	{ "tn2j-35.9f",  0x020000, 0x7a1ab87d, 1 | BRF_ESS | BRF_PRG },
	{ "tn2j-29.10e", 0x020000, 0x9c384e99, 1 | BRF_ESS | BRF_PRG },
	{ "tn2j-36.10f", 0x020000, 0x4c4b2a0a, 1 | BRF_ESS | BRF_PRG },

	{ "tn2-02m.4a",  0x080000, 0xf2016a34, 2 | BRF_GRA },
	{ "tn2-04m.6a",  0x080000, 0x094e0fb1, 2 | BRF_GRA },
	{ "tn2-01m.3a",  0x080000, 0xcb950cf9, 2 | BRF_GRA },
	{ "tn2-03m.5a",  0x080000, 0x18a5bf59, 2 | BRF_GRA },
	{ "tn2-11m.4c",  0x080000, 0xd0edd30b, 2 | BRF_GRA },
	{ "tn2-13m.6c",  0x080000, 0x426621c3, 2 | BRF_GRA },
	{ "tn2-10m.3c",  0x080000, 0xa34ece70, 2 | BRF_GRA },
	{ "tn2-12m.5c",  0x080000, 0xe04ff2f4, 2 | BRF_GRA },

	{ "tn2j-09.12a", 0x008000, 0x6d8edcef, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "tn2j-18.11c", 0x020000, 0xa40bf9a7, 4 | BRF_SND },
	{ "tn2j-19.12c", 0x020000, 0x5b3b931e, 4 | BRF_SND },
};

STD_ROM_PICK(Qtono2) STD_ROM_FN(Qtono2)

struct BurnDriver BurnDrvCpsQtono2 = {
	"qtono2", NULL, NULL, "1995",
	"Quiz Tonosama no Yabou 2 Zenkoku-ban (Japan 950123)\0", NULL, "Capcom", "CPS1",
	L"\u30AF\u30A4\u30BA\u6BBF\u69D8\u306E\u91CE\u671B\uFF12 (\u3068\u306E\u3055\u307E\u306E\u3084\u307C\u3046 2 Japan 950123)\0Quiz Tonosama no Yabou 2 Zenkoku-ban\0", NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,Qtono2RomInfo,Qtono2RomName,qadInputInfo, qtono2DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Pang! 3 (Euro 950511)

static struct BurnRomInfo Pang3RomDesc[] = {
	{ "pa3w-17.11l", 0x080000, 0x12138234, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "pa3w-16.10l", 0x080000, 0xd1ba585c, 1 | BRF_ESS | BRF_PRG },

	{ "pa3-01m.2c",  0x200000, 0x068a152c, 2 | BRF_GRA },
	{ "pa3-07m.2f",  0x200000, 0x3a4a619d, 2 | BRF_GRA },

	{ "pa3-11.11f",  0x008000, 0x90a08c46, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "pa3-05.10d",  0x020000, 0x73a10d5d, 4 | BRF_SND },
	{ "pa3-06.11d",  0x020000, 0xaffa4f82, 4 | BRF_SND },
};

STD_ROM_PICK(Pang3) STD_ROM_FN(Pang3)

struct BurnDriver BurnDrvCpsPang3 = {
	"pang3", NULL, NULL, "1995",
	"Pang! 3 (Euro 950511)\0", NULL, "Mitchell", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1_GENERIC,
	NULL,Pang3RomInfo,Pang3RomName,pang3InputInfo, pang3DIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Pang! 3 (Japan 950511)

static struct BurnRomInfo Pang3jRomDesc[] = {
	{ "pa3j-17.11l", 0x080000, 0x21f6e51f, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "pa3j-16.10l", 0x080000, 0xca1d7897, 1 | BRF_ESS | BRF_PRG },

	{ "pa3-01m.2c",  0x200000, 0x068a152c, 2 | BRF_GRA },
	{ "pa3-07m.2f",  0x200000, 0x3a4a619d, 2 | BRF_GRA },

	{ "pa3-11.11f",  0x008000, 0x90a08c46, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "pa3-05.10d",  0x020000, 0x73a10d5d, 4 | BRF_SND },
	{ "pa3-06.11d",  0x020000, 0xaffa4f82, 4 | BRF_SND },
};

STD_ROM_PICK(Pang3j) STD_ROM_FN(Pang3j)

void pang3j_decrypt()
{
	unsigned short *rom = (unsigned short *)CpsRom;
	int A,src,dst;

	for (A = 0x80000; A < 0x100000; A += 2)
	{
		// only the low 8 bits of each word are encrypted
		src = rom[A/2];
		dst = src & 0xff00;
		if ( src & 0x01) dst ^= 0x04;
		if ( src & 0x02) dst ^= 0x21;
		if ( src & 0x04) dst ^= 0x01;
		if (~src & 0x08) dst ^= 0x50;
		if ( src & 0x10) dst ^= 0x40;
		if ( src & 0x20) dst ^= 0x06;
		if ( src & 0x40) dst ^= 0x08;
		if (~src & 0x80) dst ^= 0x88;
		rom[A/2] = dst;
	}
}

int pang3jInit()
{
	pCpsInitCallback = pang3j_decrypt;

	return CpsInit();
}

struct BurnDriver BurnDrvCpsPang3j = {
	"pang3j", "pang3", NULL, "1995",
	"Pang! 3 (Japan 950511)\0", NULL, "Mitchell", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1_GENERIC,
	NULL,Pang3jRomInfo,Pang3jRomName,pang3InputInfo, pang3DIPInfo,
	pang3jInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Mega Man - The Power Battle (Asia 951006)

static struct BurnRomInfo megamanRomDesc[] = {
	{ "rcma_23b.rom", 0x080000, 0x61e4a397, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "rcma_22b.rom", 0x080000, 0x708268c4, 1 | BRF_ESS | BRF_PRG },
	{ "rcma_21a.rom", 0x080000, 0x4376ea95, 1 | BRF_ESS | BRF_PRG },

	{ "rcm_01.rom",   0x080000, 0x6ecdf13f, 2 | BRF_GRA },
	{ "rcm_02.rom",   0x080000, 0x944d4f0f, 2 | BRF_GRA },
	{ "rcm_03.rom",   0x080000, 0x36f3073c, 2 | BRF_GRA },
	{ "rcm_04.rom",   0x080000, 0x54e622ff, 2 | BRF_GRA },
	{ "rcm_05.rom",   0x080000, 0x5dd131fd, 2 | BRF_GRA },
	{ "rcm_06.rom",   0x080000, 0xf0faf813, 2 | BRF_GRA },
	{ "rcm_07.rom",   0x080000, 0x826de013, 2 | BRF_GRA },
	{ "rcm_08.rom",   0x080000, 0xfbff64cf, 2 | BRF_GRA },
	{ "rcm_10.rom",   0x080000, 0x4dc8ada9, 2 | BRF_GRA },
	{ "rcm_11.rom",   0x080000, 0xf2b9ee06, 2 | BRF_GRA },
	{ "rcm_12.rom",   0x080000, 0xfed5f203, 2 | BRF_GRA },
	{ "rcm_13.rom",   0x080000, 0x5069d4a9, 2 | BRF_GRA },
	{ "rcm_14.rom",   0x080000, 0x303be3bd, 2 | BRF_GRA },
	{ "rcm_15.rom",   0x080000, 0x4f2d372f, 2 | BRF_GRA },
	{ "rcm_16.rom",   0x080000, 0x93d97fde, 2 | BRF_GRA },
	{ "rcm_17.rom",   0x080000, 0x92371042, 2 | BRF_GRA },

	{ "rcm_09.rom",   0x020000, 0x9632d6ef, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "rcm_18.rom",   0x020000, 0x80f1f8aa, 4 | BRF_SND },
	{ "rcm_19.rom",   0x020000, 0xf257dbe1, 4 | BRF_SND },
};

STD_ROM_PICK(megaman) STD_ROM_FN(megaman)

struct BurnDriver BurnDrvCpsMegaman = {
	"megaman", NULL, NULL, "1995",
	"Mega Man - The Power Battle (Asia 951006)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPS1,
	NULL,megamanRomInfo,megamanRomName,megamanInputInfo, megamanDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Rockman - The Power Battle (Japan 950922)

static struct BurnRomInfo RockmanjRomDesc[] = {
	{ "rcm23a.bin", 0x080000, 0xefd96cb2, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "rcm22a.bin", 0x080000, 0x8729a689, 1 | BRF_ESS | BRF_PRG },
	{ "rcm21a.bin", 0x080000, 0x517ccde2, 1 | BRF_ESS | BRF_PRG },

	{ "rcm_01.rom", 0x080000, 0x6ecdf13f, 2 | BRF_GRA },
	{ "rcm_02.rom", 0x080000, 0x944d4f0f, 2 | BRF_GRA },
	{ "rcm_03.rom", 0x080000, 0x36f3073c, 2 | BRF_GRA },
	{ "rcm_04.rom", 0x080000, 0x54e622ff, 2 | BRF_GRA },
	{ "rcm_05.rom", 0x080000, 0x5dd131fd, 2 | BRF_GRA },
	{ "rcm_06.rom", 0x080000, 0xf0faf813, 2 | BRF_GRA },
	{ "rcm_07.rom", 0x080000, 0x826de013, 2 | BRF_GRA },
	{ "rcm_08.rom", 0x080000, 0xfbff64cf, 2 | BRF_GRA },
	{ "rcm_10.rom", 0x080000, 0x4dc8ada9, 2 | BRF_GRA },
	{ "rcm_11.rom", 0x080000, 0xf2b9ee06, 2 | BRF_GRA },
	{ "rcm_12.rom", 0x080000, 0xfed5f203, 2 | BRF_GRA },
	{ "rcm_13.rom", 0x080000, 0x5069d4a9, 2 | BRF_GRA },
	{ "rcm_14.rom", 0x080000, 0x303be3bd, 2 | BRF_GRA },
	{ "rcm_15.rom", 0x080000, 0x4f2d372f, 2 | BRF_GRA },
	{ "rcm_16.rom", 0x080000, 0x93d97fde, 2 | BRF_GRA },
	{ "rcm_17.rom", 0x080000, 0x92371042, 2 | BRF_GRA },

	{ "rcm_09.rom", 0x020000, 0x9632d6ef, 3 | BRF_ESS | BRF_PRG },	// z80 program

	{ "rcm_18.rom", 0x020000, 0x80f1f8aa, 4 | BRF_SND },
	{ "rcm_19.rom", 0x020000, 0xf257dbe1, 4 | BRF_SND },
};

STD_ROM_PICK(Rockmanj) STD_ROM_FN(Rockmanj)

struct BurnDriver BurnDrvCpsRockmanj = {
	"rockmanj", "megaman", NULL, "1995",
	"Rockman - The Power Battle (Japan 950922)\0", NULL, "Capcom", "CPS1",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPS1,
	NULL,RockmanjRomInfo,RockmanjRomName,megamanInputInfo, megamanDIPInfo,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};



//------------------------------------------------------------------------------------------
// CPS Changer hardware
//




// Street Fighter Zero (CPS Changer 951020)

static struct BurnRomInfo sfzchRomDesc[] = {
	{ "sfzch23", 0x080000, 0x1140743f, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sfza22",  0x080000, 0x8d9b2480, 1 | BRF_ESS | BRF_PRG },
	{ "sfzch21", 0x080000, 0x5435225d, 1 | BRF_ESS | BRF_PRG },
	{ "sfza20",  0x080000, 0x806e8f38, 1 | BRF_ESS | BRF_PRG },

	{ "sfz01",   0x080000, 0x0dd53e62, 2 | BRF_GRA },
	{ "sfz02",   0x080000, 0x94c31e3f, 2 | BRF_GRA },
	{ "sfz03",   0x080000, 0x9584ac85, 2 | BRF_GRA },
	{ "sfz04",   0x080000, 0xb983624c, 2 | BRF_GRA },
	{ "sfz05",   0x080000, 0x2b47b645, 2 | BRF_GRA },
	{ "sfz06",   0x080000, 0x74fd9fb1, 2 | BRF_GRA },
	{ "sfz07",   0x080000, 0xbb2c734d, 2 | BRF_GRA },
	{ "sfz08",   0x080000, 0x454f7868, 2 | BRF_GRA },
	{ "sfz10",   0x080000, 0x2a7d675e, 2 | BRF_GRA },
	{ "sfz11",   0x080000, 0xe35546c8, 2 | BRF_GRA },
	{ "sfz12",   0x080000, 0xf122693a, 2 | BRF_GRA },
	{ "sfz13",   0x080000, 0x7cf942c8, 2 | BRF_GRA },
	{ "sfz14",   0x080000, 0x09038c81, 2 | BRF_GRA },
	{ "sfz15",   0x080000, 0x1aa17391, 2 | BRF_GRA },
	{ "sfz16",   0x080000, 0x19a5abd6, 2 | BRF_GRA },
	{ "sfz17",   0x080000, 0x248b3b73, 2 | BRF_GRA },

	{ "sfz09",   0x010000, 0xc772628b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sfz18",   0x020000, 0x61022b2d, 4 | BRF_SND },
	{ "sfz19",   0x020000, 0x3b5886d5, 4 | BRF_SND },
};

STD_ROM_PICK(sfzch) STD_ROM_FN(sfzch)

struct BurnDriver BurnDrvCpsSfzch = {
	"sfzch", NULL, NULL, "1995",
	"Street Fighter Zero (CPS Changer 951020)\0", NULL, "Capcom", "CPS Changer",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING,2,HARDWARE_CAPCOM_CPSCHANGER,
	NULL,sfzchRomInfo,sfzchRomName,cpschInputInfo, NULL,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};


// Street Fighter Alpha (Publicity CPS Changer 951020)

static struct BurnRomInfo SfachRomDesc[] = {
	{ "sfach23", 0x080000, 0x02a1a853, 1 | BRF_ESS | BRF_PRG }, // 68k program
	{ "sfza22",  0x080000, 0x8d9b2480, 1 | BRF_ESS | BRF_PRG },
	{ "sfzch21", 0x080000, 0x5435225d, 1 | BRF_ESS | BRF_PRG },
	{ "sfza20",  0x080000, 0x806e8f38, 1 | BRF_ESS | BRF_PRG },

	{ "sfz01",   0x080000, 0x0dd53e62, 2 | BRF_GRA },
	{ "sfz02",   0x080000, 0x94c31e3f, 2 | BRF_GRA },
	{ "sfz03",   0x080000, 0x9584ac85, 2 | BRF_GRA },
	{ "sfz04",   0x080000, 0xb983624c, 2 | BRF_GRA },
	{ "sfz05",   0x080000, 0x2b47b645, 2 | BRF_GRA },
	{ "sfz06",   0x080000, 0x74fd9fb1, 2 | BRF_GRA },
	{ "sfz07",   0x080000, 0xbb2c734d, 2 | BRF_GRA },
	{ "sfz08",   0x080000, 0x454f7868, 2 | BRF_GRA },
	{ "sfz10",   0x080000, 0x2a7d675e, 2 | BRF_GRA },
	{ "sfz11",   0x080000, 0xe35546c8, 2 | BRF_GRA },
	{ "sfz12",   0x080000, 0xf122693a, 2 | BRF_GRA },
	{ "sfz13",   0x080000, 0x7cf942c8, 2 | BRF_GRA },
	{ "sfz14",   0x080000, 0x09038c81, 2 | BRF_GRA },
	{ "sfz15",   0x080000, 0x1aa17391, 2 | BRF_GRA },
	{ "sfz16",   0x080000, 0x19a5abd6, 2 | BRF_GRA },
	{ "sfz17",   0x080000, 0x248b3b73, 2 | BRF_GRA },

	{ "sfz09",   0x010000, 0xc772628b, 3 | BRF_ESS | BRF_PRG }, // z80 program

	{ "sfz18",   0x020000, 0x61022b2d, 4 | BRF_SND },
	{ "sfz19",   0x020000, 0x3b5886d5, 4 | BRF_SND },
};

STD_ROM_PICK(Sfach) STD_ROM_FN(Sfach)

struct BurnDriver BurnDrvCpsSfach = {
	"sfach", "sfzch", NULL, "1995",
	"Street Fighter Alpha (Publicity CPS Changer 951020)\0", NULL, "Capcom", "CPS Changer",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE,2,HARDWARE_CAPCOM_CPSCHANGER,
	NULL,SfachRomInfo,SfachRomName,cpschInputInfo, NULL,
	CpsInit,CpsExit,Cps1Frame,CpsRedraw,CpsAreaScan,
	&CpsRecalcPal,384,224,4,3
};
